"""
2022 Sebastian de Bone (QuTech)
https://github.com/sebastiandebone/ghz_prot_II
"""
# import sys
# import os.path
import copy

import numpy as np
import operations as op
import pickle
import dill
import os
import math
from termcolor import colored
import matplotlib as mpl
import matplotlib.pyplot as plt

# import seaborn as sns
from scipy.special import binom
from copy import deepcopy
import argparse
import time
import itertools
import signal
import platform
from datetime import datetime
import pandas as pd
from collections import defaultdict
import sys
from PIL import Image

# caution: path[0] is reserved for script path (or '' in REPL)
sys.path.insert(1, 'C:\\Users\\sebastiandebon\\surfdrive\\Code')    # Add the path to 'plot_settings'
# from plot_settings import plot_settings, plot_fonts
import da_search as das
import da_protocols as dap
import evaluate_protocols as ep
import simulate_protocol.protocol_recipe as pr
import simulate_protocol.run_auto_generated_protocol as ragp
import circuit_simulation.stabilizer_measurement_protocols.auto_generated_stabilizer_measurement_protocols as agsmp
import auxiliaries_and_help_files.calculate_minimum_GHZ_creation_time_at_target_fidelity as ghztime
import auxiliaries_and_help_files.calculate_attempts_per_echo as cape
import auxiliaries_and_help_files.circuit_simulator_example_code as csec
import auxiliaries_and_help_files.calculate_confidence_interval as cci
# import figures_d.figures_GHZ_prot_I.plots_paper as pp
import plot_protocol as plotp
from oopsc.threshold.plot import plot_thresholds as plot_fit_Mark
from oopsc.threshold.fit import fit_thresholds as fit_Mark
from oopsc.threshold.fit import zoomed_in_fit_thresholds as zoomed_fit_Mark

import plot_superoperator.analyse_simulation_data as asd
# import ancilla_rotations as ar
# import known_protocols as kp
# import average_bell_pair_generation_steps.protocols_plus_probabilities as ppp

ACCURACY = 15




class LoadFromFile(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        # for action in parser._actions:
        #     print(action)
        #     if isinstance(action, argparse._StoreAction):
        #         print(action.dest)
        with values as v:
            parser.parse_args([argument for argument in v.read().split() if "#" not in argument], namespace)


def parse_arguments():
    parser = argparse.ArgumentParser(description='Calculate protocol recipes.')

    parser.add_argument('-c_type', metavar='c_type', type=int, default=0,
                        help='An integer describing what type of calculation should be executed.')
    parser.add_argument('-n_max', required=False, type=int, default=4,
                        help='An integer describing the maximum number of parties involved in the search.')
    parser.add_argument('-k_max', required=False, type=int, default=4,
                        help='An integer describing the maximum number of Bell pairs involved in the search.')
    parser.add_argument('-nstate', required=False, type=int, default=10,
                        help='An integer describing the buffer size involved in the search.')
    parser.add_argument('-number_id_elems', required=False, type=int, default=1,
                        help='An integer describing how many (random) versions of the order in which Bell pairs are '
                             'created are taken into account.')

    parser.add_argument('-input_state', required=False, type=str, default="0.9",
                        help='A string describing the input Bell pair fidelity involved in the search. If this is set '
                             'to a float between 0 and 1, the Bell pairs are set to Werner states with this fidelity.')

    parser.add_argument('-da_type', required=False, type=str, default='mpuF',
                        help='A string describing the algorithm type involved in the search.')
    parser.add_argument('-rot_type', required=False, type=str, default='none',
                        help='A string describing the type of rotations of ancillary state involved in the search.')
    parser.add_argument('-calculate_prot', required=False, type=str, default='True',
                        help='A boolean describing whether or not noisy versions of the protocols should be evaluated'
                             'during the search.')

    parser.add_argument('-input_data_file', required=False, type=str, default=None,
                        help='Filename of data structure that is used as the basis for this search.')
    parser.add_argument('-n_start', required=False, type=int, default=2,
                        help='An integer describing for which n value the search should start.')
    parser.add_argument('-k_start', required=False, type=int, default=1,
                        help='An integer describing for which k value the search should start.')

    parser.add_argument('-number_of_threads', required=False, default='auto',
                        help='A parameter describing how many threads can be used simultaneously in case '
                             '-calculate_prot is set to True.')
    parser.add_argument('-filename_prefix', required=False, type=str, default='',
                        help='A parameter describing how the pickle file with results should start.')
    parser.add_argument('-failed_protocol_prefix', required=False, type=str, default='',
                        help='The folder in which we place failed protocols.')
    parser.add_argument('-max_calc_time', required=False, type=str, default=None,
                        help='A string in the format DD-HH:MM:SS or HH:MM:SS describing the maximum calculation time in number of hours.')
    parser.add_argument('-seeds', required=False, type=int, nargs="*",
                        help='An integer or string describing how many seeds of which seeds number should be used to '
                             'average over in case -calculate_prot is True.')
    parser.add_argument('-init_seed_search', required=False, type=str, default='True',
                        help='A boolean describing whether or not part of the seeds should first be used to get a first '
                             'idea of a protocol (and see if there is a chance that it is better than what we have in '
                             'the buffer so far), before the remaining seeds are used.')
    parser.add_argument('-skip_similar_protocols', required=False, type=str, default='identical',
                        help='A string describing if a protocol should be considered (i.e., considered in the search) '
                             'if it is either \'identical\' or \'similar\' (in its binary tree structure) to a '
                             'protocol that already has been considered in the search. If one wants to not skip '
                             'protocols, it is sufficient to pass \'None\', \'no\' or \'n\'.')
    parser.add_argument('-reshuffle_protocols', required=False, type=str, default='True',
                        help='A boolean describing whether the protocol should be \'sliced\' before given to the '
                             'available CPU cores (in order to have a more \'balanced\' distribution of them over the '
                             'available CPUs.')
    parser.add_argument('-p_g', required=False, type=float, default=None,
                        help='A float describing what the (two-qubit) gate and measurement error probability for the '
                             'protocols in the search should be.')
    parser.add_argument('-alpha', required=False, type=float, default=None,
                        help='A float describing what the bright-state population for the single-click protocol '
                             'should be.')
    parser.add_argument('-n_DD', required=False, type=str, default='None',
                        help='An integer describing how many entanglement generation attempts fit in half of a '
                             'dynamical decoupling sequence, or a string "auto" indicating that the program should'
                             'automatically find the best value for n_DD based on the other Bell state parameters'
                             'given.')
    parser.add_argument('-compare_metric', required=False, type=str, default='None',
                        help='String describing what metric we want to use to determine if one protocol is better than '
                             'another protocol: None will use the default setting, "ghz_fid" will use the fidelity of '
                             'the GHZ state, "stab_fid" will use the stabilizer fidelity, and "weighted_sum" will use '
                             'the weighted sum of the superoperator. Using a metric "weighted_sum_max" will only use '
                             '"weighted_sum" on (n_max,k_max), and "stab_fid" for protocols before that; this setting '
                             'will also overwrite a pre-calculated (n_max,k_max) buffer that was included with the '
                             'input_data_file. Further, the metric "logical_success" will compare protocols by looking '
                             'at the logical success rate over 5000 of their superoperator for a distance 4 toric '
                             'code. This will, however, only work for n=4: for all other values of n, this setting '
                             'willdefault back to using "stab_fid" as compare metric. If the "logical_success" setting '
                             'is active, "-init_seed_search" is set to False.')

    parser.add_argument('-argument_file',
                        help="Loads values from a file instead of the command line.",
                        type=open,
                        action=LoadFromFile)

    args_vars = vars(parser.parse_args())
    if args_vars['seeds'] and len(args_vars['seeds']) == 1:
        args_vars['seeds'] = args_vars['seeds'][0]
    if args_vars['number_of_threads'] != 'auto':
        args_vars['number_of_threads'] = int(args_vars['number_of_threads'])
    if args_vars['n_DD'] not in ['None', 'none', 'auto']:
        args_vars['n_DD'] = float(args_vars['n_DD'])
    calc_type = args_vars.pop('c_type')
    for arg in ['calculate_prot', 'init_seed_search', 'reshuffle_protocols']:
        args_vars[arg] = True if args_vars[arg].lower() in ['true', 'yes', '1', 'y', 't'] else False

    return calc_type, args_vars


def find_protocol(protocol, protocol_list=None, additional_protocols=None):
    if protocol_list is None:
        main_dir = "protocols/ProtocolRecipe"
        protocol_list = []
        for filename in os.listdir(main_dir):
            if "sim" in filename:
                protocol_list.append(filename)
    if additional_protocols is None:
        additional_protocols = []

    found = {'identical': [], 'similar': []}
    for type in found.keys():
        for filename_saved in protocol_list:
            protocol_saved = pickle.load(open("protocols/ProtocolRecipe/" + filename_saved, "rb")).protocol
            if ep.check_if_protocols_are_identical(protocol, protocol_saved, skip_similar_prots=type):
                found[type].append(filename_saved)
        for add_prot_name, add_protocol in additional_protocols:
            if ep.check_if_protocols_are_identical(protocol, add_protocol, skip_similar_prots=type):
                found[type].append(add_prot_name)

    return found



def calculate_number_of_protocols(n_max=4, k_max=22):
    nmbr_of_prots = 0
    cnt_stabs1 = 0
    cnt_stabs2 = 0
    cnt_fuse = 0
    for n in range(2, n_max + 1):
        for k in range(n - 1, k_max + 1):
            # Measure full-weight stabilizers
            nmbr_of_prots += max(k - 2 * n + 3, 0) * 2 ** (n - 1)
            cnt_stabs1 += max(k - 2 * n + 3, 0) * 2 ** (n - 1)
            # Measure weight-k' stabilizers
            for k_prime in range(2, n + 1, 2):
                nmbr_of_prots += binom(n, k_prime) * max(k - k_prime - n + 3, 0)
                cnt_stabs2 += binom(n, k_prime) * max(k - k_prime - n + 3, 0)
            # Fuse two smaller states
            nmbr_of_prots += (1 / 6) * n * (n ** 2 + 3 * n - 10) * (k - n + 2)
            cnt_fuse += (1 / 6) * n * (n ** 2 + 3 * n - 10) * (k - n + 2)
    print("")
    print(nmbr_of_prots)
    print(cnt_stabs1)
    print(cnt_stabs2)
    print(cnt_fuse)
    print("")

    nmbr_of_prots2 = 0
    cnt_stabs1_2 = 0
    cnt_stabs2_2 = 0
    cnt_fuse_2 = 0
    for n in range(2, n_max + 1):
        for k in range(n - 1, k_max + 1):
            for dec in range(1, 2 ** n):
                if dec < 2 ** (n - 1):  # Z_purification
                    n_prime = np.size(op.transform_list_ind(dec, n, 'Z'))
                else:  # XY_purification
                    n_prime = n
                for k_prime in range(n_prime - 1, k - n + 2):
                    nmbr_of_prots2 += 1
                    if dec < 2 ** (n - 1):
                        cnt_stabs2_2 += 1
                    else:
                        cnt_stabs1_2 += 1
            for n2 in range(2, n):
                n1 = n - n2 + 1
                for k2 in range(n2 - 1, k - n + n2 + 1):
                    k1 = k - k2
                    for i in range(n1):
                        for j in range(n2):
                            nmbr_of_prots2 += 1
                            cnt_fuse_2 += 1
    print(nmbr_of_prots2)
    print(cnt_stabs1_2)
    print(cnt_stabs2_2)
    print(cnt_fuse_2)


# def confidence_interval(confidence=0.682, minus_mean=False):
#     data = pickle.load(open("ghz_fidelity_test_expedient_expedient_swap_dec_prob_nodePur_decoupling0.013_pg0.01_pm0.01_pn0.1_lde_success0.0001_fixed_lde_attempts2000.pkl", "rb"))
#     data = data['dur'][0:100]
#
#     n = len(data)
#     mean = np.mean(data) if minus_mean else 0
#     mean2 = np.mean(data)
#     data_sorted = sorted(data)
#
#     lower_bound = math.floor(n * ((1 - confidence) / 2))
#     upper_bound = math.floor(n * (1 - (1 - confidence) / 2))
#     upper_bound = upper_bound if upper_bound < len(data) else -1
#
#     return abs(data_sorted[lower_bound] - mean), data_sorted[upper_bound] - mean, mean2


def plot_decoherence_channels():
    x = [*range(500)]
    T = 100
    n1 = 1
    n2 = 2
    line_1 = [1 - math.exp(-1 * (i / T) ** n1) for i in x]
    line_2 = [1 - math.exp(-1 * (i / T) ** n2) for i in x]
    fig, ax = plt.subplots()
    ax.plot(x, line_1, alpha=0.8, label="n = 1")
    ax.plot(x, line_2, alpha=0.8, label="n = 2")
    plt.xlabel('Time')
    plt.ylabel('Decoherence (gamma)')
    # plt.xlim([0.25, 0.8])
    # plt.ylim([0.65, 0.87])
    ax.legend()
    # save(g,'scatterplot.pdf',axes=True, axes_labels=['p_succ', 'F'],axes_labels_size=1.6, legend_font_size=16)
    plt.show()


if __name__ == "__main__":
    calc_type, args_vars = parse_arguments()

    if calc_type == 0:
        print("Running instance 0 of main.")
        # n = 4
        # k = 14
        # seeds = [*range(0, 3000)]
        # Tvalues = [1e-5]
        # input_fid = 0.9
        # thr_fid = 0.983
        # nstate = 200
        # # symm_prots = das.find_symmetric_protocols(n, k, seeds, Tvalues, input_fid, thr_fid, nstate, 0)
        # # pickle.dump(symm_prots, open("protocols/symm_prots_" + str(n) + "_" + str(k), "wb"))
        #
        # symm_prots_loaded = pickle.load(open("protocols/symm_prots_" + str(n) + "_" + str(k), "rb"))
        # print(len(symm_prots_loaded))

        # dyn_prot = pickle.load(open("protocols/dyn_prot_4_14_1", "rb"))
        # _, _, swap_count, node_list = ep.create_recipe(dyn_prot, True)

        # i = 8
        # prot_name = "prot_save_3_" + str(i)
        # protocol = pickle.load(open("protocols/BinaryTree/old/" + prot_name, "rb"))
        # protocol_recipe = pr.ProtocolRecipe(protocol)
        # char = ragp.simulate_protocol_recipe(protocol_recipe)
        # print(char)

        t = time.time()
        n_max = args_vars['n_max']
        k_max = args_vars['k_max']
        nstate = args_vars['nstate']
        nmb_id_elems = args_vars['number_id_elems']
        input_state = args_vars['input_state']
        da_type = args_vars['da_type']
        rot_type = args_vars['rot_type']
        calculate_prot = args_vars['calculate_prot']
        number_of_threads = args_vars['number_of_threads']
        seeds = args_vars['seeds']
        init_seed_search = args_vars['init_seed_search']
        skip_similar_prots = args_vars['skip_similar_protocols']
        reshuffle_protocols = args_vars['reshuffle_protocols']
        input_data_file = args_vars['input_data_file']
        p_g = args_vars['p_g']
        alpha = args_vars['alpha']
        n_DD = args_vars['n_DD']
        metric = args_vars['compare_metric'] if args_vars['compare_metric'] != 'None' else None
        max_calc_time = args_vars['max_calc_time']
        if max_calc_time is not None:
            max_calc_time_elements = max_calc_time.split(":")
            if len(max_calc_time_elements) == 3:
                numb_hours = max_calc_time_elements[0]
                numb_days = 0
                if len(numb_hours.split("-")) == 2:
                    numb_days = int(numb_hours.split("-")[0])
                    numb_hours = int(numb_hours.split("-")[1])
                numb_hours = int(numb_hours)
                if numb_hours > 23:
                    numb_days += int(int(numb_hours) / 24)
                    numb_hours -= int(int(numb_hours) / 24) * 24
                max_calc_time = f"{numb_days + 1}-{numb_hours}:{max_calc_time_elements[1]}:{max_calc_time_elements[2]}"
            elif len(max_calc_time_elements) == 1:
                max_calc_time = f"1-00:00:{max_calc_time_elements[0]}"
            elif len(max_calc_time_elements) == 2:
                max_calc_time = f"1-00:{max_calc_time_elements[0]}:{max_calc_time_elements[1]}"
            elif len(max_calc_time_elements) == 3:
                max_calc_time = f"1-{max_calc_time_elements[0]}:{max_calc_time_elements[1]}:{max_calc_time_elements[2]}"
            else:
                raise ValueError("max_calc_time is not understood.")

            print(f"Time limit is set at: {max_calc_time}, where the day number should be subtracted by 1.")

            date_time = datetime.strptime(max_calc_time, "%d-%H:%M:%S")
            a_timedelta = date_time - datetime(1900, 1, 1)
            max_calc_time_in_seconds = a_timedelta.total_seconds()

            print(f"This is converted to {max_calc_time_in_seconds} seconds.")

        if n_DD == 'auto' and isinstance(input_state, str):
            bell_pars = ragp.simulate_protocol_recipe(None, set_name=input_state, return_bell_par=True, alpha=alpha)
            if 'p_link' in bell_pars and bell_pars['p_link'] is not None:
                p_link = bell_pars['p_link']
                F_link = bell_pars['F_link']
            else:
                p_link, F_link = cape.calculate_bell_state_properties(bell_pars)
            n_DD = cape.number_of_entanglement_attempts_per_echo_pulse(p_link=p_link,
                                                                       t_link=bell_pars['t_link'],
                                                                       t_pulse=bell_pars['t_pulse'])
            print(f"The parameter 'n_DD' is automatically set at {n_DD}, for p_link = {p_link}, "
                  f"t_link = {bell_pars['t_link']} and t_pulse = {bell_pars['t_pulse']}, with F_link = {F_link}")

        failed_protocol_prefix = args_vars['failed_protocol_prefix']
        if input_data_file is not None:
            n_start, k_start = args_vars['n_start'], args_vars['k_start']
        else:
            n_start, k_start = 2, 1

        failed_protocol_prefix = "" if failed_protocol_prefix is None \
            else (failed_protocol_prefix if failed_protocol_prefix[-1] != "/" else failed_protocol_prefix[:-1])
        main_folder = f"protocols/{failed_protocol_prefix}/"
        if not os.path.exists(main_folder):
            os.mkdir(main_folder)

        # characteristics = ragp.simulate_protocol_recipe(None, numb_of_iter=10,
        #                                                 protocol_name='dyn_prot_4_4_1_swap', set=0)
        # print(np.average(characteristics))

        # # Close session
        # def handler(signum, frame):
        #     raise Exception("Action took too much time")
        #
        # if platform.system() == "Linux" and max_calc_time is not None:
        #     signal.signal(signal.SIGALRM, handler)
        #     signal.alarm(int(max_calc_time_in_seconds) - 1 * 60)  # Set the parameter to the amount of seconds you want to wait

        # try:
        print(f"Calculating dynamic algorithm with the following parameters: \n{args_vars}.")
        data, number_of_prots = das.dynamic_algorithm(n_max, k_max, input_state, input_data_file, n_start, k_start,
                                                      da_type=da_type, nstate=nstate, rot_type=rot_type,
                                                      show_or_not=1,
                                                      calculate_prot=calculate_prot, nmb_id_elems=nmb_id_elems,
                                                      number_of_threads=number_of_threads, seeds=seeds,
                                                      failed_protocol_prefix=failed_protocol_prefix,
                                                      init_seed_search=init_seed_search,
                                                      skip_similar_prots=skip_similar_prots,
                                                      reshuffle_protocols=reshuffle_protocols,
                                                      p_g=p_g, alpha=alpha, n_DD=n_DD, metric=metric)
        print("")
        for k in range(n_max - 1, k_max + 1):
            print(f"Maximal metric for (n, k) = ({n_max}, {k}) is {data[n_max][k][0].state}.")
        print(f"Elapsed time is {time.time() - t}.")
        print(f"Number of protocols considered = {number_of_prots}.")
        file_name_addition = "_" + str(nmb_id_elems) if calculate_prot else ""
        pickle.dump(data, open(f"{args_vars['filename_prefix']}data_{n_max}_{k_max}_{nstate}" +
                               file_name_addition + f"_{input_state}_{seeds}_{da_type}_algv1.pkl", "wb"))
        # except:
        #     print("Program crashed and was aborted.")
        #     file_name_addition = "_" + str(nmb_id_elems) if calculate_prot else ""
        #     try:
        #         pickle.dump(data, open(f"{args_vars['filename_prefix']}data_{n_max}_{k_max}_{nstate}" +
        #                                file_name_addition + f"_{input_state}_{seeds}_{da_type}_algv1_CRASHED.pkl", "wb"))
        #         print(f"Data structure dump was succesful.")
        #     except:
        #         print(f"Data structure dump failed.")

        # signal.alarm(10)  # Resets the alarm to 10 new seconds
        # signal.alarm(0)  # Disables the alarm



        # print(ghztime.minimum_GHZ_creation_time())

        # numb_of_iter = 100
        # data = pickle.load(open("data_supercomputer.pkl", "rb"))
        # print(data[4][4][0].state)
        # protocol = dap.identify_protocol(data, 4, 4, 0)
        # protocol_recipe = pr.ProtocolRecipe(protocol)
        # char1 = ragp.simulate_protocol_recipe(protocol_recipe, set=0, numb_of_iter=numb_of_iter)
        # print(np.average(char1))
        # # char2 = ragp.simulate_protocol_recipe(protocol_recipe, set=1, numb_of_iter=numb_of_iter)
        # # print(np.average(char2))
        # protocol2 = pickle.load(open("protocols/BinaryTree/old/prot_save_4_4", "rb"))
        # protocol_recipe2 = pr.ProtocolRecipe(protocol2)
        # char3 = ragp.simulate_protocol_recipe(protocol_recipe2, set=0, numb_of_iter=numb_of_iter)
        # print(np.average(char3))
        # # char4 = ragp.simulate_protocol_recipe(protocol_recipe2, set=1, numb_of_iter=numb_of_iter)
        # # print(np.average(char4))

        # da_type = 'sp'
        # data, marker_cnt = das.dynamic_algorithm(n_max, k_max, fid, da_type=da_type, nstate=1, rot_type="none", show_or_not=1)
        # print(marker_cnt)
        # print(data[n_max][k_max][0].state[0])



        # print(data[n_max][k_max][1].state[0])
        # protocol = dap.identify_protocol(data, n_max, k_max, 0)
        # pickle.dump(protocol, open("protocols/test_prot_" + str(n_max) + "_" + str(k_max) + "_0", "wb"))
        # protocol = pickle.load(open("protocols/test_prot_" + str(n_max) + "_" + str(k_max) + "_0", "rb"))
        # _, _, swap_count, _, marker = ep.create_recipe(protocol, print_prot=True, return_marker=True)
        # print(marker)

        # data = das.dynamic_algorithm(n_max, k_max, fid, da_type=da_type, nstate=10, inc_rot=1, show_or_not=1)
        # print("")
        # print(data[n_max][16][0].state[0])
        # print(data[n_max][16][1].state[0])
        # print(data[n_max][16][2].state[0])
        # print(data[n_max][16][3].state[0])
        # print(data[n_max][16][4].state[0])
        # print(data[n_max][16][5].state[0])
        # print(data[n_max][16][6].state[0])
        # print(data[n_max][16][7].state[0])
        # print(data[n_max][16][8].state[0])
        # print(data[n_max][16][9].state[0])
        # print("")
        # print(data[n_max][k_max][0].state[0])
        # print(data[n_max][k_max][1].state[0])
        # print(data[n_max][k_max][2].state[0])
        # print(data[n_max][k_max][3].state[0])
        # print(data[n_max][k_max][4].state[0])
        # print(data[n_max][k_max][5].state[0])
        # print(data[n_max][k_max][6].state[0])
        # print(data[n_max][k_max][7].state[0])
        # print(data[n_max][k_max][8].state[0])
        # print(data[n_max][k_max][9].state[0])



        # protocol = pickle.load(open("protocols/old/dyn_prot_4_14_50", "rb"))
        # protocol = pickle.load(open("protocols/old/prot_save_4_6", "rb"))
        # protocol = pickle.load(open("protocols/old/prot_save_4_12", "rb"))
        # protocol = pickle.load(open("protocols/old/dyn_prot_4_4_1", "rb"))
        # protocol_recipe = sp.run_simulator(protocol)
        # protocol_recipe.seed = 20
        # print(ragp.simulate_protocol_recipe(protocol_recipe, protocol_name="auto_generated_swap"))
        # print(ragp.simulate_protocol_recipe(protocol_recipe, protocol_name="dyn_prot_4_4_1_swap"))
        # special_seeds = []

    elif calc_type == 1:
        print("Running instance 1 of main.")
        # fids_auto = []
        # fids_manual = []
        # for prot_k in range(4, 5):
        #     # protocol_name = "protocols/old/prot_save_4_" + str(prot_k)
        #     # protocol = pickle.load(open(protocol_name, "rb"))
        #     prot_name = "dyn_prot_4_4_1"
        #     if os.path.isfile("protocols/BinaryTree/" + prot_name):
        #         protocol = pickle.load(open("protocols/BinaryTree/" + prot_name, "rb"))
        #         # plotp.plot_protocol(protocol, 1, name="test_figure_can_be_deleted")
        #         protocol_recipe = pr.ProtocolRecipe(protocol)
        #
        #         print(os.path.dirname(os.path.abspath(__file__)))
        #
        #         prot_new_name = prot_name
        #         if not os.path.isfile("protocols/ProtocolRecipeExport/" + prot_new_name):
        #             dill.dump(protocol_recipe, open("protocols/ProtocolRecipe/" + prot_new_name, "wb"))
        #         protocol_recipe.print_protocol_recipe()
        #         # protocol_recipe_export = dill.load(open("protocols/ProtocolRecipeExport/" + prot_new_name, "rb"))
        #         # protocol_recipe_export.print_protocol_recipe()
        #
        #         ragp.simulate_protocol_recipe(protocol_recipe, print_results=True)

        for i in range(1):
            # print(f"Exploring prot_unexpected_SetYves_9_{i}")
            # prot_name = "rec_failed_SetYves_" + str(i)
            prot_name = "ProtocolRecipe/dyn_prot_simv3_sIIIc_4_7_1"
            protocol = pickle.load(open("protocols/ProtocolRecipe/dyn_prot_simv3_sIIIc_4_7_1", "rb"))
            # plotp.plot_protocol(protocol, 1)
            # protocol_recipe = pr.ProtocolRecipe(protocol)
            # print(protocol.qubit_memory)
            # protocol.print_protocol_recipe()
            # list_of_id_elems = ep.other_versions_of_binary_tree_protocol(protocol, number_of_orderings_checked=100)

            # I HAVE TO CHANGE THE CODE WHERE WE MOVE A FUSION CORRECTION TO A LATER TIME STEP: IT BASICALLY ONLY HAS TO BE
            # UPDATED WHEN IT COMES ACROSS OPERATION WITH AN "i_op" NUMBER HIGHER THAN THE LOWEST "i_op" NUMBER IN THE
            # DEPENDENCY LIST (THE LIST "condition") BEFORE THE NEW DEPENDENCIES AS A RESULT OF THE NEW OPERATION ARE ADDED.

            # protocol_recipe = pr.ProtocolRecipe(protocol, id_elem_test_low_9_seed_32)
            # protocol_recipe.print_protocol_recipe(print_intermediate_memory=True)
            # for seed in range(40, 100):
            #     characteristics = ragp.simulate_protocol_recipe(protocol_recipe, print_results=False, seed=seed)
            #     print(seed, ":", characteristics[0])]

            # for seed in range(1):
            #     print(f"Exploring seed {seed}")
            #     # for id_elem in list_of_id_elems:
            #     #     try:
            #     #         protocol_recipe = pr.ProtocolRecipe(protocol, id_elem)
            #     #         # protocol_recipe.print_protocol_recipe(print_intermediate_memory=False)
            #     #     except:
            #     #         print(f"id_elem_failed_recipe_{i}_seed_{seed} = {id_elem}")
            #
            #         # plotp.plot_protocol(protocol, 1)
            #         # characteristics = ragp.simulate_protocol_recipe(protocol_recipe, seed=seed, set=1)
            #         # print(characteristics[0])
            #
            #     # try:
            #     # set_name = "SetYves"
            #     set_name = "SetIIIc"
            #     characteristics = ragp.simulate_protocol_recipe(protocol, seeds=[*range(80)], set_name=set_name,
            #                                                     metric="logical_success", p_g=0.0009, print_results=False,
            #                                                     lattices=[6, 8, 10], sc_iters=800, multithreading=True)
            #     print(characteristics)
            #         # if characteristics[0] < 0.4:
            #         #     print(f"id_elem_low_fid_{i}_seed_{seed} = {id_elem}")
                # except:
                #     print(f"id_elem_failed_sim_{i}_seed_{seed} = {id_elem}")

        prot_name = "ProtocolRecipe/dyn_prot_simv3_sIIIc_4_7_1"
        protocol = pickle.load(open("protocols/ProtocolRecipe/dyn_prot_simv3_sIIIc_4_7_1", "rb"))
        set_name = "SetIIIc"

        a = time.time()
        lattices = [4, 6, 8, 10, 12]
        so_iters = 5000
        sc_iters = 50000
        # p_g_values = [1e-12, 0.00025, 0.0005, 0.00075, 0.001, 0.00125, 0.0015, 0.00175, 0.002]
        p_g_values = [0.0008, 0.00105, 0.0013, 0.00155]

        show_fits = True
        multithreading = True
        results_dictionary = None
        # results_dictionary = {0.0008: {4: {'success': 45956.0, 'N': 50000}, 6: {'success': 47430.0, 'N': 50000}, 8: {'success': 47980.0, 'N': 50000}, 10: {'success': 48333.0, 'N': 50000}, 12: {'success': 48621.0, 'N': 50000}}, 0.00105: {4: {'success': 44837.0, 'N': 50000}, 6: {'success': 46251.0, 'N': 50000}, 8: {'success': 46666.0, 'N': 50000}, 10: {'success': 46892.0, 'N': 50000}, 12: {'success': 47033.0, 'N': 50000}}, 0.0013: {4: {'success': 43607.0, 'N': 50000}, 6: {'success': 44643.0, 'N': 50000}, 8: {'success': 44606.0, 'N': 50000}, 10: {'success': 44609.0, 'N': 50000}, 12: {'success': 44398.0, 'N': 50000}}, 0.00155: {4: {'success': 42139.0, 'N': 50000}, 6: {'success': 42547.0, 'N': 50000}, 8: {'success': 41938.0, 'N': 50000}, 10: {'success': 41107.0, 'N': 50000}, 12: {'success': 40117.0, 'N': 50000}}}
        # results_characteristics = pickle.load(open("results_characteristics.pkl", "rb"))
        # results_dictionary = {0: {4: {'success': 6005.0, 'N': 6400}, 6: {'success': 6194.0, 'N': 6400}, 8: {'success': 6270.0, 'N': 6400}, 10: {'success': 6292.0, 'N': 6400}}, 0.00025: {4: {'success': 6122.0, 'N': 6400}, 6: {'success': 6296.0, 'N': 6400}, 8: {'success': 6334.0, 'N': 6400}, 10: {'success': 6370.0, 'N': 6400}}, 0.0005: {4: {'success': 6025.0, 'N': 6400}, 6: {'success': 6221.0, 'N': 6400}, 8: {'success': 6260.0, 'N': 6400}, 10: {'success': 6337.0, 'N': 6400}}, 0.00075: {4: {'success': 5918.0, 'N': 6400}, 6: {'success': 6141.0, 'N': 6400}, 8: {'success': 6191.0, 'N': 6400}, 10: {'success': 6243.0, 'N': 6400}}, 0.001: {4: {'success': 5826.0, 'N': 6400}, 6: {'success': 5958.0, 'N': 6400}, 8: {'success': 6062.0, 'N': 6400}, 10: {'success': 6063.0, 'N': 6400}}, 0.00125: {4: {'success': 5699.0, 'N': 6400}, 6: {'success': 5808.0, 'N': 6400}, 8: {'success': 5849.0, 'N': 6400}, 10: {'success': 5820.0, 'N': 6400}}, 0.0015: {4: {'success': 5473.0, 'N': 6400}, 6: {'success': 5594.0, 'N': 6400}, 8: {'success': 5500.0, 'N': 6400}, 10: {'success': 5456.0, 'N': 6400}}, 0.00175: {4: {'success': 5290.0, 'N': 6400}, 6: {'success': 5323.0, 'N': 6400}, 8: {'success': 5158.0, 'N': 6400}, 10: {'success': 4938.0, 'N': 6400}}, 0.002: {4: {'success': 5045.0, 'N': 6400}, 6: {'success': 4998.0, 'N': 6400}, 8: {'success': 4663.0, 'N': 6400}, 10: {'success': 4210.0, 'N': 6400}}}

        # Results for (52 hours on desktop pc):
        # lattices = [4, 6, 8, 10]
        # so_iters = 1280
        # sc_iters = 12800
        # p_g_values = [1e-12, 0.00025, 0.0005, 0.00075, 0.001, 0.00125, 0.0015, 0.00175, 0.002]
        # results_dictionary = {1e-12: {4: {'success': 12415.0, 'N': 12800}, 6: {'success': 12660.0, 'N': 12800},
        #                               8: {'success': 12738.0, 'N': 12800}, 10: {'success': 12786.0, 'N': 12800}},
        #                       0.00025: {4: {'success': 12272.0, 'N': 12800}, 6: {'success': 12585.0, 'N': 12800},
        #                                 8: {'success': 12666.0, 'N': 12800}, 10: {'success': 12738.0, 'N': 12800}},
        #                       0.0005: {4: {'success': 12086.0, 'N': 12800}, 6: {'success': 12418.0, 'N': 12800},
        #                                8: {'success': 12557.0, 'N': 12800}, 10: {'success': 12638.0, 'N': 12800}},
        #                       0.00075: {4: {'success': 11820.0, 'N': 12800}, 6: {'success': 12213.0, 'N': 12800},
        #                                 8: {'success': 12398.0, 'N': 12800}, 10: {'success': 12460.0, 'N': 12800}},
        #                       0.001: {4: {'success': 11606.0, 'N': 12800}, 6: {'success': 11925.0, 'N': 12800},
        #                               8: {'success': 12059.0, 'N': 12800}, 10: {'success': 12101.0, 'N': 12800}},
        #                       0.00125: {4: {'success': 11290.0, 'N': 12800}, 6: {'success': 11534.0, 'N': 12800},
        #                                 8: {'success': 11614.0, 'N': 12800}, 10: {'success': 11605.0, 'N': 12800}},
        #                       0.0015: {4: {'success': 10894.0, 'N': 12800}, 6: {'success': 11124.0, 'N': 12800},
        #                                8: {'success': 10967.0, 'N': 12800}, 10: {'success': 10763.0, 'N': 12800}},
        #                       0.00175: {4: {'success': 10494.0, 'N': 12800}, 6: {'success': 10477.0, 'N': 12800},
        #                                 8: {'success': 10143.0, 'N': 12800}, 10: {'success': 9652.0, 'N': 12800}},
        #                       0.002: {4: {'success': 10151.0, 'N': 12800}, 6: {'success': 9837.0, 'N': 12800},
        #                               8: {'success': 9207.0, 'N': 12800}, 10: {'success': 8511.0, 'N': 12800}}}

        # Results for (half an hour on supercomputer):
        # lattices = [4, 6, 8, 10, 12]
        # so_iters = 5000
        # sc_iters = 50000
        # p_g_values = [0.0008, 0.00105, 0.0013, 0.00155]
        # results_dictionary = {0.0008: {4: {'success': 46042.0, 'N': 50000}, 6: {'success': 47360.0, 'N': 50000},
        #                                8: {'success': 47831.0, 'N': 50000}, 10: {'success': 48251.0, 'N': 50000},
        #                                12: {'success': 48427.0, 'N': 50000}},
        #                       0.00105: {4: {'success': 45079.0, 'N': 50000}, 6: {'success': 46049.0, 'N': 50000},
        #                                 8: {'success': 46471.0, 'N': 50000}, 10: {'success': 46612.0, 'N': 50000},
        #                                 12: {'success': 46584.0, 'N': 50000}},
        #                       0.0013: {4: {'success': 43944.0, 'N': 50000}, 6: {'success': 44493.0, 'N': 50000},
        #                                8: {'success': 44276.0, 'N': 50000}, 10: {'success': 43998.0, 'N': 50000},
        #                                12: {'success': 43718.0, 'N': 50000}},
        #                       0.00155: {4: {'success': 42634.0, 'N': 50000}, 6: {'success': 42553.0, 'N': 50000},
        #                                 8: {'success': 41438.0, 'N': 50000}, 10: {'success': 40331.0, 'N': 50000},
        #                                 12: {'success': 39000.0, 'N': 50000}}}

        par, perr, results_dictionary = ragp.initial_threshold_search(prot_name, protocol, set_name,
                                                                      p_g_values=p_g_values, lattices=lattices,
                                                                      so_iters=so_iters, sc_iters=sc_iters,
                                                                      results_dictionary=results_dictionary,
                                                                      multithreading=multithreading,
                                                                      show_fits=show_fits, zoomed_in_fit=False,
                                                                      zoom_number=3)

        print(f"Found threshold value: {par[0], (par[0] - perr, par[0] + perr)}.")
        print(f"results_dictionary = {results_dictionary}")
        print(f"Calculation time: {(time.time() - a)/60} min.")






        #     for seed in range(20, 24):
        #         if seed % 5 == 0:
        #             print((prot_k, seed))
        #         protocol_recipe.seed = seed
        #         # fids_auto.append(ragp.simulate_protocol_recipe(protocol_recipe))
        #         fids_manual.append(ragp.simulate_protocol_recipe(protocol_recipe, protocol_name="dyn_prot_4_4_1_swap"))
        #         # if characteristics[0] < 0.6:
        #         #     special_seeds.append((prot_k, seed))
        # # print(special_seeds)
        # # print(f"Auto-generated protocol fidelities: {fids_auto}.")
        # print(f"Manually created protocol fidelities: {fids_manual}.")
        # # print(f"The average for the auto-generated protocol is {np.average(fids_auto)}.")
        # print(f"The average for the manual protocol is {np.average(fids_manual)}.")


        # # GHZ protocols:

        # calculate_number_of_protocols()
        # conf_int = confidence_interval()
        # print(conf_int)

        # plot_decoherence_channels()

    elif calc_type == 2:
        print("Running instance 2 of main.")
        folder = "C:\\Users\\sebastiandebon\\surfdrive\\Code\\Sebastian\\oopsc\\superoperator\\SetIIIc_statistics\\"
        filename1 = "40_20211004_dyn_prot_4_4_1_swap_dec_prob_nodePur_decoupling0.013_pg0.000175_pm0.000175_pn0.015_lde_success0.017_fixed_lde_attempts100_cut_off_time0.3023.pkl"
        filename2 = "41_20211004_dyn_prot_4_4_1_swap_dec_prob_nodePur_decoupling0.013_pg0.000175_pm0.000175_pn0.015_lde_success0.017_fixed_lde_attempts100_cut_off_time0.3023.pkl"
        filename3 = "40_20211004_dyn_prot_4_4_1_swap_dec_prob_nodePur_decoupling0.013_pg0.000175889_pm0.000175889_pn0.015_lde_success0.017_fixed_lde_attempts100_cut_off_time0.3023.pkl"
        filename4 = "41_20211004_dyn_prot_4_4_1_swap_dec_prob_nodePur_decoupling0.013_pg0.000175889_pm0.000175889_pn0.015_lde_success0.017_fixed_lde_attempts100_cut_off_time0.3023.pkl"
        filename5 = "62_20211020_dyn_prot_4_4_1_swap_dec_prob_nodePur_decoupling0.013_pg0.000175_pm0.000175_pn0.015_lde_success0.017_fixed_lde_attempts100_cut_off_time0.3023.pkl"
        filename6 = "62_20211020_dyn_prot_4_4_1_swap_dec_prob_nodePur_decoupling0.013_pg0.000175889_pm0.000175889_pn0.015_lde_success0.017_fixed_lde_attempts100_cut_off_time0.3023.pkl"
        filename7 = "63_20211020_dyn_prot_4_4_1_swap_dec_prob_nodePur_decoupling0.013_pg0.000175_pm0.000175_pn0.015_lde_success0.017_fixed_lde_attempts100_cut_off_time0.3023.pkl"
        filename8 = "63_20211020_dyn_prot_4_4_1_swap_dec_prob_nodePur_decoupling0.013_pg0.000175001_pm0.000175001_pn0.015_lde_success0.017_fixed_lde_attempts100_cut_off_time0.3023.pkl"
        filename9 = "dyn_prot_simv2_sIIIc_4_10_1_node-s3c_p_g-0.0006_p_m-0.0006"
        data = pickle.load(open(folder + filename9, "rb"))
        print(np.average(data['stab_fid']))
        low_stab_fids = [(i, d) for i, d in enumerate(data['stab_fid']) if d > 0.9]
        print(low_stab_fids)


    elif calc_type == 3:
        print("Running instance 3 of main.")
        csec.run_circuit_simulator_threaded()

    elif calc_type == 4 or calc_type == 104 or calc_type == 204 or calc_type == 304 or calc_type == 404 \
            or calc_type == 504 or calc_type == 604:
        print(f"Running instance {calc_type} of main.")
        # Export and test out protocols found in the dynamic algorithm
        test_small_nr_iters = False

        # prot_new_name = "dyn_prot_simv1_sIIa_4_6_1"
        n = 4
        # k = 5
        # p_g = 0.0031
        # superop_iters = 1
        settings_version = 1 if calc_type == 4 else \
            (2 if calc_type == 104 else
             (3 if calc_type == 204 else
              (4 if calc_type == 304 else
               (5 if calc_type == 404 else
                (6 if calc_type == 504 else 7)))))

        if platform.system() == "Linux":
            multithreading = True
        else:
            multithreading = False
        show_fits = False
        show_fits_high_threshold = False
        threshold_show_fits_high_threshold = 0.00068
        max_export_prots_per_class = 2
        threshold_export_threshold = 0.00009
        export_protocols = False
        skip_threshold_calculations = False
        exclude_negative_intervals = False
        interactive_plots = False

        results_dictionary_full = {}

        if settings_version == 1:
            # Scenario 4
            multiplier_iters = 6
            max_nr_threads = 192
            set_number = "NeutralAtom1EqualCoh"
            version_number = "1"
            k_max = 15
            k_max_eval = 8
            number_of_prots_per_k = 10
            nr_seeds_used = 50
            date = "20240508"
            data = pickle.load(open(f"results_0/results/{date}_data_4_{k_max}_{number_of_prots_per_k}_1_{set_number}_{nr_seeds_used}_mpuP_algv1.pkl","rb"))
            p_g_values = [0.0005, 0.001, 0.0015, 0.002, 0.0025, 0.003, 0.0035]
            alpha = None
            n_DD = 0
            # p_g_values = [1e-12, 0.00025, 0.0005, 0.00075, 0.001, 0.00125, 0.0015, 0.00175, 0.002]
            k_range = range(8, k_max_eval + 1)
        elif settings_version == 2:
            # 104, CAN BE REMOVED:
            multiplier_iters = 1
            max_nr_threads = 1
            set_number = "XXXtest"
            version_number = "1"
            k_max = 15
            k_max_eval = 9
            number_of_prots_per_k = 10
            nr_seeds_used = 50
            date = "20230106"
            # data = pickle.load(open(f"calc_data/Set{set_number}_v{version_number}/{date}_data_4_{k_max}_{number_of_prots_per_k}_1_Set{set_number}_{nr_seeds_used}_mpuP_algv1.pkl","rb"))
            data = pickle.load(open("calc_data/crash_file/20231002_data_4_15_10_1_NeutralAtom1EqualCoh_50_mpuP_algv1.pkl", "rb"))
            p_g_values = [0.0001, 0.00025, 0.00075, 0.0010, 0.00125]
            alpha = None
            n_DD = 0
            k_range = range(9, k_max_eval + 1)
            test_small_nr_iters = True
        elif settings_version == 3:
            # 204, CAN BE REMOVED:
            multiplier_iters = 4
            max_nr_threads = 128
            set_number = "IIIr"
            version_number = "1"
            k_max = 8
            k_max_eval = 8
            number_of_prots_per_k = 30
            nr_seeds_used = 80
            date = "20230106"
            data = pickle.load(open(f"calc_data/Set{set_number}_v{version_number}/{date}_data_4_{k_max}_{number_of_prots_per_k}_1_Set{set_number}_{nr_seeds_used}_mpuP_algv1.pkl","rb"))
            p_g_values = [0.00005, 0.0001, 0.0003, 0.0005, 0.0007]
            alpha = None
            n_DD = 18
            k_range = range(5, k_max_eval + 1)
        elif settings_version == 4:
            # 304, CAN BE REMOVED:
            multiplier_iters = 4
            max_nr_threads = 128
            set_number = "Vr"
            version_number = "1"
            k_max = 15
            k_max_eval = 15
            number_of_prots_per_k = 10
            nr_seeds_used = 40
            date = "20230106"
            data = pickle.load(open(f"calc_data/Set{set_number}_v{version_number}/{date}_data_4_{k_max}_{number_of_prots_per_k}_1_Set{set_number}_{nr_seeds_used}_mpuP_algv1.pkl","rb"))
            p_g_values = [0.00005, 0.0002, 0.0005, 0.0007, 0.0009]
            alpha = None
            n_DD = 18
            k_range = range(8, k_max_eval + 1)
        elif settings_version == 5:
            # 404, CAN BE REMOVED:
            multiplier_iters = 4
            max_nr_threads = 128
            set_number = "Vq"
            version_number = "1"
            k_max = 15
            k_max_eval = 15
            number_of_prots_per_k = 10
            nr_seeds_used = 40
            date = "20230106"
            data = pickle.load(open(f"calc_data/Set{set_number}_v{version_number}/{date}_data_4_{k_max}_{number_of_prots_per_k}_1_Set{set_number}_{nr_seeds_used}_mpuP_algv1.pkl","rb"))
            p_g_values = [0.00005, 0.0002, 0.0005, 0.0007, 0.0009]
            alpha = None
            n_DD = 18
            k_range = range(7, k_max_eval + 1)
        elif settings_version == 6:
            # 504, CAN BE REMOVED:
            multiplier_iters = 4
            max_nr_threads = 128
            set_number = "Vp"
            version_number = "2"
            k_max = 15
            k_max_eval = 15
            number_of_prots_per_k = 10
            nr_seeds_used = 40
            date = "20230106"
            data = pickle.load(open(f"calc_data/Set{set_number}_v{version_number}/{date}_data_4_{k_max}_{number_of_prots_per_k}_1_Set{set_number}_{nr_seeds_used}_mpuP_algv1.pkl", "rb"))
            p_g_values = [0.0001, 0.0003, 0.0005, 0.0007, 0.0009]
            alpha = None
            n_DD = 18
            k_range = range(7, k_max_eval + 1)
        else:
            # 604, CAN BE REMOVED:
            multiplier_iters = 4
            max_nr_threads = 128
            set_number = "Ve"
            version_number = "2"
            k_max = 10
            k_max_eval = 10
            number_of_prots_per_k = 25
            nr_seeds_used = 50
            date = "20230106"
            data = pickle.load(open(f"calc_data/Set{set_number}_v{version_number}/{date}_data_4_{k_max}_{number_of_prots_per_k}_1_Set{set_number}_{nr_seeds_used}_mpuP_algv1.pkl", "rb"))
            p_g_values = [0.0012, 0.0015, 0.0018, 0.0021, 0.0024]
            alpha = None
            n_DD = 18
            k_range = range(4, k_max_eval + 1)

        lattices = [4, 6, 8, 10]
        if platform.system() == "Linux":
            so_iters = 320 * multiplier_iters
            sc_iters = 3200 * multiplier_iters
        elif test_small_nr_iters:
            so_iters = 4
            sc_iters = 100
        else:
            so_iters = 80
            sc_iters = 1600

        # print(len(results_dictionary_full.keys()))
        # exit(0)

        thresholds_found = {}

        new_protocols = []
        protocols_current = defaultdict(lambda: defaultdict(list))
        start_time = time.time()
        for k in k_range:
            prot_new_name_prefix = f"dyn_prot_simv{version_number}_s{set_number}_{n}_{k}_"
            # protocols_new = []
            # prot_number = number_of_prots_per_k
            try:
                len_data_n_k = len(data[n][k])
            except IndexError:
                continue
            for i in range(min(len_data_n_k, 100)):
                print(f"\nk = {k}, b = {i + 1}, stab_fid = {data[n][k][i].state}.")
                prot_name = prot_new_name_prefix + str(i + 1)
                protocol = dap.identify_protocol(data, n, k, i)
                protocol = dap.protocol_add_meta_data(protocol)
                sim_prots = find_protocol(protocol, additional_protocols=new_protocols)
                protocol_not_yet_used = True
                if len(sim_prots['identical']) == 0:
                    print(f"Protocol {prot_name} ({data[n][k][i].state}) is not yet in the list. It has similar protocols: {sim_prots['similar']}.")
                else:
                    print(f"Protocol {prot_name} ({data[n][k][i].state}) is already {len(sim_prots['identical'])} times in the list: {sim_prots['identical']}.")
                    print(f"It has similar protocols: {sim_prots['similar']}.")
                    for prots_sim in sim_prots['identical']:
                        if set_number in prots_sim:
                            protocol_not_yet_used = False
                            break
                new_protocols.append((prot_name, protocol))

                if skip_threshold_calculations:
                    thresholds_found[prot_name] = {}
                    thresholds_found[prot_name]["threshold"] = None
                    protocol_recipe = pr.ProtocolRecipe(protocol)
                    thresholds_found[prot_name]["protocol_recipe"] = protocol_recipe
                    thresholds_found[prot_name]["similar_prots"] = sim_prots

                elif protocol_not_yet_used:
                    thresholds_found[prot_name] = {}
                    results_dictionary = results_dictionary_full[prot_name] if \
                        results_dictionary_full is not None and prot_name in results_dictionary_full.keys() else None
                    # threshold_tries = [0.0028, 0.0030, 0.0032, 0.0034]
                    # pg_below = []
                    # pg_above = []

                    # pg_prev = 0
                    # pg_new = p_g
                    # pg_tried = [pg_new]
                    # for i_threshold in range(6):
                    #     result = ragp.simulate_protocol_recipe(pr.ProtocolRecipe(protocol), superop_iters, print_results=False,
                    #                                            seeds=None, set_name=f"Set{set_number}", p_g=pg_new,
                    #                                            metric="logical_success", lattices=[4, 10], sc_iters=500)
                    #     pg_previous = pg_new
                    #     if result[0] < result[1] and pg_prev > pg_new:
                    #         # Below threshold and previous p_g value was higher
                    #         pg_new = pg_new + 0.5*(pg_prev - pg_new)
                    #         pg_below.append(pg_previous)
                    #     elif result[0] < result[1] and pg_prev < pg_new:
                    #         pg_new = pg_new + (pg_new - pg_prev)
                    #         pg_below.append(pg_previous)
                    #     elif result[0] > result[1] and pg_prev > pg_new:
                    #         # Above threshold and previous p_g value was higher
                    #         if pg_new - (pg_prev - pg_new) not in pg_tried:
                    #             pg_new = pg_new - (pg_prev - pg_new)
                    #         else:
                    #             pg_new = pg_new - 0.5*(pg_prev - pg_new)
                    #         pg_above.append(pg_previous)
                    #     elif result[0] > result[1] and pg_prev < pg_new:
                    #         pg_new = pg_new - 0.5*(pg_new - pg_prev)
                    #         pg_above.append(pg_previous)
                    #     elif result[0] == result[1]:
                    #         # Threshold found
                    #         pg_below.append(pg_previous)
                    #         break
                    #     pg_tried.append(pg_new)
                    #     pg_prev = pg_previous
                    #     # print(f"Succes at p_g = {pg_prev}: {result}. New p_g = {pg_new}.")
                    # max_pg_below = max(pg_below) if len(pg_below) > 0 else 0
                    # min_pg_above = min(pg_above) if len(pg_above) > 0 else 0
                    # print(f"Highest threshold found {max_pg_below}. Lowest p_g value above threshold {min_pg_above}. All tries: pg_below = {pg_below}, pg_above = {pg_above}.")

                    # for p_g_use in threshold_tries:
                    #     result = ragp.simulate_protocol_recipe(pr.ProtocolRecipe(protocol), superop_iters, print_results=False,
                    #                                            seeds=None, set_name=f"Set{set_number}", p_g=p_g_use,
                    #                                            metric="logical_success", lattices=[6, 12], sc_iters=1000)
                    #     if result[0] <= 0.5:
                    #         pg_above.append((p_g_use, result[0]))
                    #     else:
                    #         pg_below.append((p_g_use, result[0]))
                    # print(f" Above threshold: {pg_above}. Below threshold: {pg_below}.")

                    # if (k, i + 1) in [(6, 2), (6, 3), (7, 5), (10, 5), (6, 5), (5, 2), (9, 2), (6, 9)]:
                    #     show_fits = True
                    # else:
                    #     show_fits = False

                    set_name = set_number if (set_number in ["SolidState1EqualCoh", "SolidState2EqualCoh",
                                                            "NeutralAtom1EqualCoh", "NeutralAtom2EqualCoh"]
                                              or set_number[:3] == "Set") \
                        else "Set" + set_number
                    run_loop = True
                    show_fits_inside = show_fits
                    while run_loop is True:
                        par, perr, results_dictionary = ragp.initial_threshold_search(prot_name,
                                                                                      pr.ProtocolRecipe(protocol),
                                                                                      set_name,
                                                                                      p_g_values=p_g_values,
                                                                                      lattices=lattices,
                                                                                      so_iters=so_iters, sc_iters=sc_iters,
                                                                                      results_dictionary=results_dictionary,
                                                                                      multithreading=multithreading,
                                                                                      max_nr_threads=max_nr_threads,
                                                                                      show_fits=show_fits_inside,
                                                                                      zoomed_in_fit=False,
                                                                                      alpha=alpha,
                                                                                      n_DD=n_DD,
                                                                                      zoom_number=2,
                                                                                      interactive_plot=interactive_plots)

                        if isinstance(results_dictionary_full, dict):
                            results_dictionary_full[prot_name] = results_dictionary
                        print(f"Found threshold value: {par[0], (par[0] - perr, par[0] + perr)}.")
                        thresholds_found[prot_name]["threshold"] = par[0], (par[0] - perr, par[0] + perr)
                        if show_fits_inside == show_fits:
                            protocol_recipe = pr.ProtocolRecipe(protocol)
                            thresholds_found[prot_name]["protocol_recipe"] = protocol_recipe
                            thresholds_found[prot_name]["similar_prots"] = sim_prots
                        if show_fits_inside is False and show_fits_high_threshold and par[0] > threshold_show_fits_high_threshold \
                                and (exclude_negative_intervals is False
                                     or (exclude_negative_intervals and par[0] - perr > 0)):
                            show_fits_inside = True
                        else:
                            run_loop = False

                # protocols_new.append(protocol)
                # protocol_recipe = pr.ProtocolRecipe(protocol)
                #     protocol_recipe.print_protocol_recipe()
                # if not os.path.isfile("protocols/ProtocolRecipeExport/" + prot_name):
                #     plotp.plot_protocol(protocol, 1, name="protocols_pdf/" + prot_name)
                # # protocol_recipe.print_protocol_recipe()
                # if check is False:
                # plotp.plot_protocol(protocol, 1, name="protocols_pdf/" + prot_name)
                # dill.dump(protocol_recipe, open("protocols/ProtocolRecipe/" + prot_name, "wb"))
                # prot_number += 1
                # if check is False:
                #     print(n, k, i + 1, data[n][k][i].state, prot_number)
                #     print(check)
                #     print("")

            # # prot_test_name_prefix = f"dyn_prot_simv1_sY_{n}_{k}_"
            # checks = []
            # for protocol in list_of_protocols:
            #     check = ep.check_if_protocols_are_identical(protocol, protocols_current)
            #     checks.append(check)
            # #     # if check is True:
            # #     prot_name = prot_test_name_prefix + str(current_protocol_number_index)
            # #     plotp.plot_protocol(protocol, 1, name="protocols_pdf/" + prot_name)
            # #     current_protocol_number_index += 1

            # checks = []
            # check_with_protocol_current = []
            # for i_prot, protocol in enumerate(protocols_new):
            #     # check = ep.check_if_protocols_are_identical(protocol, list_of_protocols[:i_prot] + list_of_protocols[(i_prot+1):])
            #     check = ep.check_if_protocols_are_identical(protocol, protocols_current)
            #     checks.append(check)
            #     # check_with_protocol_current.append(ep.check_if_protocols_are_identical(protocol_current, protocol))

            # print(checks)
            # # # print(check_with_protocol_current)
        print(f"results_dictionary_full = {results_dictionary_full}")
        print(f"Calculation time: {(time.time() - start_time)/60} minutes.")
        print("")
        if not skip_threshold_calculations:
            thresholds_found = {k: v for k, v in sorted(thresholds_found.items(), key=lambda item: item[1]["threshold"][0], reverse=True)}
            print({k: v["threshold"] for k, v in thresholds_found.items()})
        # thresholds_found = {k: v for k, v in sorted(thresholds_found.items(), key=lambda item: item[1][1][1], reverse=True)}
        # print(thresholds_found)
        print("")

        exported_protocols = []
        exported_protocols_identical_to_others = {}
        for prot_name in thresholds_found.keys():
            protocol_is_candidate_for_export = False
            if skip_threshold_calculations:
                protocol_is_candidate_for_export = True
            else:
                threshold_value = thresholds_found[prot_name]["threshold"][0]
                threshold_value_left_bound = thresholds_found[prot_name]["threshold"][1][0]
                if threshold_export_threshold is not None and threshold_value > threshold_export_threshold and \
                        (exclude_negative_intervals is False or
                         (exclude_negative_intervals and threshold_value_left_bound > 0)):
                    protocol_is_candidate_for_export = True

            if protocol_is_candidate_for_export:
                similar_protocols_to_candidate_protocol = thresholds_found[prot_name]["similar_prots"]["similar"]

                # Check if current protocol is similar to protocols that are already exported
                prot_class = None
                for i_type, export_type in enumerate(exported_protocols):
                    for prot_name_exp in export_type["protocols"]:
                        if prot_name_exp in similar_protocols_to_candidate_protocol:
                            prot_class = i_type
                            break
                    if prot_class is not None:
                        break
                    else:
                        for sim_prot_name in export_type["similar_prots"]:
                            if prot_name == sim_prot_name:
                                prot_class = i_type
                                break

                # We allow 'max_export_prots_per_class' similar protocols to be exported (not more):
                if prot_class is not None and \
                        len(exported_protocols[prot_class]["protocols"]) >= max_export_prots_per_class:
                    export_current_protocol = False
                else:
                    export_current_protocol = True

                # if not (prot_name in similar_protocols_to_exported_protocols or
                #         len(set(exported_protocols) & set(similar_protocols_to_candidate_protocol)) > 0):
                if export_current_protocol is True:
                    if prot_class is None:
                        exported_protocols.append({"protocols": [prot_name],
                                                   "similar_prots": similar_protocols_to_candidate_protocol})
                    else:
                        exported_protocols[prot_class]["protocols"].append(prot_name)
                        if prot_name in exported_protocols[prot_class]["similar_prots"]:
                            index_prot_name = exported_protocols[prot_class]["similar_prots"].index(prot_name)
                            del exported_protocols[prot_class]["similar_prots"][index_prot_name]
                        exported_protocols[prot_class]["similar_prots"] += similar_protocols_to_candidate_protocol
                        exported_protocols[prot_class]["similar_prots"] = list(set(exported_protocols[prot_class]["similar_prots"]))
                    # Export new protocol
                    if export_protocols:
                        protocol_recipe = thresholds_found[prot_name]["protocol_recipe"]
                        dill.dump(protocol_recipe, open("protocols/ProtocolRecipe/" + prot_name, "wb"))
                        print(colored(f"Protocol {prot_name} is exported", "yellow"))
                    else:
                        print(colored(f"Protocol {prot_name} will be exported", "yellow"))
                    print(f"Identical protocols: {thresholds_found[prot_name]['similar_prots']['identical']}.")
                    print(f"Similar protocols: {thresholds_found[prot_name]['similar_prots']['similar']}.")
                    if len(thresholds_found[prot_name]['similar_prots']['identical']) > 0:
                        exported_protocols_identical_to_others[prot_name] = thresholds_found[prot_name]['similar_prots']['identical']
                else:
                    if export_protocols:
                        print(colored(f"Protocol {prot_name} is not exported", "red"))
                    else:
                        print(colored(f"Protocol {prot_name} will not be exported", "red"))
                    print(f"Identical protocols: {thresholds_found[prot_name]['similar_prots']['identical']}.")
                    print(f"Similar protocols: {thresholds_found[prot_name]['similar_prots']['similar']}.")
                print("Updated exported protocols:")
                for i_type in range(len(exported_protocols)):
                    print(f"Protocols {exported_protocols[i_type]['protocols']} that have similar protocols {exported_protocols[i_type]['similar_prots']}.")

        print("")
        for prot_name in exported_protocols_identical_to_others.keys():
            print(f"Exported protocol {prot_name} has identical protocols: {exported_protocols_identical_to_others[prot_name]}.")

        print("")
        if not skip_threshold_calculations:
            for i_type in range(len(exported_protocols)):
                for prot_name in exported_protocols[i_type]["protocols"]:
                    print(f"({prot_name}, None, [{thresholds_found[prot_name]['threshold'][0]}])")



    elif calc_type == 5:
        print("Running instance 5 of main.")
        # Analyse protocol simulations that have failed.

        start_time = time.time()

        stop_loops = False
        main_dir = "protocols/040504/"
        for filename in os.listdir(main_dir):
            if filename.isdigit():
                for filename_failed in os.listdir(main_dir + filename):
                    if filename_failed.startswith("rec_failed"):
                        # prot_name = "rec_failed_SetIIIc_32"
                        # prot_name = "ProtocolRecipe/dyn_prot_simv1_sIIIc_4_7_2"
                        prot_name = filename_failed
                        protocols = []
                        for i in [0]:
                            prot_name_full = prot_name # + str(i) + "_1"
                            protocol_recipe = dill.load(open(main_dir + filename + "/" + prot_name_full, "rb"))
                            # plotp.plot_protocol(protocol_recipe.protocol, 1, name="protocols_pdf/" + prot_name)
                            # protocols.append(protocol_recipe.protocol)
                            protocol = protocol_recipe.protocol
                            id_elem = protocol_recipe.id_elem
                            protocol_recipe = pr.ProtocolRecipe(protocol, id_elem)
                            # protocol_recipe.print_protocol_recipe()
                            start_time = time.time()
                            for seed in [prot_name.split("_")[-1]]:
                                # print(f"Seed {seed} for i {i}.")
                                print(f"Seed {seed} for protocol {prot_name_full} in folder {filename}.")
                                result = ragp.simulate_protocol_recipe(protocol_recipe, print_results=False,
                                                                       seeds=[seed], set_name="SetIXf", p_g=0.0001,
                                                                       alpha=0.0667, n_DD=78)
                                print(result)
                                print(f"Calculation time: {time.time() - start_time}.")
                                if result[0] == 0:
                                    stop_loops = True
                                    break
                            if stop_loops:
                                break
                        if stop_loops:
                            break
                if stop_loops:
                    break

        # result = check_if_protocols_are_identical(protocols[0], protocols[1])
        # print(result)

        print(f"Running time: {time.time() - start_time}")

    elif calc_type == 6:
        print("Running instance 6 of main.")

        def majority_function(x):
            x = list(x)
            len_x = len(x)
            # # This part doesn't work:
            # sum = 0
            # indices_set = {*range(len_x)}
            # weight_ops = math.ceil(len_x / 2)
            # for subset in set(itertools.combinations(indices_set, 3)):
            #     subproduct = 1
            #     for index in subset:
            #         subproduct *= x[index]
            #     sum += subproduct
            number_of_ones = x.count(1)
            if number_of_ones >= len_x / 2:
                return 1
            else:
                return 0

        def inner_product_function(x):
            len_x = len(x)
            sum = 0
            half_len_x = math.floor(len_x / 2)
            for ind in range(half_len_x):
                sum += x[ind] * x[ind + half_len_x]
            return sum % 2

        def test_function(x):
            len_x = len(x)
            sum = 0
            sum_txt = ""
            for ind_1 in range(len_x - 1, -1, -1):
                prod = 1
                prod_txt = ""
                for ind_2 in range(ind_1, -1, -1):
                    prod *= x[ind_2]
                    prod_txt += "x[" + str(ind_2) + "]*"
                sum_txt += prod_txt[:-1] + " + "
                sum += prod
            # if x == [0] * len_x:
            #     print(sum_txt[:-3])
            return sum % 2

        def mult(x):
            len_x = len(x)
            prod = 1
            for ind_1 in range(len_x):
                prod *= x[ind_1]
            return prod % 2

        def mult_plus_one(x):
            len_x = len(x)
            prod = 1
            for ind_1 in range(len_x):
                prod *= x[ind_1]
            return prod % 2

        def test_function_1(x):
            # return x[0]*x[1]*x[3]*x[4] + x[0]*x[1]*x[5] + x[2]*x[3]*x[4] + x[2]*x[5]
            # return x[0]*x[1] + x[2]
            return x[0]*x[1]

        def test_function_2(x):
            return x[3]*x[2]

        def test_function_3(x):
            return x[1]*x[3]

        def test_function_4(x):
            return (test_function_1(x) + test_function_2(x)) % 2

        def fourier_transform(g1, g2, g3, g4, n):
            fourier_coefficients_1 = [None] * (2 ** n)
            fourier_coefficients_2 = [None] * (2 ** n)
            # fourier_coefficients_3 = [None] * (2 ** n)
            fourier_coefficients_4 = [None] * (2 ** n)
            for y_dec in range(2 ** n):
                y = np.array(op.dec2signs(y_dec, n))
                sum1 = 0
                sum2 = 0
                sum3 = 0
                sum4 = 0
                for x_dec in range(2 ** n):
                    x = np.array(op.dec2signs(x_dec, n))
                    sum1 += (-1) ** (g1(x) + np.inner(x, y))
                    sum2 += (-1) ** (g2(x) + np.inner(x, y))
                    # sum3 += (-1) ** (g3(x) + np.inner(x, y))
                    sum4 += (-1) ** (g4(x) + np.inner(x, y))
                fourier_coefficients_1[y_dec] = sum1 / (2 ** n)
                fourier_coefficients_2[y_dec] = sum2 / (2 ** n)
                # fourier_coefficients_3[y_dec] = sum3 / (2 ** n)
                fourier_coefficients_4[y_dec] = sum4 / (2 ** n)

            # main_coeff = 2**(-(n-1))
            # count_main_coeff = 0
            # for y, coeff in enumerate(fourier_coefficients):
            #     print(op.dec2signs(y, n), coeff)
            #     if abs(coeff) == main_coeff:
            #         count_main_coeff += 1
            # print(f"The main coefficient {main_coeff} comes up {count_main_coeff} times.")

            for y_dec in range(2 ** n):
                print(f"{op.dec2signs(y_dec, n)}, "
                      f"{fourier_coefficients_1[y_dec]}, "
                      f"{fourier_coefficients_2[y_dec]}, "
                      # f"{fourier_coefficients_3[y_dec]}, "
                      f"{fourier_coefficients_4[y_dec]}.")

            return fourier_coefficients_1

        n = 4
        # for x_dec in range(2 ** n):
        #     x = op.dec2signs(x_dec, n)
        #     g_x = test_function(x)
        #     # for i in range(len(x)):
        #     #     x[i] = 1 if x[i] == 0 else -1
        #     # f_x = f(x)
        #     print(x, g_x)

        d = fourier_transform(test_function_1, test_function_2, test_function_3, test_function_4, n)

        # fig = plt.bar([*range(2**n)], d)
        # ax = plt.gca()
        # plt.grid(axis='y', alpha=0.75)
        # plt.grid(axis='x', alpha=0.75)
        # plt.xlabel(r'$y$')
        # plt.ylabel('Value')
        # plt.title(r'Fourier coefficients of TF' + str(n) + '$(x)$')
        # plt.xticks([*range(2**n)])
        # label_distance = math.floor(2**n / 10)
        # tick_labels = np.arange(0, 2**n + 1, label_distance)
        # [l.set_visible(False) for (i, l) in enumerate(ax.xaxis.get_ticklabels()) if i not in tick_labels]
        # plt.show()

    elif calc_type == 7:
        print("Running instance 7 of main.")
        from circuit_simulation.stabilizer_measurement_protocols.run_protocols import update_result_files
        # Calculate and plot the running time for a certain protocol compared to how many GHZ states finished within
        # those times.

        # folder = "C:\\Users\\sebastiandebon\\surfdrive\\Code\\Sebastian\\oopsc\\superoperator\\Thresholds_best_prots_sVp\\"
        # prot_name = "dyn_prot_simv2_sVp_4_15_1"
        folder = "C:\\Users\\sebastiandebon\\surfdrive\\Code\\Sebastian\\oopsc\\superoperator\\Thresholds_fix_weight_3_error\\"
        # folder = "C:\\Users\\sebastiandebon\\surfdrive\\Code\\Sebastian\\oopsc\\superoperator\\Thresholds_known_prots_in_Set3g\\"
        # prot_name = "dyn_prot_simv5_sIIIf_4_11_2"    #"hc_stringent"   #"dyn_prot_simv3_sIIIc_4_7_1"    #"hc_expedient"
        prot_name = "hc_weight_3_direct"  # "hc_stringent"   #"dyn_prot_simv3_sIIIc_4_7_1"    #"hc_expedient"
        set_number = "s3e"
        error_prob = 0.002

        # # # folder = "C:\\Users\\sebastiandebon\\surfdrive\\Code\\Sebastian\\oopsc\\superoperator\\Old\\Best_prots_sIIIc_v2_SAME_STATISTICS\\Statistics_best_prots_sIIIc_v2\\"
        # folder = "C:\\Users\\sebastiandebon\\surfdrive\\Code\\Sebastian\\oopsc\\superoperator\\Thresholds_known_prots\\20221107_154133_"
        # prot_name = "hc_plain"
        # set_number = "s3g"
        # error_prob = "1e-12"

        # statistics = pickle.load(open(folder + "20220708_205041_" + prot_name + "_node-" + set_number + "_p_g-" + str(error_prob) + "_p_m-" + str(error_prob) + ".pkl", "rb"))
        try:
            if set_number in ["s3h", "s3k"]:
                filename = prot_name + "_node-" + set_number + "_p_g-" + str(error_prob) + "_p_m-" + str(error_prob) + "_no-dec_no-prob.pkl"
            else:
                filename = prot_name + "_node-" + set_number + "_bell_pair_type-30_network_noise_type-80_n_DD-8.0" + "_p_g-" + str(error_prob) + "_p_m-" + str(error_prob) + ".pkl"
            update_result_files(folder, filename.split(".pkl")[0])
            filename = folder + filename
            statistics = pickle.load(open(filename, "rb"))
        except FileNotFoundError:
            if set_number in ["s3h", "s3k"]:
                filename = prot_name + "_node-" + set_number + "_no-dec_no-prob.pkl"
            else:
                filename = prot_name + "_node-" + set_number + ".pkl"
            update_result_files(folder, filename.split(".pkl")[0])
            filename = folder + filename
            statistics = pickle.load(open(filename, "rb"))

        durations = statistics['dur']
        total_number_of_iterations = len(durations)

        count_durations = {}
        for duration in durations:
            duration = round(duration, ACCURACY)
            if duration not in count_durations:
                count_durations[duration] = 1
            else:
                count_durations[duration] += 1
        count_durations = dict(sorted(count_durations.items(), key=lambda item: item[0]))

        difference = list(count_durations.keys())[1] - list(count_durations.keys())[0] if len(count_durations.keys()) > 1 else 0
        previous_duration = list(count_durations.keys())[0]
        prob = count_durations[previous_duration]/total_number_of_iterations
        accum_prob = prob
        print("Duration; #Iterations; Scenario prob.; Comm. scenario prob.; Delta_t with previous; #Time_steps with previous")
        print(f"{previous_duration}; {count_durations[previous_duration]}; "
              f"{round(prob, ACCURACY)}; {round(accum_prob, ACCURACY)}")
        x_data = []
        y_data = []
        for duration in list(count_durations.keys())[1:]:
            diff_w_previous = duration - previous_duration
            prob = count_durations[duration]/total_number_of_iterations
            accum_prob += prob
            print(f"{duration}; {count_durations[duration]}; {round(prob, ACCURACY)}; {round(accum_prob, ACCURACY)}; "
                  f"{round(diff_w_previous, ACCURACY)}; {diff_w_previous/difference}")
            previous_duration = duration
            x_data.append(duration)
            y_data.append(accum_prob*100)

        fig = plt.plot(x_data, y_data)
        ax = plt.gca()
        # plt.grid(axis='y', alpha=0.75)
        # plt.grid(axis='x', alpha=0.75)
        plt.xlabel('Cut-off time (seconds)')
        plt.ylabel('GHZ success probability')
        plt.title(prot_name)
        # plt.xticks([*range(2**n)])
        # label_distance = math.floor(2**n / 10)
        # tick_labels = np.arange(0, 2**n + 1, label_distance)
        # [l.set_visible(False) for (i, l) in enumerate(ax.xaxis.get_ticklabels()) if i not in tick_labels]
        # plt.yscale("log")
        # locs, labels = plt.yticks()  # Get the current locations and labels.
        # plt.yticks()  # Set label locations.
        ax.set_ylim([78, 102])
        ax.set_xlim([0, 0.23])
        ax.grid(True, which='major', linewidth=1)
        ax.grid(True, which='minor')

        ax.xaxis.set_minor_locator(mpl.ticker.MultipleLocator(0.1))
        ax.yaxis.set_minor_locator(mpl.ticker.MultipleLocator(1))
        plt.show()

    elif calc_type == 8:
        print("Running instance 8 of main.")

        from line_profiler import LineProfiler
        import random

        def do_stuff(numbers):
            s = sum(numbers)
            l = [numbers[i] / 43 for i in range(len(numbers))]
            m = ['hello' + str(numbers[i]) for i in range(len(numbers))]
            return 5

        numbers = [random.randint(1, 100) for i in range(1000)]
        lp = LineProfiler()
        lp_wrapper = lp(do_stuff)
        b = lp_wrapper(numbers)
        lp.print_stats()
        print(b)

    elif calc_type == 9:
        print("Running instance 9 of main.")

        c = 10
        for i in range(10):
            if c == 10:
                try:
                    b = a[0]
                except:
                    print(i)
                    break

    elif calc_type == 10:
        print("Running instance 10 of main.")
        # Calculate and plot the error probabilities of double excitation and off-resonant excitation.

        p_DE_values = np.linspace(0.001, 0.1, 1000)
        p_WE_values = [None] * len(p_DE_values)
        delta_values = [None] * len(p_DE_values)
        for i, p_DE in enumerate(p_DE_values):
            delta_values[i] = math.sqrt(3)*math.pi/8 * (1/p_DE - 1)
            p_WE_values[i] = 13/16 * (1/p_DE - 1)**(-1) * (1 - (1/p_DE - 1)**(-1))

        fig = plt.plot(p_DE_values, p_WE_values)
        ax = plt.gca()
        plt.grid(axis='y', alpha=0.75)
        plt.grid(axis='x', alpha=0.75)
        plt.xlabel(r'$p_{DE}$')
        plt.ylabel(r'$p_{WE}$')
        plt.title(r'Relation between excitation errors')
        plt.show()

        delta_values = np.linspace(0.7, 100, 1000)
        p_DE_values = [None] * len(delta_values)
        p_WE_values = [None] * len(delta_values)
        for i, delta in enumerate(delta_values):
            alpha = math.sqrt(3)*math.pi/(8*delta)
            p_DE_values[i] = 1/(1+1/alpha)
            p_WE_values[i] = 13/16*alpha*(1-alpha)

        fig = plt.plot(delta_values, p_DE_values, label=r'$p_{DE}$')
        plt.plot(delta_values, p_WE_values, label=r'$p_{WE}$')
        ax = plt.gca()
        plt.grid(axis='y', alpha=0.75)
        plt.grid(axis='x', alpha=0.75)
        plt.xlabel(r'$\tilde{\Delta}$')
        plt.ylabel('Error probability')
        plt.title(r'Relation between $\tilde{\Delta}$ and excitation errors')
        plt.legend()
        plt.show()

    elif calc_type == 11:
        print("Running instance 11 of main.")
        # Calculate the influence of each of the decoherence channels for a certain protocol at a certain cut-off time
        # and error probability.
        from oopsc.superoperator.superoperator import Superoperator
        from circuit_simulation.stabilizer_measurement_protocols.run_protocols import update_result_files
        from circuit_simulation._superoperator.superoperator_methods import calculate_weighted_sum

        folder = "C:\\Users\\sebastiandebon\\surfdrive\\Code\\Sebastian\\oopsc\\superoperator\\Thresholds_best_prots_sIIIc\\"
        prot_name = "dyn_prot_simv3_sIIIc_4_7_1"
        set_number = "s10n"
        error_prob = 0.0001
        cut_off_time = None

        # # # folder = "C:\\Users\\sebastiandebon\\surfdrive\\Code\\Paul\\oopsc\\superoperator\\Old\\Best_prots_sIIIc_v2_SAME_STATISTICS\\Statistics_best_prots_sIIIc_v2\\"
        # folder = "C:\\Users\\sebastiandebon\\surfdrive\\Code\\Paul\\oopsc\\superoperator\\Thresholds_known_prots\\"
        # prot_name = "hc_dyn_prot_4_4_1"
        # set_number = "s3g"
        # error_prob = "1e-12"
        # cut_off_time = None

        file_name_start = prot_name + "_node-" + set_number + "_p_g-" + str(error_prob) + "_p_m-" + str(error_prob)

        if cut_off_time is not None:
            file_name_start += "_cut_off_time-" + str(cut_off_time)

        print(file_name_start)

        update_result_files(folder, file_name_start.split(".csv")[0])

        selected_files = []
        for file in os.listdir(folder):
            if file.startswith(file_name_start) and not file.endswith(".pkl") and not file.endswith("_failed.csv"):
                selected_files.append(file)

        if len(selected_files) != 1:
            raise ImportError(f"The program didn't find one single file, but instead {len(selected_files)} of them.")

        print(f'\nSelected file is {selected_files[0]}.')

        superoperator = Superoperator(folder + selected_files[0])
        #                                  GHZ_success=1.0,
        #                                  additional_superoperators=additional, failed_ghz_superoperator=failed,
        #                                  supop_date_time=supop_date_time)

        print(calculate_weighted_sum(superoperator))
        print(superoperator.sup_op_elements_p)
        element = next((x.p for x in superoperator.sup_op_elements_p if x.error_array == ['I','I','I','I'] and x.lie is False), None)
        print(element)

        # raise ValueError

        file = folder + selected_files[0]

        data = pd.read_csv(file, sep=';', index_col=[0, 1])

        average_prot_time = data.loc[('IIII', False), 'avg_duration']
        total_link_attempts = data.loc[('IIII', False), 'total_link_attempts']
        total_iterations = data.loc[('IIII', False), 'written_to']
        link_attempt_time = data.loc[('IIII', False), 't_link']
        T1n_link = data.loc[('IIII', False), 'T1n_link']
        T1n_idle = data.loc[('IIII', False), 'T1n_idle']
        T2n_link = data.loc[('IIII', False), 'T2n_link']
        T2n_idle = data.loc[('IIII', False), 'T2n_idle']

        average_link_generating_time = total_link_attempts/total_iterations/2*link_attempt_time

        print(f'\nAverage protocol time        = {average_prot_time} seconds.')
        print(f'Total amount of iterations   = {total_iterations}.')
        print(f'Total link attempts          = {total_link_attempts}.')
        print(f'Link generation attempt time = {link_attempt_time} seconds.')
        print(f'Average time spend on links  = {average_link_generating_time} seconds.')

        print(f'\nLink generation happened on average {average_link_generating_time/average_prot_time*100}% of the time.')
        print(f'Effect of T1n_link = {1-math.exp(-average_link_generating_time/T1n_link)}.')
        print(f'Effect of T1n_idle = {1-math.exp(-(average_prot_time-average_link_generating_time)/T1n_idle)}.')
        print(f'Effect of T2n_link = {1-math.exp(-average_link_generating_time/T2n_link)}.')
        print(f'Effect of T2n_idle = {1-math.exp(-(average_prot_time-average_link_generating_time)/T2n_idle)}.')

    elif calc_type == 12:
        print("Running instance 12 of main.")
        # Calculate if protocols are below threshold for various cut-off times and error probabilities.

        from oopsc.threshold.sim import update_result_files as update_result_files_sc
        from oopsc.superoperator.superoperator import Superoperator
        from circuit_simulation.stabilizer_measurement_protocols.run_protocols import update_result_files as update_result_files_so
        from circuit_simulation._superoperator.superoperator_methods import calculate_weighted_sum
        import calculate_thresholds as ct

        #FENDER

        # Settings used:
        n = 4
        max_prots_per_n = 30
        max_prots_per_n_k = 30
        p_g = [0, 1]
        target_ghz_success_rate = None   # [0.92, 0.95, 0.975, 0.99, 0.995, 0.9975, 0.999]:  # 92 95 97.5 99 99.5 99.75 99.9
        set_name = "Set3r"
        set_for_which_prots_are_found = "all_sV"
        nDD = None
        decoder = "uf"
        folder_known_protocols = "Thresholds_known_prots"   # "in_Set..." is automatically added to the folder name
        # folder_known_protocols = "Thresholds_direct_GHZ_40_original_superoperator_MWPM\\"
        # decoder = "mwpm"
        # set_name = "SetNickerson"
        # folder_known_protocols = "Paper_test_comparisons\\Test_known_prots_Nickerson_results\\Thresholds_known_prots_Stringent_MWPM_correct_lattices\\"

        explore_new_protocols = True
        review_old_results = False
        update_csv_files = True
        output_file = True
        new_p_g = [0.00005, 0.0001, 0.00015, 0.0002, 0.00025, 0.0003]
        new_p_g_increment = 1e-05
        interactive_plot = False
        plot_style = "paper"
        update_protocol_information = False

        show_fits = True
        output_fits = ""
        zoomed_in_fit = True
        zoom_number = 3

        calculate_cut_off_times = False
        GHZ_success_percentages = [0.92, 0.95, 0.975, 0.99, 0.995, 0.9975, 0.999, 0.9999]
        cut_off_times_per_protocol_and_per_percentage_for_10m12 = {}

        include_superoperator_statistics_decoherence = True
        include_superoperator_statistics_stab_fids = True
        include_superoperator_statistics_weighted_sum = False
        output_threshold_success_rates_dataframe = False

        plot_cut_off_time_dependence = False
        plot_cut_off_time_dependence_subplots = False
        plot_cut_off_subplot_one_figure = True
        prot_cut_off_time_dependence = "simv2_sVp_4_14_1"  # "refined2" # "simv3_sIIIc_4_7_1" # "simv2_sVp_4_15_1"
        plot_cut_off_time_filtered_out_cots = [0.0164681, 0.0382001, 0.0418481, 0.0783281, 0.0844081, 0.21783452, 0.23440386, 0.26349148]
        plot_cut_off_time_add_cut_off_to_location = False
        # Setting one of the following to True overrules "explore_new_protocols", "set_name" and
        # "set_for_which_prots_are_found" and loops over a number of settings:
        plot_link_efficiency_thresholds = False
        plot_bell_succ_prob_thresholds = False
        plot_bell_quality_thresholds = False
        only_show_best_new_protocols_in_plot = True
        font_size_big_start = 16 if plot_style == "paper" else 15.06
        font_size_small_start = 14 if plot_style == "paper" else 15.06
        figure_scale_factor = 1
        number_scaling_tries = 3 if plot_style == "paper" else 1

        if interactive_plot is True and sys.platform != "linux":
            mpl.use('TkAgg')

        inf = np.inf
        protocols_to_be_plotted_per_set = None
        cut_off_dict_set_plot = None

        # For plot coherence time (link efficiency) dependence:
        # protocols_to_be_plotted_per_set = {'Set3e': [('basic', 0.0361161, (0.000896803033659248, (0.0008395148641440959, 0.0009540912031744001))), ('minimum4x_22', 0.0434841, (0.00015787152024141458, (0.00013785075634828515, 0.00017789228413454401))), ('simv3_sIIIc_4_7_1', 0.0226081, (0.0014562245787016648, (0.0013710084720609176, 0.001541440685342412)))], 'Set3f': [('basic', 0.0361161, (0.0010639633781494033, (0.0008912055183059751, 0.0012367212379928315))), ('medium', 0.08353444, (0.0009683053779076497, (0.0008669843386521982, 0.0010696264171631011))), ('minimum4x_22', 0.0422561, (0.0011245633147777985, (0.0008961718496082406, 0.0013529547799473564))), ('expedient', 0.0619041, (0.0013379999999999998, (0.0013296664760233086, 0.001346333523976691))), ('refined', 0.13658286, (0.0007824007181837058, (0.0007097551159547833, 0.0008550463204126283))), ('minimum4x_40', 0.09911554, (0.0008029496722335393, (0.0007345510958240679, 0.0008713482486430107))), ('stringent', 0.2301401, (0.00014833241901584217, (0.0001355286249872131, 0.00016113621304447123))), ('simv3_sIIIc_4_7_1', 0.0262921, (0.002184178974869103, (0.002175070368204364, 0.0021932875815338423))), ('simv5_sIIIf_4_8_1', 0.0324321, (0.002188545001194171, (0.0021756393889543224, 0.00220145061343402)))], 'Set3g': [('medium', 0.08322, (0.000890573424106605, (0.0007687567505194151, 0.0010123900976937948))), ('expedient', 0.0619041, (0.0014954335646524546, (0.0014908700116214914, 0.0014999971176834178))), ('refined', 0.1343561, (0.000891181000709884, (0.0008010543381046125, 0.0009813076633151555))), ('minimum4x_40', 0.09787172, (0.0008114839592807536, (0.0007261533068251798, 0.0008968146117363273))), ('stringent', 0.2289121, (0.0005322459367644097, (0.0005093020651652352, 0.0005551898083635843))), ('simv3_sIIIc_4_7_1', 0.0262921, (0.0022276726208300904, (0.002215910212782088, 0.0022394350288780927))), ('simv5_sIIIf_4_8_1', 0.0324321, (0.0022480613816715763, (0.0022353263865144287, 0.002260796376828724)))], 'Set3k': [('basic', inf, (0.003268080852612509, (0.003094642771266074, 0.003441518933958944))), ('medium', inf, (0.0034336429124418194, (0.00326804303181371, 0.0035992427930699286))), ('minimum4x_22', inf, (0.005154461383948146, (0.0051231783720424124, 0.005185744395853879))), ('expedient', inf, (0.004669408571574415, (0.004471300738054236, 0.004867516405094595))), ('refined', inf, (0.005316880828919273, (0.004978859857435549, 0.005654901800402997))), ('minimum4x_40', inf, (0.005046726571507892, (0.004389336327497114, 0.0057041168155186695))), ('stringent', inf, (0.006142539685223198, (0.005965266699220217, 0.00631981267122618))), ('simv3_sIIIc_4_7_1', inf, (0.004153302951797575, (0.003773023117694556, 0.004533582785900594)))], 'Set3d': [('simv4_sIIId_4_6_9', 0.0189241, (0.0001156567935136613, (8.221732293352998e-05, 0.00014909626409379263)))], 'Set3c': [('simv3_sIIIc_4_7_1', 0.0226081, (0.0007704610580540276, (0.0007541058694578195, 0.0007868162466502357)))]}
        # protocols_to_be_plotted_per_set = {'Set3g': [('basic', 0.0361161, (0.0016376877267297604, (0.0014561571570282082, 0.0018192182964313127))), ('minimum4x_22', 0.0422561, (0.0014205539871802548, (0.001209592922526744, 0.0016315150518337656))), ('expedient', 0.0631321, (0.0014049780068228074, (0.0011620020172776963, 0.0016479539963679185))), ('refined', 0.1343561, (0.001105685571594368, (0.0005175643062356025, 0.0016938068369531337))), ('simv3_sIIIc_4_7_1', 0.0262921, (0.002056476988292664, (0.001912536185985618, 0.0022004177905997103))), ('simv5_sIIIg_4_9_13', 0.0410281, (0.002121116595060285, (0.002055555952734411, 0.002186677237386159)))], 'Set3p': [('simv1_sIIIp_4_6_9', 0.0180281, (0.0007503872885331377, (0.0006455276464316685, 0.000855246930634607)))], 'Set3c': [('simv3_sIIIc_4_7_1', 0.0226081, (0.0011246010698722714, (0.0010924825029141264, 0.0011567196368304164)))], 'Set3e': []}
        # For WB presentation Jan 2023:
        # protocols_to_be_plotted_per_set = {'Set3e': [('basic', 0.0361161, (0.000896803033659248, (0.0008395148641440959, 0.0009540912031744001))), ('minimum4x_22', 0.0434841, (0.00015787152024141458, (0.00013785075634828515, 0.00017789228413454401))), ('simv4_sIIIe_4_7_1', 0.0261161, (0.0017861990091229028, (0.001715882919527722, 0.0018565150987180836)))], 'Set3f': [('basic', 0.0361161, (0.0010639633781494033, (0.0008912055183059751, 0.0012367212379928315))), ('medium', 0.08353444, (0.0009683053779076497, (0.0008669843386521982, 0.0010696264171631011))), ('minimum4x_22', 0.0422561, (0.0011245633147777985, (0.0008961718496082406, 0.0013529547799473564))), ('expedient', 0.0619041, (0.0013379999999999998, (0.0013296664760233086, 0.001346333523976691))), ('refined', 0.13658286, (0.0007824007181837058, (0.0007097551159547833, 0.0008550463204126283))), ('minimum4x_40', 0.09911554, (0.0008029496722335393, (0.0007345510958240679, 0.0008713482486430107))), ('stringent', 0.2301401, (0.00014833241901584217, (0.0001355286249872131, 0.00016113621304447123))), ('simv5_sIIIf_4_11_14', 0.0686001, (0.001958496966091688, (0.0018653348721224747, 0.0020516590600609015)))], 'Set3g': [('medium', 0.08322, (0.000890573424106605, (0.0007687567505194151, 0.0010123900976937948))), ('expedient', 0.0619041, (0.0014954335646524546, (0.0014908700116214914, 0.0014999971176834178))), ('refined', 0.1343561, (0.000891181000709884, (0.0008010543381046125, 0.0009813076633151555))), ('minimum4x_40', 0.09787172, (0.0008114839592807536, (0.0007261533068251798, 0.0008968146117363273))), ('stringent', 0.2289121, (0.0005322459367644097, (0.0005093020651652352, 0.0005551898083635843))), ('simv4_sIIIg_4_8_18', 0.0369841, (0.002129465228543283, (0.002041812752784184, 0.002217117704302382)))], 'Set3k': [('basic', inf, (0.003268080852612509, (0.003094642771266074, 0.003441518933958944))), ('medium', inf, (0.0034336429124418194, (0.00326804303181371, 0.0035992427930699286))), ('minimum4x_22', inf, (0.005154461383948146, (0.0051231783720424124, 0.005185744395853879))), ('expedient', inf, (0.004669408571574415, (0.004471300738054236, 0.004867516405094595))), ('refined', inf, (0.005316880828919273, (0.004978859857435549, 0.005654901800402997))), ('minimum4x_40', inf, (0.005046726571507892, (0.004389336327497114, 0.0057041168155186695))), ('stringent', inf, (0.006142539685223198, (0.005965266699220217, 0.00631981267122618)))], 'Set3r': [('simv1_sIIIr_4_6_1', 0.0187441, (0.0001996196894892424, (0.0001232580675259688, 0.000275981311452516)))], 'Set3d': [('simv4_sIIId_4_6_17', 0.0187441, (0.00046063578674533737, (0.0003746136651059909, 0.0005466579083846839)))], 'Set3q': [('simv1_sIIIq_4_6_9', 0.0187441, (0.0006497738084981, (0.0005645003157271258, 0.0007350473012690743)))], 'Set3p': [('simv1_sIIIp_4_6_9', 0.0180281, (0.0007503872885331377, (0.0006455276464316685, 0.000855246930634607)))], 'Set3c': [('simv3_sIIIc_4_7_6', 0.0262921, (0.001101306286956197, (0.0009812022838407014, 0.0012214102900716924)))]}
        # For Fujitsu deliverable document Jan 2023:
        # protocols_to_be_plotted_per_set = {'Set3g': [('basic', 0.0382001, (0.0014601743419022878, (0.0014106714472183312, 0.0015096772365862444))), ('medium', 0.0941361, (0.0009187691463983787, (0.0008086103454422766, 0.001028927947354481))), ('minimum4x_22', 0.0503601, (0.0018676048031442266, (0.0017907643425306764, 0.0019444452637577768))), ('expedient', 0.0722481, (0.0014652059648732026, (0.001373721376543775, 0.0015566905532026303))), ('refined', 0.15934372, (0.0009179598183979425, (0.0007704603068771166, 0.0010654593299187685))), ('minimum4x_40', 0.13547996, (0.0008696338551967673, (0.0007458532620526788, 0.000993414448340856))), ('simv4_sIIIg_4_8_18', 0.0369841, (0.002129465228543283, (0.002041812752784184, 0.002217117704302382)))], 'Set3h': [('basic', inf, (0.0017313326423497497, (0.001683717481699056, 0.0017789478030004434))), ('medium', inf, (0.001613676190938412, (0.0015748355959919523, 0.0016525167858848719))), ('minimum4x_22', inf, (0.002360001769887041, (0.002312812226860952, 0.00240719131291313))), ('expedient', inf, (0.0020635544747414926, (0.0019782807602726706, 0.0021488281892103147))), ('refined', inf, (0.0025870925342684004, (0.002536471154153283, 0.002637713914383518))), ('minimum4x_40', inf, (0.0024773662430965966, (0.002394483550571871, 0.0025602489356213224))), ('stringent', inf, (0.0030926719159396696, (0.00300665486648711, 0.003178688965392229)))], 'Set3k': [('plain', inf, (0.004320269710968336, (0.0008355924324274741, 0.007804946989509199))), ('basic', inf, (0.003655147954589184, (0.0035671064633631564, 0.003743189445815211))), ('medium', inf, (0.0035204278452161774, (0.0034468155601528124, 0.0035940401302795425))), ('minimum4x_22', inf, (0.005168358643637008, (0.005091817204560112, 0.005244900082713904))), ('expedient', inf, (0.004939702627174901, (0.004859767439793643, 0.005019637814556159))), ('refined', inf, (0.005380112981118627, (0.005289885747794558, 0.005470340214442695))), ('minimum4x_40', inf, (0.005436090213877059, (0.0053322919627976735, 0.005539888464956444))), ('stringent', inf, (0.00601440160789508, (0.0057229099336595125, 0.006305893282130647)))], 'Set3r': [('simv1_sIIIr_4_6_1', 0.0187441, (0.0001996196894892424, (0.0001232580675259688, 0.000275981311452516)))], 'Set3d': [('simv4_sIIId_4_6_17', 0.0187441, (0.00046063578674533737, (0.0003746136651059909, 0.0005466579083846839)))], 'Set3q': [('simv1_sIIIq_4_6_9', 0.0187441, (0.0006497738084981, (0.0005645003157271258, 0.0007350473012690743)))], 'Set3p': [('simv1_sIIIp_4_6_9', 0.0180281, (0.0007503872885331377, (0.0006455276464316685, 0.000855246930634607)))], 'Set3c': [('simv3_sIIIc_4_7_1', 0.0248241, (0.001121848792582397, (0.0010220556913012806, 0.0012216418938635135)))], 'Set3e': [('simv4_sIIIe_4_7_1', 0.0261161, (0.0017861990091229028, (0.001715882919527722, 0.0018565150987180836)))], 'Set3f': [('simv5_sIIIf_4_11_14', 0.0686001, (0.001958496966091688, (0.0018653348721224747, 0.0020516590600609015)))]}
        # protocols_to_be_plotted_per_set = {'Set3c': [('basic', 0.0394161, (0.0002912535604712159, (0.00015736867411362073, 0.000425138446828811))), ('simv3_sIIIc_4_7_1', 0.0248241, (0.001121848792582397, (0.0010220556913012806, 0.0012216418938635135)))], 'Set3e': [('basic', 0.0382001, (0.0010133700047587194, (0.0009327982693102191, 0.00109394174020722))), ('expedient', 0.0771121, (0.00019048475933784167, (0.00012000040262817566, 0.00026096911604750765))), ('simv4_sIIIe_4_7_1', 0.0261161, (0.0017861990091229028, (0.001715882919527722, 0.0018565150987180836)))], 'Set3f': [('basic', 0.0382001, (0.0014418216977431693, (0.0013758278317199525, 0.001507815563766386))), ('medium', 0.0904881, (0.0008349003770320473, (0.0007473840913549363, 0.0009224166627091582))), ('expedient', 0.0686001, (0.0014099878170523003, (0.0013560737729891297, 0.001463901861115471))), ('refined', 0.1518883, (0.0007370937338724404, (0.0006082108689228966, 0.0008659765988219843))), ('simv5_sIIIf_4_11_14', 0.0686001, (0.001958496966091688, (0.0018653348721224747, 0.0020516590600609015)))], 'Set3g': [('basic', 0.0382001, (0.0014601743419022878, (0.0014106714472183312, 0.0015096772365862444))), ('medium', 0.0941361, (0.0009187691463983787, (0.0008086103454422766, 0.001028927947354481))), ('minimum4x_22', 0.0503601, (0.0018676048031442266, (0.0017907643425306764, 0.0019444452637577768))), ('expedient', 0.0722481, (0.0014652059648732026, (0.001373721376543775, 0.0015566905532026303))), ('refined', 0.15934372, (0.0009179598183979425, (0.0007704603068771166, 0.0010654593299187685))), ('minimum4x_40', 0.13547996, (0.0008696338551967673, (0.0007458532620526788, 0.000993414448340856))), ('simv4_sIIIg_4_8_18', 0.0369841, (0.002129465228543283, (0.002041812752784184, 0.002217117704302382)))], 'Set3h': [('basic', inf, (0.0017313326423497497, (0.001683717481699056, 0.0017789478030004434))), ('medium', inf, (0.001613676190938412, (0.0015748355959919523, 0.0016525167858848719))), ('minimum4x_22', inf, (0.002360001769887041, (0.002312812226860952, 0.00240719131291313))), ('expedient', inf, (0.0020635544747414926, (0.0019782807602726706, 0.0021488281892103147))), ('refined', inf, (0.0025870925342684004, (0.002536471154153283, 0.002637713914383518))), ('minimum4x_40', inf, (0.0024773662430965966, (0.002394483550571871, 0.0025602489356213224))), ('stringent', inf, (0.0030926719159396696, (0.00300665486648711, 0.003178688965392229)))], 'Set3k': [('plain', inf, (0.004320269710968336, (0.0008355924324274741, 0.007804946989509199))), ('basic', inf, (0.003655147954589184, (0.0035671064633631564, 0.003743189445815211))), ('medium', inf, (0.0035204278452161774, (0.0034468155601528124, 0.0035940401302795425))), ('minimum4x_22', inf, (0.005168358643637008, (0.005091817204560112, 0.005244900082713904))), ('expedient', inf, (0.004939702627174901, (0.004859767439793643, 0.005019637814556159))), ('refined', inf, (0.005380112981118627, (0.005289885747794558, 0.005470340214442695))), ('minimum4x_40', inf, (0.005436090213877059, (0.0053322919627976735, 0.005539888464956444))), ('stringent', inf, (0.00601440160789508, (0.0057229099336595125, 0.006305893282130647)))], 'Set3r': [('simv1_sIIIr_4_6_1', 0.0187441, (0.0001996196894892424, (0.0001232580675259688, 0.000275981311452516)))], 'Set3d': [('simv4_sIIId_4_6_17', 0.0187441, (0.00046063578674533737, (0.0003746136651059909, 0.0005466579083846839)))], 'Set3q': [('simv1_sIIIq_4_6_9', 0.0187441, (0.0006497738084981, (0.0005645003157271258, 0.0007350473012690743)))], 'Set3p': [('simv1_sIIIp_4_6_9', 0.0180281, (0.0007503872885331377, (0.0006455276464316685, 0.000855246930634607)))]}
        # Draft April 14, 2023:
        # protocols_to_be_plotted_per_set = {'Set3c': [('basic', 0.0394161, (0.0002912535604712159, (0.00015736867411362073, 0.000425138446828811))), ('simv1_sIIIr_4_6_1', 0.0187441, (0.0008050085254869597, (0.0007652977312892646, 0.0008447193196846548))), ('simv3_sIIIc_4_7_1', 0.0248241, (0.001121848792582397, (0.0010220556913012806, 0.0012216418938635135))), ('simv3_sIIIc_4_7_6', 0.0262921, (0.001101306286956197, (0.0009812022838407014, 0.0012214102900716924))), ('simv4_sIIIc_4_7_14', 0.0274641, (0.0010986062521766424, (0.001003676408290987, 0.001193536096062298))), ('simv5_sIIIf_4_11_2', 0.0564401, (0.00016591276736908518, (0.0001178089661081392, 0.00021401656863003117)))], 'Set3e': [('basic', 0.0382001, (0.0010133700047587194, (0.0009327982693102191, 0.00109394174020722))), ('minimum4x_22', 0.0442801, (0.0005158491002163272, (0.00045318304264418133, 0.000578515157788473))), ('expedient', 0.0661681, (0.00023450261856885219, (0.00016480850724761702, 0.00030419672989008735))), ('simv1_sIIIr_4_6_1', 0.0187441, (0.0011402929415552643, (0.0011220417891078711, 0.0011585440940026575))), ('simv3_sIIIc_4_7_1', 0.0236081, (0.001808573478860709, (0.0017766901741722023, 0.0018404567835492158))), ('simv3_sIIIc_4_7_6', 0.0236081, (0.0017979683246157462, (0.0017594025082338025, 0.00183653414099769))), ('simv4_sIIIc_4_7_14', 0.0274641, (0.001772275575253924, (0.0017101136731642666, 0.0018344374773435813))), ('simv4_sIIIe_4_7_3', 0.0261161, (0.001758219036096265, (0.001681485176444344, 0.0018349528957481858))), ('simv5_sIIIf_4_11_2', 0.0588721, (0.0012465080037756804, (0.0011678502395172622, 0.0013251657680340987)))], 'Set3f': [('basic', 0.0382001, (0.0014418216977431693, (0.0013758278317199525, 0.001507815563766386))), ('medium', 0.0783281, (0.0009031308096932074, (0.0008203211355515595, 0.0009859404838348552))), ('minimum4x_22', 0.0442801, (0.0017592833424916006, (0.0017147039202514507, 0.0018038627647317504))), ('expedient', 0.0686001, (0.0014099878170523003, (0.0013560737729891297, 0.001463901861115471))), ('refined', 0.1162963, (0.0008826041667199956, (0.0007976667540840397, 0.0009675415793559515))), ('minimum4x_40', 0.0941012, (0.0007779909864811638, (0.0006921083428341987, 0.0008638736301281288))), ('stringent', 0.1743921, (0.00032535039639465453, (0.0002775532898485434, 0.0003731475029407657))), ('simv1_sIIIr_4_6_1', 0.0187441, (0.0013191465263666705, (0.001301461772826312, 0.0013368312799070291))), ('simv3_sIIIc_4_7_1', 0.0248241, (0.002184897180262896, (0.002144626176419992, 0.0022251681841058))), ('simv3_sIIIc_4_7_6', 0.0248241, (0.0021698510294745417, (0.0021274952308028053, 0.002212206828146278))), ('simv4_sIIIc_4_7_14', 0.0274641, (0.0021719077744496327, (0.002136517297132182, 0.0022072982517670836))), ('simv4_sIIIe_4_7_3', 0.0248241, (0.0021853370404023935, (0.002155482103402791, 0.002215191977401996))), ('simv5_sIIIf_4_11_2', 0.0686001, (0.001953541221818235, (0.0018622097048385463, 0.0020448727387979235)))], 'Set3g': [('basic', 0.0382001, (0.0014601743419022878, (0.0014106714472183312, 0.0015096772365862444))), ('medium', 0.07743772, (0.0009930082725113786, (0.0009208549032399436, 0.0010651616417828135))), ('minimum4x_22', 0.0442801, (0.0018750796488213675, (0.0018462923494586552, 0.0019038669481840798))), ('expedient', 0.0661681, (0.00150864048474626, (0.0014586471703844032, 0.0015586337991081167))), ('refined', 0.11555358, (0.0011183421480504459, (0.0010251734716493744, 0.0012115108244515173))), ('minimum4x_40', 0.09496762, (0.0010020923577775055, (0.000877201257467098, 0.0011269834580879131))), ('stringent', 0.1792561, (0.0005656048992565089, (0.000510001025748853, 0.0006212087727641649))), ('simv1_sIIIr_4_6_1', 0.0187441, (0.001337232934554783, (0.0013117518336566703, 0.0013627140354528957))), ('simv3_sIIIc_4_7_1', 0.0248241, (0.0022197823877677337, (0.0021729075615401503, 0.002266657213995317))), ('simv3_sIIIc_4_7_6', 0.0248241, (0.0021985883406806083, (0.002166931987735185, 0.0022302446936260318))), ('simv4_sIIIc_4_7_14', 0.0274641, (0.002212898763223719, (0.0021759536815612866, 0.0022498438448861516))), ('simv4_sIIIe_4_7_3', 0.0248241, (0.002209262747106016, (0.0021744522472305363, 0.0022440732469814957))), ('simv5_sIIIf_4_11_2', 0.06340308, (0.002068398723950525, (0.002013459285202496, 0.0021233381626985536)))], 'Set3h': [('basic', inf, (0.0017313326423497497, (0.001683717481699056, 0.0017789478030004434))), ('medium', inf, (0.001613676190938412, (0.0015748355959919523, 0.0016525167858848719))), ('minimum4x_22', inf, (0.002360001769887041, (0.002312812226860952, 0.00240719131291313))), ('expedient', inf, (0.0020635544747414926, (0.0019782807602726706, 0.0021488281892103147))), ('refined', inf, (0.0025870925342684004, (0.002536471154153283, 0.002637713914383518))), ('minimum4x_40', inf, (0.0024773662430965966, (0.002394483550571871, 0.0025602489356213224))), ('stringent', inf, (0.0030926719159396696, (0.00300665486648711, 0.003178688965392229))), ('simv1_sIIIr_4_6_1', inf, (0.0014102880483641265, (0.0013908854968992922, 0.0014296905998289608))), ('simv3_sIIIc_4_7_1', inf, (0.00235483127926136, (0.0022717791100665095, 0.0024378834484562106))), ('simv3_sIIIc_4_7_6', inf, (0.0024374500238284772, (0.002399702709694824, 0.0024751973379621307))), ('simv4_sIIIc_4_7_14', inf, (0.002448842299963221, (0.00241912021068163, 0.002478564389244812))), ('simv4_sIIIe_4_7_3', inf, (0.0024402347179365174, (0.0024049376271137405, 0.0024755318087592943))), ('simv5_sIIIf_4_11_2', inf, (0.002672419662149637, (0.002643187635296757, 0.0027016516890025167)))], 'Set3k': [('plain', inf, (0.004320269710968336, (0.0008355924324274741, 0.007804946989509199))), ('basic', inf, (0.003655147954589184, (0.0035671064633631564, 0.003743189445815211))), ('medium', inf, (0.0035204278452161774, (0.0034468155601528124, 0.0035940401302795425))), ('minimum4x_22', inf, (0.005168358643637008, (0.005091817204560112, 0.005244900082713904))), ('expedient', inf, (0.004939702627174901, (0.004859767439793643, 0.005019637814556159))), ('refined', inf, (0.005380112981118627, (0.005289885747794558, 0.005470340214442695))), ('minimum4x_40', inf, (0.005436090213877059, (0.0053322919627976735, 0.005539888464956444))), ('stringent', inf, (0.00601440160789508, (0.0057229099336595125, 0.006305893282130647))), ('simv1_sIIIr_4_6_1', inf, (0.003232544002392771, (0.0031849796711691768, 0.003280108333616365))), ('simv3_sIIIc_4_7_1', inf, (0.004911604191557384, (0.004831454815065121, 0.004991753568049646))), ('simv3_sIIIc_4_7_6', inf, (0.004916490556576346, (0.0048546958828165196, 0.004978285230336172))), ('simv4_sIIIc_4_7_14', inf, (0.004938686064760127, (0.004877554096492128, 0.004999818033028127))), ('simv4_sIIIe_4_7_3', inf, (0.004911089497139029, (0.00485844336144481, 0.004963735632833248))), ('simv5_sIIIf_4_11_2', inf, (0.004792715175491646, (0.0046974279719672754, 0.004888002379016016)))], 'Set3r': [('simv1_sIIIr_4_6_1', 0.0187441, (0.0001996196894892424, (0.0001232580675259688, 0.000275981311452516)))], 'Set3d': [('simv1_sIIIr_4_6_1', 0.0187441, (0.000444049164888717, (0.0003807363331928895, 0.0005073619965845445))), ('simv4_sIIId_4_6_17', 0.0187441, (0.00046063578674533737, (0.0003746136651059909, 0.0005466579083846839))), ('simv3_sIIIc_4_7_1', 0.0236081, (0.00045796930762364115, (0.00037947785682522777, 0.0005364607584220545))), ('simv3_sIIIc_4_7_6', 0.0236081, (0.0004497906219954717, (0.000383493878443328, 0.0005160873655476154))), ('simv4_sIIIc_4_7_14', 0.0236081, (0.00041960645951207145, (0.00034169366828472726, 0.0004975192507394156))), ('simv4_sIIIe_4_7_3', 0.0236081, (0.00042893011562771547, (0.0003384729053914307, 0.0005193873258640002)))], 'Set3q': [('simv1_sIIIr_4_6_1', 0.0187441, (0.0006063295719860612, (0.0005558614175771729, 0.0006567977263949496))), ('simv3_sIIIc_4_7_1', 0.0236081, (0.0007336911881945261, (0.0006702940885892379, 0.0007970882877998143))), ('simv3_sIIIc_4_7_6', 0.0236081, (0.0007585424224963638, (0.0006991847501963854, 0.0008179000947963422))), ('simv4_sIIIc_4_7_14', 0.0236081, (0.0007223599898541096, (0.0006490955576924944, 0.0007956244220157248))), ('simv4_sIIIe_4_7_3', 0.0236081, (0.0007573931898384736, (0.0006963342095710587, 0.0008184521701058885)))], 'Set3p': [('simv1_sIIIp_4_6_9', 0.0180281, (0.0007503872885331377, (0.0006455276464316685, 0.000855246930634607)))]}

        # For plot Bell success probability dependence:
        # protocols_to_be_plotted_per_set = {'Set6g': [('simv1_sVIg_4_7_10', 0.0293681, (0.0007651971204719757, (0.0006735604261282876, 0.0008568338148156638)))], 'Set6h': [('simv1_sVIh_4_7_2', 0.0269041, (0.0014268471351021727, (0.0013503104019224725, 0.001503383868281873)))]}
        # protocols_to_be_plotted_per_set = {'Set6k': [('simv1_sVIk_4_6_9', 0.0241681, (0.00036128138375584717, (0.0002635977293936293, 0.00045896503811806505)))], 'Set6g': [('simv1_sVIg_4_7_10', 0.0293681, (0.0007651971204719757, (0.0006735604261282876, 0.0008568338148156638)))], 'Set6m': [('simv1_sVIm_4_6_1', 0.0204241, (0.000822928647399164, (0.000763196900181399, 0.000882660394616929)))], 'Set6h': [('simv1_sVIh_4_7_2', 0.0269041, (0.0014268471351021727, (0.0013503104019224725, 0.001503383868281873)))], 'Set3e': [('simv4_sIIIe_4_7_3', 0.0261161, (0.0018079224977718474, (0.0017519690203877695, 0.0018638759751559253)))]}
        # For WB presentation Jan 2023:
        # protocols_to_be_plotted_per_set = {'Set6k': [('simv1_sVIk_4_6_9', 0.0241681, (0.0003463568984010415, (0.00026985805404553366, 0.0004228557427565494)))], 'Set6g': [('simv1_sVIg_4_7_9', 0.0311721, (0.0007432780262888303, (0.0006357750968980202, 0.0008507809556796403)))], 'Set6m': [('simv1_sVIm_4_6_1', 0.0204241, (0.0008116269984038758, (0.0007303384542137732, 0.0008929155425939784)))], 'Set6h': [('simv1_sVIh_4_7_2', 0.0269041, (0.0014483825421433354, (0.001376588426301714, 0.001520176657984957)))], 'Set3e': [('simv4_sIIIe_4_7_1', 0.0261161, (0.0017861990091229028, (0.001715882919527722, 0.0018565150987180836)))]}
        # For Fujitsu deliverable document Jan 2023:
        # protocols_to_be_plotted_per_set = {'Set6g': [('basic', 0.0501361, (4.119424605672646e-06, (1.2971312243597158e-06, 6.941717986985576e-06))), ('simv1_sVIg_4_7_9', 0.0311721, (0.0007432780262888303, (0.0006357750968980202, 0.0008507809556796403)))], 'Set6m': [('basic', 0.0455601, (0.00023859620795924896, (0.00015564095386338823, 0.00032155146205510967))), ('simv2_sVIm_4_7_6', 0.0286641, (0.001159434774968616, (0.0010509638248445417, 0.0012679057250926903)))], 'Set6h': [('basic', 0.0419201, (0.0006009473089571123, (0.0004875336762080086, 0.000714360941706216))), ('simv1_sVIh_4_7_2', 0.0269041, (0.0014483825421433354, (0.001376588426301714, 0.001520176657984957)))], 'Set3e': [('basic', 0.0382001, (0.0010133700047587194, (0.0009327982693102191, 0.00109394174020722))), ('expedient', 0.0771121, (0.00019048475933784167, (0.00012000040262817566, 0.00026096911604750765))), ('simv4_sIIIe_4_7_1', 0.0261161, (0.0017861990091229028, (0.001715882919527722, 0.0018565150987180836)))], 'Set6k': [('simv1_sVIk_4_6_9', 0.0241681, (0.0003463568984010415, (0.00026985805404553366, 0.0004228557427565494)))]}
        # Draft April 14, 2023:
        # protocols_to_be_plotted_per_set = {'Set6g': [('basic', 0.0501361, (4.119424605672646e-06, (1.2971312243597158e-06, 6.941717986985576e-06))), ('simv4_sIIIe_4_7_3', 0.0290801, (0.0007689976895498639, (0.0007136831441563166, 0.0008243122349434113)))], 'Set6m': [('basic', 0.0455601, (0.00023859620795924896, (0.00015564095386338823, 0.00032155146205510967))), ('simv2_sVIm_4_7_6', 0.0286641, (0.001159434774968616, (0.0010509638248445417, 0.0012679057250926903))), ('simv4_sIIIe_4_7_3', 0.0272561, (0.0011511421048539532, (0.0010868465341166201, 0.0012154376755912862)))], 'Set6h': [('basic', 0.0419201, (0.0006009473089571123, (0.0004875336762080086, 0.000714360941706216))), ('simv1_sVIh_4_7_2', 0.0269041, (0.0014483825421433354, (0.001376588426301714, 0.001520176657984957))), ('simv4_sIIIe_4_7_3', 0.0272241, (0.001434230655112338, (0.0013890563101060594, 0.0014794050001186165)))], 'Set3e': [('basic', 0.0382001, (0.0010133700047587194, (0.0009327982693102191, 0.00109394174020722))), ('minimum4x_22', 0.0442801, (0.0005158491002163272, (0.00045318304264418133, 0.000578515157788473))), ('expedient', 0.0661681, (0.00023450261856885219, (0.00016480850724761702, 0.00030419672989008735))), ('simv4_sIIIe_4_7_1', 0.0261161, (0.0017861990091229028, (0.001715882919527722, 0.0018565150987180836))), ('simv4_sIIIe_4_7_3', 0.0261161, (0.001758219036096265, (0.001681485176444344, 0.0018349528957481858)))], 'Set6k': [('simv1_sVIk_4_6_9', 0.0241681, (0.0003463568984010415, (0.00026985805404553366, 0.0004228557427565494))), ('simv4_sIIIe_4_7_3', 0.0313601, (0.0002722450359903902, (0.0002161034812808449, 0.0003283865906999355)))]}

        # For plot Bell quality dependence:
        # protocols_to_be_plotted_per_set = {'Set5a': [('simv1_sVa_4_9_17', 0.0324321, (0.000697654100984241, (0.0006455878531591717, 0.0007497203488093103)))], 'Set5b': [('simv1_sVb_4_9_1', 0.0287481, (0.0009447406620455676, (0.0008843119686656597, 0.0010051693554254755)))], 'Set5c': [('simv3_sVc_4_11_7', 0.0373441, (0.0014041392866953623, (0.0013279490457591024, 0.0014803295276316223)))], 'Set5d': [('simv1_sVd_4_11_2', 0.0361161, (0.0015627387458778217, (0.0014614383052444623, 0.001664039186511181)))]}
        # protocols_to_be_plotted_per_set = {'Set5p': [('simv1_sVp_4_11_18', 0.0406321, (0.0004773288204565467, (0.00038012973943435033, 0.000574527901478743)))], 'Set5a': [('simv1_sVa_4_9_17', 0.0324321, (0.000697654100984241, (0.0006455878531591717, 0.0007497203488093103)))], 'Set5b': [('simv1_sVb_4_9_1', 0.0287481, (0.0009447406620455676, (0.0008843119686656597, 0.0010051693554254755)))], 'Set5c': [('simv3_sVc_4_11_7', 0.0373441, (0.0014041392866953623, (0.0013279490457591024, 0.0014803295276316223)))], 'Set5d': [('simv1_sVd_4_11_2', 0.0361161, (0.0015627387458778217, (0.0014614383052444623, 0.001664039186511181)))], 'Set5e': [('simv1_sVe_4_10_3', 0.0571201, (0.0013277101436103907, (0.0012162572971762859, 0.0014391629900444955)))]}
        # For WB presentation Jan 2023:
        # protocols_to_be_plotted_per_set = {'Set5q': [('simv1_sVq_4_12_9', 0.0540081, (0.00027382404013529676, (0.00020396822967561784, 0.0003436798505949757)))], 'Set5p': [('simv1_sVp_4_11_18', 0.0406321, (0.0004773288204565467, (0.00038012973943435033, 0.000574527901478743)))], 'Set5a': [('simv1_sVa_4_10_10', 0.0361161, (0.0006909179383204923, (0.0006229453207516252, 0.0007588905558893593)))], 'Set5b': [('simv1_sVb_4_9_1', 0.0287481, (0.0009386658508314452, (0.0008199351191303687, 0.0010573965825325217)))], 'Set5c': [('simv3_sVc_4_11_7', 0.0373441, (0.0014006366627153206, (0.00132461335420166, 0.001476659971228981)))], 'Set5d': [('simv1_sVd_4_11_2', 0.0361161, (0.0015852720273923126, (0.0015052669387059154, 0.0016652771160787097)))], 'Set5e': [('simv2_sVe_4_7_7', 0.0236081, (0.001863661820490014, (0.0017211070450299495, 0.0020062165959500783)))]}
        # For Fujitsu deliverable document Jan 2023:
        # protocols_to_be_plotted_per_set = {'Set5c': [('minimum4x_22', 0.05679396, (8.448719831053967e-05, (1.1647594377155477e-05, 0.00015732680224392386))), ('simv3_sVc_4_11_7', 0.0373441, (0.0014006366627153206, (0.00132461335420166, 0.001476659971228981)))], 'Set5d': [('basic', 0.05679396, (0.00016969666993983242, (0.00010181067461985065, 0.00023758266525981419))), ('minimum4x_22', 0.05013948, (0.00035754848492805736, (0.00030205205398005234, 0.0004130449158760624))), ('simv1_sVd_4_11_2', 0.0361161, (0.0015852720273923126, (0.0015052669387059154, 0.0016652771160787097)))], 'Set5e': [('basic', 0.0357681, (0.0012394835460776822, (0.0011250966371592773, 0.001353870454996087))), ('medium', 0.07991772, (0.00027050555915955756, (0.0001385941434152648, 0.0004024169749038503))), ('minimum4x_22', 0.04190158, (0.0005832778334514696, (0.0004867422219357838, 0.0006798134449671555))), ('expedient', 0.0613041, (0.0003737233005360482, (0.00030527192508813785, 0.00044217467598395857))), ('simv2_sVe_4_7_7', 0.0236081, (0.001863661820490014, (0.0017211070450299495, 0.0020062165959500783)))], 'Set5r': [('simv1_sVr_4_13_9', 0.0613041, (6.369237932503925e-05, (1.1046177262590574e-05, 0.00011633858138748791)))], 'Set5q': [('simv1_sVq_4_12_1', 0.0540081, (0.00025764599793864316, (0.00019500485178584344, 0.00032028714409144287)))], 'Set5p': [('simv1_sVp_4_11_18', 0.0406321, (0.0004773288204565467, (0.00038012973943435033, 0.000574527901478743)))], 'Set5a': [('simv1_sVa_4_10_10', 0.0361161, (0.0006909179383204923, (0.0006229453207516252, 0.0007588905558893593)))], 'Set5b': [('simv1_sVb_4_9_1', 0.0287481, (0.0009386658508314452, (0.0008199351191303687, 0.0010573965825325217)))]}
        # Draft April 14, 2023:
        # protocols_to_be_plotted_per_set = {'Set5c': [('minimum4x_22', 0.04770748, (0.00011171032991459728, (8.311878582826577e-05, 0.0001403018740009288))), ('simv3_sVc_4_11_7', 0.0373441, (0.0014006366627153206, (0.00132461335420166, 0.001476659971228981)))], 'Set5d': [('basic', 0.0430641, (0.00021922503642695054, (0.00015021293899830917, 0.0002882371338555919))), ('minimum4x_22', 0.05013948, (0.00035754848492805736, (0.00030205205398005234, 0.0004130449158760624))), ('simv1_sVd_4_11_2', 0.0361161, (0.0015852720273923126, (0.0015052669387059154, 0.0016652771160787097)))], 'Set5e': [('basic', 0.0357681, (0.0012394835460776822, (0.0011250966371592773, 0.001353870454996087))), ('medium', 0.0698161, (0.00028683894124130997, (0.00019778163095784224, 0.0003758962515247777))), ('minimum4x_22', 0.04190158, (0.0005832778334514696, (0.0004867422219357838, 0.0006798134449671555))), ('expedient', 0.0613041, (0.0003737233005360482, (0.00030527192508813785, 0.00044217467598395857))), ('simv2_sVe_4_7_9', 0.0260401, (0.0018726414049063077, (0.0017902697334154338, 0.0019550130763971814)))], 'Set5r': [('simv1_sVr_4_13_9', 0.0613041, (6.369237932503925e-05, (1.1046177262590574e-05, 0.00011633858138748791)))], 'Set5q': [('simv1_sVq_4_12_1', 0.0540081, (0.00025764599793864316, (0.00019500485178584344, 0.00032028714409144287)))], 'Set5p': [('simv1_sVp_4_11_18', 0.0406321, (0.0004773288204565467, (0.00038012973943435033, 0.000574527901478743)))], 'Set5a': [('simv1_sVa_4_10_10', 0.0361161, (0.0006909179383204923, (0.0006229453207516252, 0.0007588905558893593)))], 'Set5b': [('simv1_sVb_4_9_1', 0.0287481, (0.0009386658508314452, (0.0008199351191303687, 0.0010573965825325217)))]}


        # protocols_to_be_plotted_per_set = {'Set3e': [('basic', 0.0361161, [0.00088, (0.000855, 0.000895)]), ('simv3_sIIIc_4_7_1', 0.0226081, [0.00166, (0.00164, 0.001685)])], 'Set3f': [('basic', 0.0361161, [0.00144, (0.00142, 0.00148)]), ('medium', 0.08353444, [0.00094, (0.0009, 0.00096)]), ('expedient', 0.0631321, [0.00135, (0.00132, 0.001375)]), ('refined', 0.13658286, [0.00075, (0.00075, 0.00085)]), ('stringent', 0.2301401, [0.00018, (0.00012, 0.0002)]), ('simv3_sIIIc_4_7_1', 0.0262921, [0.00219, (0.00216, 0.00222)])], 'Set3g': [('basic', 0.0361161, [0.0015, (0.00146, 0.00154)]), ('medium', 0.08322, [0.00105, (0.001025, 0.0011)]), ('expedient', 0.0631321, [0.0015, (0.00147, 0.00153)]), ('refined', 0.1343561, [0.001075, (0.00105, 0.001125)]), ('stringent', 0.2289121, [0.00056, (0.00052, 0.0006)]), ('simv3_sIIIc_4_7_1', 0.0262921, [0.00222, (0.0022, 0.00226)])], 'Set3k': [('basic', inf, [0.0036, (0.003575, 0.003675)]), ('medium', inf, [0.003475, (0.003425, 0.00355)]), ('expedient', inf, [0.00484, (0.00482, 0.0049)]), ('refined', inf, [0.00535, (0.0053, 0.005425)]), ('stringent', inf, [0.00615, (0.006, 0.0063)]), ('simv3_sIIIc_4_7_1', inf, [0.004775, (0.004775, 0.004925)])], 'Set3c': [('simv3_sIIIc_4_7_1', 0.0226081, [0.000775, (0.00075, 0.00078)])]}
        # protocols_to_be_plotted_per_set = {'Set3e': [('simv3_sIIIc_4_7_1', 0.0226081, [0.00166, (0.00164, 0.001685)]), ('basic', 0.0361161, [0.00088, (0.000855, 0.000895)]), ('minimum4x_22', 0.0434841, [0.00017, (0.00014, 0.00022)])], 'Set3f': [('basic', 0.0361161, [0.00144, (0.00142, 0.00148)]), ('medium', 0.08353444, [0.00094, (0.0009, 0.00096)]), ('minimum4x_22', 0.0422561, [0.00169, (0.00166, 0.00172)]), ('expedient', 0.0631321, [0.00135, (0.00132, 0.001375)]), ('refined', 0.13658286, [0.00075, (0.00075, 0.00085)]), ('minimum4x_40', 0.09911554, [0.0008, (0.00078, 0.00085)]), ('stringent', 0.2301401, [0.00018, (0.00012, 0.0002)]), ('simv3_sIIIc_4_7_1', 0.0262921, [0.00219, (0.00216, 0.00222)])], 'Set3g': [('basic', 0.0361161, [0.0015, (0.00146, 0.00154)]), ('medium', 0.08322, [0.00105, (0.001025, 0.0011)]), ('minimum4x_22', 0.0422561, [0.00184, (0.00182, 0.00188)]), ('expedient', 0.0631321, [0.0015, (0.00147, 0.00153)]), ('refined', 0.1343561, [0.001075, (0.00105, 0.001125)]), ('minimum4x_40', 0.09787172, [0.00105, (0.001025, 0.0011)]), ('stringent', 0.2289121, [0.00056, (0.00052, 0.0006)]), ('simv3_sIIIc_4_7_1', 0.0262921, [0.00222, (0.0022, 0.00226)])], 'Set3k': [('basic', inf, [0.0036, (0.003575, 0.003675)]), ('medium', inf, [0.003475, (0.003425, 0.00355)]), ('minimum4x_22', inf, [0.00512, (0.0051, 0.00515)]), ('expedient', inf, [0.00484, (0.00482, 0.0049)]), ('refined', inf, [0.00535, (0.0053, 0.005425)]), ('minimum4x_40', inf, [0.00535, (0.00535, 0.005475)]), ('stringent', inf, [0.00615, (0.006, 0.0063)]), ('simv3_sIIIc_4_7_1', inf, [0.004775, (0.004775, 0.004925)])], 'Set3c': [('simv3_sIIIc_4_7_1', 0.0226081, [0.000775, (0.00075, 0.00078)])]}


        # For plot cut-off dependence:
        # Draft July 4, 2023:
        # cut_off_information_plot = {0: ('simv3_sIIIc_4_7_1', 0, '\\eta^{\\ast}_\\mathrm{link}=8\\cdot10^2', 'Fig. 7'), 1: ('expedient', 1, '\\eta^{\\ast}_\\mathrm{link}=2\\cdot10^5', 'Fig. 7'), 2: ('refined2', 2, '\\eta^{\\ast}_\\mathrm{link}=2\\cdot10^5', 'Fig. 7'), 3: ('simv2_sVp_4_14_1', 3, 'F_\\mathrm{link}\\approx0.828', 'Fig. 6')}
        # cut_off_dict_set_plot = {0: {0.0150961: {'threshold': (0.0005743938988740482, (0.000503583904522898, 0.0006452038932251984)), 'GHZ_success_rates': {0.0004: 0.9132, 0.00055: 0.9105}, 'number_of_points': 6, 'minimum_number_iterations': 50000}, 0.0163121: {'threshold': (0.0007418498451764077, (0.0007029939877914512, 0.0007807057025613643)), 'GHZ_success_rates': {0.0004: 0.9366, 0.0007: 0.9323}, 'number_of_points': 8, 'minimum_number_iterations': 50000}, 0.0164681: {'threshold': (0.0007194873195265122, (0.0006066865764756691, 0.0008322880625773553)), 'GHZ_success_rates': {0.0007: 0.932}, 'number_of_points': 2, 'minimum_number_iterations': 50000}, 0.0175281: {'threshold': (0.0008638314394854218, (0.0008242935227431411, 0.0009033693562277025)), 'GHZ_success_rates': {0.0007: 0.9542, 0.00085: 0.9526}, 'number_of_points': 5, 'minimum_number_iterations': 150000}, 0.0187441: {'threshold': (0.000975985776436757, (0.0009254578146491928, 0.001026513738224321)), 'GHZ_success_rates': {0.0007: 0.9717, 0.00095: 0.9694}, 'number_of_points': 5, 'minimum_number_iterations': 150000}, 0.0199601: {'threshold': (0.0010368568211900563, (0.0009840547737837602, 0.0010896588685963525)), 'GHZ_success_rates': {0.0007: 0.982, 0.00095: 0.9802}, 'number_of_points': 5, 'minimum_number_iterations': 150000}, 0.0211761: {'threshold': (0.0010423153998406747, (0.000988224781940247, 0.0010964060177411025)), 'GHZ_success_rates': {0.0009: 0.9869, 0.00105: 0.986}, 'number_of_points': 4, 'minimum_number_iterations': 150000}, 0.0223921: {'threshold': (0.0011004188676653776, (0.0010575790391057587, 0.0011432586962249966)), 'GHZ_success_rates': {0.00095: 0.9922, 0.0011: 0.9917}, 'number_of_points': 6, 'minimum_number_iterations': 50000}, 0.0236081: {'threshold': (0.001104709609520086, (0.0010502475852307504, 0.0011591716338094217)), 'GHZ_success_rates': {0.00095: 0.9947, 0.00105: 0.9944}, 'number_of_points': 6, 'minimum_number_iterations': 150000}, 0.0248241: {'threshold': (0.001109688681917355, (0.0010568449457351746, 0.0011625324180995354)), 'GHZ_success_rates': {0.0009: 0.9964, 0.0011: 0.996}, 'number_of_points': 5, 'minimum_number_iterations': 100000}, 0.0284721: {'threshold': (0.0010882341402350185, (0.0010453949586948306, 0.0011310733217752063)), 'GHZ_success_rates': {0.0009: 0.9986, 0.00105: 0.9984}, 'number_of_points': 5, 'minimum_number_iterations': 150000}, 0.0309041: {'threshold': (0.0010888649431051482, (0.001044822993070359, 0.0011329068931399374)), 'GHZ_success_rates': {0.0009: 0.9994, 0.00105: 0.9994}, 'number_of_points': 5, 'minimum_number_iterations': 150000}, 0.0369841: {'threshold': (0.0010000253285013577, (0.000936776433303844, 0.0010632742236988714)), 'GHZ_success_rates': {0.00085: 0.9999, 0.001: 0.9999}, 'number_of_points': 7, 'minimum_number_iterations': 100000}, 0.0382001: {'threshold': (0.0009993136492000149, (0.0009389974965369364, 0.0010596298018630934)), 'GHZ_success_rates': {0.00075: 1.0, 0.00095: 1.0}, 'number_of_points': 5, 'minimum_number_iterations': 100000}, 0.0442801: {'threshold': (0.000933881175572018, (0.0008800237461495761, 0.0009877386049944599)), 'GHZ_success_rates': {0.00075: 1.0, 0.0009: 1.0}, 'number_of_points': 7, 'minimum_number_iterations': 100000}}, 1: {0.0406321: {'threshold': (0.0009129720252904336, (0.0008642747878660966, 0.0009616692627147705)), 'GHZ_success_rates': {0.0007: 0.8496, 0.0009: 0.8406}, 'number_of_points': 5, 'minimum_number_iterations': 100000}, 0.0430641: {'threshold': (0.0010148283164787323, (0.00095999590233126, 0.0010696607306262046)), 'GHZ_success_rates': {0.0007: 0.8878, 0.001: 0.8751}, 'number_of_points': 4, 'minimum_number_iterations': 100000}, 0.0454961: {'threshold': (0.00114768246047089, (0.0010932997147837307, 0.0012020652061580494)), 'GHZ_success_rates': {0.0007: 0.9154, 0.0011: 0.8981}, 'number_of_points': 4, 'minimum_number_iterations': 100000}, 0.0479281: {'threshold': (0.0012670559443590688, (0.0012438217803170924, 0.0012902901084010452)), 'GHZ_success_rates': {0.00095: 0.926, 0.00125: 0.9146}, 'number_of_points': 6, 'minimum_number_iterations': 100000}, 0.0503601: {'threshold': (0.0013395041046610496, (0.0013124187984340804, 0.0013665894108880188)), 'GHZ_success_rates': {0.0012: 0.9332, 0.0013: 0.9297}, 'number_of_points': 4, 'minimum_number_iterations': 100000}, 0.0527921: {'threshold': (0.0013896077604916078, (0.0013603350845382816, 0.001418880436444934)), 'GHZ_success_rates': {0.0012: 0.9481, 0.0013: 0.9447}, 'number_of_points': 4, 'minimum_number_iterations': 100000}, 0.0552241: {'threshold': (0.0014224740714775506, (0.0013898212734623455, 0.0014551268694927557)), 'GHZ_success_rates': {0.0012: 0.9587, 0.0014: 0.9532}, 'number_of_points': 3, 'minimum_number_iterations': 100000}, 0.0576561: {'threshold': (0.0014623763764196826, (0.0014340366307544021, 0.001490716122084963)), 'GHZ_success_rates': {0.0012: 0.9672, 0.0014: 0.9624}, 'number_of_points': 4, 'minimum_number_iterations': 100000}, 0.0613041: {'threshold': (0.0014858751835952135, (0.0014465668291962937, 0.0015251835379941332)), 'GHZ_success_rates': {0.0012: 0.9772, 0.00145: 0.9729}, 'number_of_points': 4, 'minimum_number_iterations': 100000}, 0.0649521: {'threshold': (0.0015135593393655416, (0.0014776283811759926, 0.0015494902975550905)), 'GHZ_success_rates': {0.0012: 0.9833, 0.0015: 0.9798}, 'number_of_points': 3, 'minimum_number_iterations': 100000}, 0.0673841: {'threshold': (0.0015153780040470964, (0.0014755253051425572, 0.0015552307029516356)), 'GHZ_success_rates': {0.0012: 0.9873, 0.0015: 0.984}, 'number_of_points': 4, 'minimum_number_iterations': 100000}, 0.0698161: {'threshold': (0.0014950800453340108, (0.001465063963889723, 0.0015250961267782987)), 'GHZ_success_rates': {0.0007: 0.994, 0.001475: 0.9875}, 'number_of_points': 8, 'minimum_number_iterations': 100000}, 0.0722481: {'threshold': (0.0014948764852672524, (0.0014365904675488128, 0.001553162502985692)), 'GHZ_success_rates': {0.0012: 0.9924, 0.0015: 0.9904}, 'number_of_points': 3, 'minimum_number_iterations': 100000}, 0.0795441: {'threshold': (0.0014573395450224401, (0.0014167691077747772, 0.001497909982270103)), 'GHZ_success_rates': {0.0005: 0.9988, 0.00145: 0.9954}, 'number_of_points': 7, 'minimum_number_iterations': 100000}, 0.0831921: {'threshold': (0.0014407461495117146, (0.001406411857029798, 0.001475080441993631)), 'GHZ_success_rates': {0.0005: 0.9994, 0.00135: 0.9972}, 'number_of_points': 5, 'minimum_number_iterations': 100000}, 0.1014321: {'threshold': (0.0012762507875131686, (0.0012270484472763292, 0.0013254531277500081)), 'GHZ_success_rates': {0.0003: 1.0, 0.00125: 0.9995}, 'number_of_points': 6, 'minimum_number_iterations': 100000}, 0.1038641: {'threshold': (0.001244166745910367, (0.0011892505821870685, 0.0012990829096336655)), 'GHZ_success_rates': {0.0001: 1.0, 0.0012: 0.9998}, 'number_of_points': 7, 'minimum_number_iterations': 100000}}, 2: {0.0758961: {'threshold': (0.0007177424993600469, (0.0006614962231737328, 0.0007739887755463609)), 'GHZ_success_rates': {0.0003: 0.8606, 0.00065: 0.8377}, 'number_of_points': 5, 'minimum_number_iterations': 100000}, 0.0783281: {'threshold': (0.0008113961700268459, (0.00076875218638052, 0.0008540401536731719)), 'GHZ_success_rates': {0.00045: 0.8743, 0.0007: 0.86}, 'number_of_points': 4, 'minimum_number_iterations': 100000}, 0.08117396: {'threshold': (0.0008378182621479549, (0.0007896694552839515, 0.0008859670690119582)), 'GHZ_success_rates': {0.00065: 0.8731, 0.0008: 0.8664}, 'number_of_points': 5, 'minimum_number_iterations': 100000}, 0.0844081: {'threshold': (0.000909616362038195, (0.0008609211587329378, 0.0009583115653434523)), 'GHZ_success_rates': {0.00065: 0.8952, 0.0009: 0.8834}, 'number_of_points': 6, 'minimum_number_iterations': 100000}, 0.0868401: {'threshold': (0.0010196094549152694, (0.0009710666157393488, 0.0010681522940911901)), 'GHZ_success_rates': {0.0001: 0.9477, 0.00095: 0.9076}, 'number_of_points': 5, 'minimum_number_iterations': 100000}, 0.09411158: {'threshold': (0.0010803299491927067, (0.0010389902278407639, 0.0011216696705446496)), 'GHZ_success_rates': {0.0001: 0.9674, 0.0011: 0.9258}, 'number_of_points': 4, 'minimum_number_iterations': 100000}, 0.10038958: {'threshold': (0.001118947051354998, (0.001076611782363526, 0.00116128232034647)), 'GHZ_success_rates': {0.0001: 0.9741, 0.0011: 0.9413}, 'number_of_points': 5, 'minimum_number_iterations': 100000}, 0.1099441: {'threshold': (0.0010886735198406845, (0.0010391673990460214, 0.0011381796406353475)), 'GHZ_success_rates': {0.0001: 0.9783, 0.00105: 0.9528}, 'number_of_points': 4, 'minimum_number_iterations': 100000}, 0.1148081: {'threshold': (0.0011117448700772503, (0.0010442818361102002, 0.0011792079040443004)), 'GHZ_success_rates': {0.0001: 0.9832, 0.0011: 0.962}, 'number_of_points': 6, 'minimum_number_iterations': 100000}, 0.12496706: {'threshold': (0.001065966677647953, (0.0009950482344146243, 0.0011368851208812816)), 'GHZ_success_rates': {0.0005: 0.9817, 0.00105: 0.9722}, 'number_of_points': 5, 'minimum_number_iterations': 100000}, 0.13223344: {'threshold': (0.001071989628059657, (0.0010104203785211853, 0.0011335588775981287)), 'GHZ_success_rates': {0.0005: 0.9879, 0.0011: 0.981}, 'number_of_points': 5, 'minimum_number_iterations': 100000}, 0.1395961: {'threshold': (0.0010237078333689327, (0.0009555676550909709, 0.0010918480116468945)), 'GHZ_success_rates': {0.0005: 0.9919, 0.001: 0.9869}, 'number_of_points': 6, 'minimum_number_iterations': 100000}, 0.14804948: {'threshold': (0.0009839475753385557, (0.0009245182445141664, 0.001043376906162945)), 'GHZ_success_rates': {0.0005: 0.9957, 0.001: 0.9915}, 'number_of_points': 6, 'minimum_number_iterations': 100000}, 0.1518883: {'threshold': (0.0009424533261367481, (0.0008854221294227116, 0.0009994845228507847)), 'GHZ_success_rates': {0.0005: 0.9969, 0.001: 0.9931}, 'number_of_points': 7, 'minimum_number_iterations': 100000}, 0.15934372: {'threshold': (0.0008905216468922002, (0.0008243140156462037, 0.0009567292781381967)), 'GHZ_success_rates': {0.0005: 0.9978, 0.0009: 0.9949}, 'number_of_points': 7, 'minimum_number_iterations': 50000}, 0.18245158: {'threshold': (0.000723810368547834, (0.0006635175445174944, 0.0007841031925781736)), 'GHZ_success_rates': {0.0003: 0.9997, 0.0007: 0.9989}, 'number_of_points': 6, 'minimum_number_iterations': 50000}, 0.21783452: {'threshold': (0.00040849241254773354, (0.0003588152684849564, 0.0004581695566105107)), 'GHZ_success_rates': {0.0002: 1.0, 0.0004: 1.0}, 'number_of_points': 6, 'minimum_number_iterations': 100000}, 0.23440386: {'threshold': (0.00022787301894340773, (0.00016401392328445151, 0.00029173211460236395)), 'GHZ_success_rates': {1e-05: 1.0, 0.00025: 1.0}, 'number_of_points': 6, 'minimum_number_iterations': 100000}, 0.26349148: {'threshold': (1.0000000000000002e-06, (-6.558492753819893e-05, 6.758492753819893e-05)), 'GHZ_success_rates': {1e-06: 1.0}, 'number_of_points': 9, 'minimum_number_iterations': 100000}}, 3: {0.0345521: {'threshold': (0.0001574728196519305, (0.0001177159925421144, 0.0001972296467617466)), 'GHZ_success_rates': {7.5e-05: 0.9133, 0.0002: 0.9111}, 'number_of_points': 11, 'minimum_number_iterations': 100000}, 0.0382001: {'threshold': (0.0003175921281470081, (0.0002960599592352736, 0.00033912429705874257)), 'GHZ_success_rates': {0.000125: 0.9424, 0.000315: 0.9396}, 'number_of_points': 23, 'minimum_number_iterations': 100000}, 0.0406321: {'threshold': (0.0003756529027387934, (0.0003392292078322504, 0.0004120765976453364)), 'GHZ_success_rates': {0.000225: 0.9556, 0.0004: 0.9538}, 'number_of_points': 21, 'minimum_number_iterations': 100000}, 0.0418481: {'threshold': (0.0004087867484949176, (0.00036941595694321277, 0.0004481575400466224)), 'GHZ_success_rates': {0.000275: 0.9611, 0.00041: 0.9601}, 'number_of_points': 17, 'minimum_number_iterations': 100000}, 0.0430641: {'threshold': (0.00043255006653365117, (0.00038997259324096675, 0.0004751275398263356)), 'GHZ_success_rates': {0.000275: 0.967, 0.000455: 0.9655}, 'number_of_points': 14, 'minimum_number_iterations': 100000}, 0.0467121: {'threshold': (0.00046000000000000007, (0.0003976710097634948, 0.0005223289902365054)), 'GHZ_success_rates': {0.000325: 0.9792, 0.00047: 0.978}, 'number_of_points': 16, 'minimum_number_iterations': 100000}, 0.0503601: {'threshold': (0.00046499999999999997, (0.0004266903826683576, 0.0005033096173316424)), 'GHZ_success_rates': {0.000375: 0.9868, 0.000495: 0.9863}, 'number_of_points': 16, 'minimum_number_iterations': 100000}, 0.0527921: {'threshold': (0.00046999999999999993, (0.00041208703430651846, 0.0005279129656934814)), 'GHZ_success_rates': {0.000375: 0.9907, 0.00049: 0.9903}, 'number_of_points': 16, 'minimum_number_iterations': 100000}, 0.0552241: {'threshold': (0.0004978567163333487, (0.0004547832934370526, 0.0005409301392296449)), 'GHZ_success_rates': {0.000375: 0.9932, 0.000505: 0.9928}, 'number_of_points': 17, 'minimum_number_iterations': 100000}, 0.0576561: {'threshold': (0.0004730951118354175, (0.00044426026267366063, 0.0005019299609971743)), 'GHZ_success_rates': {0.000375: 0.9949, 0.000495: 0.9947}, 'number_of_points': 22, 'minimum_number_iterations': 100000}, 0.0600881: {'threshold': (0.00046004323537756367, (0.0004264523202207015, 0.0004936341505344259)), 'GHZ_success_rates': {0.000375: 0.9963, 0.000485: 0.9962}, 'number_of_points': 18, 'minimum_number_iterations': 100000}, 0.0649521: {'threshold': (0.00043437266672449585, (0.00041065970285197287, 0.00045808563059701883)), 'GHZ_success_rates': {0.000375: 0.9983, 0.000445: 0.9983}, 'number_of_points': 12, 'minimum_number_iterations': 100000}, 0.0746801: {'threshold': (0.00033479404327924745, (0.00029640705383581287, 0.00037318103272268203)), 'GHZ_success_rates': {0.0003: 0.9997, 0.00036: 0.9997}, 'number_of_points': 15, 'minimum_number_iterations': 100000}, 0.0904881: {'threshold': (0.00018000000000000004, (0.00013879599830476504, 0.00022120400169523504)), 'GHZ_success_rates': {0.00018: 1.0, 0.000215: 1.0}, 'number_of_points': 17, 'minimum_number_iterations': 100000}, 0.0977841: {'threshold': (0.00014244045771621202, (0.00011007950078901964, 0.0001748014146434044)), 'GHZ_success_rates': {0.0001: 1.0, 0.000175: 1.0}, 'number_of_points': 11, 'minimum_number_iterations': 100000}}}


        # For plot that only shows the data points of the best protocols:
        if only_show_best_new_protocols_in_plot and protocols_to_be_plotted_per_set is not None:
            protocols_to_be_plotted_per_set_new = {}
            for set in protocols_to_be_plotted_per_set.keys():
                if set != "Set3p":
                    protocols_to_be_plotted_per_set_new[set] = []
                    old_prots = []
                    best_new_prot = None
                    best_new_prot_fid = 0
                    best_old_prot = None
                    best_old_prot_fid = 0
                    for protocol in protocols_to_be_plotted_per_set[set]:
                        if protocol[0][:3] == "sim":
                            if protocol[2][0] > best_new_prot_fid:
                                best_new_prot = ('Best protocol GHZ optimization', None, protocol[2])
                                best_new_prot_fid = protocol[2][0]
                        elif protocol[0] not in ["plain", "minimum4x_22", "minimum4x_40"]:
                            best_old_prot = True
                            old_prots.append(protocol)
                        # else:
                        #     if protocol[2][0] > best_old_prot_fid:
                        #         best_old_prot = ('Best previously known protocol', None, protocol[2])
                        #         best_old_prot_fid = protocol[2][0]
                    if best_new_prot is not None:
                        protocols_to_be_plotted_per_set_new[set].append(best_new_prot)
                    if best_old_prot is not None:
                        for old_prot in old_prots:
                            protocols_to_be_plotted_per_set_new[set].append(old_prot)
            for set in protocols_to_be_plotted_per_set_new.keys():
                print(set)
                for protocol in protocols_to_be_plotted_per_set_new[set]:
                    print(protocol[0])
                print("")
            protocols_to_be_plotted_per_set = protocols_to_be_plotted_per_set_new

        protocol_information = ep.find_protocol_information(update_protocol_information)

        best_prots_found_per_set = {"Set3r": ["simv1_sIIIr_4_6_1", "simv1_sIIIr_4_6_13", "simv1_sIIIr_4_6_2",
                                              "simv1_sIIIr_4_6_12"],
                                    "Set3d": ["simv4_sIIId_4_6_17", "simv4_sIIId_4_6_9", "simv4_sIIId_4_6_1",
                                              "simv4_sIIId_4_6_5"],
                                    "Set3q": ["simv1_sIIIq_4_6_9", "simv1_sIIIq_4_6_10", "simv1_sIIIq_4_6_1"],
                                    "Set3c": ["simv3_sIIIc_4_7_1", "simv3_sIIIc_4_7_6", "simv4_sIIIc_4_7_14"],
                                    "Set3e": ["simv4_sIIIe_4_7_1", "simv4_sIIIe_4_7_3"],
                                    "Set3f": ["simv5_sIIIf_4_11_14", "simv5_sIIIf_4_11_2"],
                                    "Set5r": ["simv1_sVr_4_12_9", "simv1_sVr_4_12_1", "simv1_sVr_4_13_9"],
                                    "Set5q": ["simv1_sVq_4_12_9", "simv1_sVq_4_11_1", "simv1_sVq_4_12_1"],
                                    "Set5p": ["simv1_sVp_4_11_18", "simv1_sVp_4_10_2", "simv2_sVp_4_12_1",
                                              "simv2_sVp_4_14_1", "simv2_sVp_4_15_1"],
                                    "Set5a": ["simv1_sVa_4_10_10", "simv1_sVa_4_11_5", "simv1_sVa_4_9_17",
                                              "simv1_sVa_4_9_16"],
                                    "Set5b": ["simv1_sVb_4_9_1", "simv1_sVb_4_10_1"],
                                    "Set5c": ["simv1_sVc_4_11_5", "simv1_sVc_4_11_8", "simv3_sVc_4_11_7",
                                              "simv3_sVc_4_10_4"],
                                    "Set5d": ["simv1_sVd_4_11_1", "simv1_sVd_4_11_2"],
                                    "Set5e": ["simv2_sVe_4_7_7", "simv2_sVe_4_7_9", "simv2_sVe_4_7_4"],
                                    "Set6k": ["simv1_sVIk_4_6_9", "simv1_sVIk_4_6_1"],
                                    "Set6g": ["simv1_sVIg_4_7_9"],
                                    "Set6m": ["simv2_sVIm_4_7_6", "simv2_sVIm_4_7_9"],
                                    "Set6h": ["simv1_sVIh_4_7_2", "simv2_sVIh_4_7_3", "simv1_sVIh_4_7_1",
                                              "simv2_sVIh_4_7_2"]}

        best_prots_found = []
        best_prots_found_Set3and5 = []
        for set_name_key in best_prots_found_per_set.keys():
            if set_name_key[:4] in ["Set3", "Set5", "Set6"]:
                for protocol_name in best_prots_found_per_set[set_name_key]:
                    best_prots_found.append(protocol_name)
                    if set_name_key[:4] in ["Set3", "Set5"]:
                        best_prots_found_Set3and5.append(protocol_name)
        print(best_prots_found)
        print("")

        for protocol_name in best_prots_found:
            identical_protocols = protocol_information[protocol_name]["identical"]
            for identical_protocol in identical_protocols:
                if identical_protocol in best_prots_found:
                    print(f"Protocol {protocol_name} and {identical_protocol} are identical.")
        print("")

        # for protocol_name in best_prots_found:
        #     similar_protocols = protocol_information[protocol_name]["similar"]
        #     sim_prots = f"Protocol {protocol_name} and "
        #     for similar_protocol in similar_protocols:
        #         if similar_protocol in best_prots_found:
        #             sim_prots += f"{similar_protocol}, "
        #             # print(f"Protocol {protocol_name} and {similar_protocol} are similar.")
        #     print(sim_prots + " are similar")
        # print("")

        # for protocol_name in best_prots_found_Set3and5:
        #     similar_best_prots = []
        #     for similar_protocol in protocol_information[protocol_name]["similar"]:
        #         if similar_protocol in best_prots_found:
        #             similar_best_prots.append(similar_protocol)
        #     print(f"Protocol {protocol_name} has similar in list of best prots: {similar_best_prots}.")
        # print("")

        all_prots_found_set_five = []
        for protocol_name in protocol_information.keys():
            if protocol_name[6:8] == "sV" and protocol_name[6:9] != "sVI":
                all_prots_found_set_five.append(protocol_name)

        convert_name = {'simv3_sIIIc_4_7_1': 'Septimum ($k=7$)', 'basic': 'Basic ($k=8$)', 'minimum4x_22': 'Minimum4x22 ($k=22$)',
                        'medium': 'Medium ($k=16$)', 'expedient': 'Expedient ($k=22$)', 'refined': 'Refined ($k=40$)',
                        'refined2': 'Refined ($k=40$)', 'minimum4x_40': 'Minimum4x40 ($k=40$)',
                        'stringent': 'Stringent ($k=42$)', 'simv2_sVp_4_14_1': 'New protocol ($k=14$)',
                        'simv2_sVp_4_15_1': 'New protocol ($k=15$)'}

        re_order_threshold_df_columns = 0
        def yield_settings(explore_new_protocols, set_name, set_for_which_prots_are_found, cut_off_dep_plot_data=None):
            if plot_link_efficiency_thresholds or plot_bell_quality_thresholds or plot_bell_succ_prob_thresholds \
                    or plot_cut_off_time_dependence_subplots:
                for round_yield in [1, 2]:
                    for explore_new_protocols_yield in [True, False]:
                        if round_yield == 2 and explore_new_protocols_yield is False:
                            continue
                        elif (round_yield == 2 or explore_new_protocols_yield is False) and plot_cut_off_time_dependence_subplots:
                            continue
                        else:
                            if plot_link_efficiency_thresholds:
                                associated_sets = ["Set3r", "Set3d", "Set3q", "Set3p", "Set3c", "Set3e", "Set3f", "Set3g", "Set3m", "Set3h", "Set3k"]
                                protocol_found_sets = ["sIIIr", "sIIId", "sIIIq", "sIIIp", "sIIIc", "sIIIe", "sIIIf", "sIIIg", "sIIIm", "sIIIh", "sIIIk"]
                            elif plot_bell_succ_prob_thresholds:
                                # associated_sets = ["Set6d", "Set6e", "Set6f", "Set6k", "Set6g", "Set6m", "Set6h", "Set3e"]
                                # protocol_found_sets = ["sVId", "sVIe", "sVIf", "sVIk", "sVIg", "sVIm", "sVIh", "sIIIe"]
                                associated_sets = ["Set6k", "Set6g", "Set6m", "Set6h", "Set3e"]
                                protocol_found_sets = ["sVIk", "sVIg", "sVIm", "sVIh", "sIIIe"]
                            elif plot_bell_quality_thresholds:
                                associated_sets = ["Set5r", "Set5q", "Set5p", "Set5a", "Set5b", "Set5c", "Set5d", "Set5e"]
                                protocol_found_sets = ["sVr", "sVq", "sVp", "sVa", "sVb", "sVc", "sVd", "sVe"]
                            # for set_name_yield in ["Set3d", "Set3c", "Set3e", "Set3f", "Set3g", "Set3m", "Set3h", "Set3k"]:
                            elif plot_cut_off_time_dependence_subplots:
                                associated_sets = ["Set3c", "Set3g", "Set3g", "Set5p"] #, "Set5p"]
                                protocol_found_sets = ["sIIIc", None, None, "sVp"] #, "sVp"]
                                cut_off_dep_plot_data_set = [("simv3_sIIIc_4_7_1", 0, r"\eta^{\ast}_\mathrm{link}=8\cdot10^2", "Fig. 7"),
                                                             ("expedient", 1, r"\eta^{\ast}_\mathrm{link}=2\cdot10^5", "Fig. 7"),
                                                             ("refined2", 2, r"\eta^{\ast}_\mathrm{link}=2\cdot10^5", "Fig. 7"),
                                                             ("simv2_sVp_4_14_1", 3, r"F_\mathrm{link}\approx0.828", "Fig. 6")] #,
                                                             # ("simv2_sVp_4_15_1", 4, r"F_\mathrm{link}\approx0.828", "Fig. 6")]
                                for combo in zip(associated_sets, protocol_found_sets, cut_off_dep_plot_data_set):
                                    set_name_yield = combo[0]
                                    set_for_which_prots_are_found_yield = combo[1]
                                    explore_new_protocols_yield_co = False if set_for_which_prots_are_found_yield is None else True
                                    cut_off_dep_plot_data_yield = combo[2]
                                    yield explore_new_protocols_yield_co, set_name_yield, set_for_which_prots_are_found_yield, round_yield, cut_off_dep_plot_data_yield
                                continue
                            for set_name_yield in associated_sets:
                                if explore_new_protocols_yield is True:
                                    for set_for_which_prots_are_found_yield in protocol_found_sets:
                                        yield explore_new_protocols_yield, set_name_yield, set_for_which_prots_are_found_yield, round_yield, cut_off_dep_plot_data
                                else:
                                    yield explore_new_protocols_yield, set_name_yield, set_for_which_prots_are_found, round_yield, cut_off_dep_plot_data
            else:
                if set_for_which_prots_are_found == "all_sV":
                    protocol_found_sets = ["sVr", "sVq", "sVp", "sVa", "sVb", "sVc", "sVd", "sVe"]
                else:
                    protocol_found_sets = [set_for_which_prots_are_found]
                round_yield = 1
                for set_prots_found in protocol_found_sets:
                    yield explore_new_protocols, set_name, set_prots_found, round_yield, cut_off_dep_plot_data

        def yield_protocols(n, set_name, explore_new_protocols, set_for_which_prots_are_found):
            if explore_new_protocols:
                for version in ['simv1', 'simv2', 'simv3', 'simv4', 'simv5', 'simv6', 'simv7', 'simv8', 'simv8T']:
                    for k in range(5, max_prots_per_n + 1):
                        for b in range(1, max_prots_per_n_k + 1):
                            if 1 == 1:
                                prot_name_short = f"{version}_{set_for_which_prots_are_found}_{n}_{k}_{b}"
                                prot_name = f'dyn_prot_{version}_{set_for_which_prots_are_found}_{n}_{k}_{b}'
                                file_name = "recipe_" + prot_name + "_swap_" + set_name + "_toric_graph_3D_" + decoder + "_thres.csv"
                                yield prot_name_short, prot_name, file_name
            else:
                for prot in ['plain', 'modicum', 'basic1', 'basic2', 'medium1', 'medium2', 'minimum4x_22', 'expedient',
                             'refined1', 'refined2', 'minimum4x_40_1', 'minimum4x_40_2', 'stringent']:
                # for prot in ['plain', 'modicum', 'basic1', 'basic2', 'medium1', 'medium2', 'expedient', 'refined1',
                #              'refined2', 'stringent', 'direct_ghz']:
                # for prot in ['basic1', 'medium2', 'minimum4x_22', 'expedient', 'refined1', 'minimum4x_40_1', 'stringent']:
                    prot_name_short = prot
                    prot_name = "hc_" + prot
                    file_name = prot + "_swap_" + set_name + "_toric_graph_3D_" + decoder + "_thres.csv"
                    yield prot_name_short, prot_name, file_name

        new_protocol_settings = {}
        protocols_calculated_per_set = defaultdict(list)
        best_new_prot_per_set = {}
        cut_off_dict_set = {}
        cut_off_information = {}

        plot_fonts(style=plot_style)
        i_output_fits = 0

        previous_set_name = None
        first_set = True
        for explore_new_protocols_yield, set_name_yield, set_for_which_prots_are_found_yield, round_yield, \
            cut_off_dep_plot_data_yield in yield_settings(explore_new_protocols, set_name, set_for_which_prots_are_found):

            if (plot_link_efficiency_thresholds
                        or plot_bell_succ_prob_thresholds
                        or plot_bell_quality_thresholds
                        or plot_cut_off_time_dependence_subplots) \
                    and (protocols_to_be_plotted_per_set is not None
                        or cut_off_dict_set_plot is not None):
                continue

            if first_set is False:
                print("\n\n\n\n\n\n")
            first_set = False
            if explore_new_protocols_yield:
                print_type_prots = f"new protocols found with set {set_for_which_prots_are_found_yield}"
            else:
                print_type_prots = "known protocols"
            print(f"{colored(f'Calculating {print_type_prots} in set {set_name_yield}:', 'green', 'on_grey', attrs=['bold'])}")

            if round_yield == 2:
                best_new_prot_list = [prot[0] for prot in best_new_prot_per_set.values()]
            else:
                best_new_prot_list = []
            print(colored(f"best_new_prot_list: {best_new_prot_list}", "yellow"))
            print(colored(f"best_new_prot_per_set: {best_new_prot_per_set}", "yellow"))

            if set_name_yield != previous_set_name:
                best_thresholds_per_set = {}

                columns = ["protocol", "cut_off_time", "est_GHZ_success_rate", "p_g", "GHZ_success_rate",
                           "below_threshold", "possibly_below_threshold", "possibly_above_threshold",
                           "order", "success_rates", "number_of_iterations"]

                re_order_threshold_df_columns = 0
                if include_superoperator_statistics_decoherence:
                    re_order_threshold_df_columns += 7
                    columns += ["avg_prot_time", "link_gen_perc", "dec_T1n_link", "dec_T1n_idle", "dec_T2n_link", "dec_T2n_idle", "n_DD"]
                    # re_order_threshold_df_columns += 5
                    # columns += ["link_gen_perc", "dec_T1n_link", "dec_T1n_idle", "dec_T2n_link", "dec_T2n_idle"]
                if include_superoperator_statistics_stab_fids:
                    re_order_threshold_df_columns += 2
                    columns += ["stab_fid_p", "stab_fid_s"]
                if include_superoperator_statistics_weighted_sum:
                    re_order_threshold_df_columns += 12
                    columns += ["weighted_sum", "success_rate_12", "success_rate_12_int", "success_rate_4",
                                "weighted_sum_p_x", "weighted_sum_p_z", "weighted_sum_p_y", "weighted_sum_p_meas",
                                "weighted_sum_s_x", "weighted_sum_s_z", "weighted_sum_s_y", "weighted_sum_s_meas"]
                threshold_df = pd.DataFrame(columns=columns)
                thresholds_dict = defaultdict(lambda: defaultdict(dict))
                thresholds_fit_dict = defaultdict(lambda: defaultdict(dict))

                if calculate_cut_off_times is False:
                    threshold_df.drop(columns="est_GHZ_success_rate", inplace=True)
                    threshold_df = threshold_df.set_index(['protocol', 'cut_off_time', 'p_g'])
                else:
                    threshold_df = threshold_df.set_index(['protocol', 'cut_off_time', 'est_GHZ_success_rate', 'p_g'])

            previous_set_name = set_name_yield

            folder = "C:\\Users\\sebastiandebon\\surfdrive\\Code\\Sebastian\\oopsc\\superoperator\\"
            folder = folder + "Old\\Wrong_network_architecture\\" if review_old_results is True else folder
            if explore_new_protocols_yield:
                if plot_cut_off_time_add_cut_off_to_location:
                    folder_stats = f"{folder}Thresholds_best_prots_{set_for_which_prots_are_found_yield}_vary_cut_off{'_nDD_' + str(nDD) if nDD is not None else ''}\\"
                    folder_rel = f"Thresholds_best_prots_{set_for_which_prots_are_found_yield}_vary_cut_off{'_nDD_' + str(nDD) if nDD is not None else ''}\\threshold_sim\\"
                elif set_for_which_prots_are_found == "all_sV":
                    folder_stats = f"{folder}Thresholds_best_prots_{set_for_which_prots_are_found}_in_{set_name_yield}\\"
                    folder_rel = f"Thresholds_best_prots_{set_for_which_prots_are_found}_in_{set_name_yield}\\threshold_sim\\"
                else:
                    folder_stats = f"{folder}Thresholds_best_prots_{set_for_which_prots_are_found_yield}{'_nDD_' + str(nDD) if nDD is not None else ''}\\"
                    folder_rel = f"Thresholds_best_prots_{set_for_which_prots_are_found_yield}{'_nDD_' + str(nDD) if nDD is not None else ''}\\threshold_sim\\"
                # folder_stats = f"{folder}Old\\Compare_different_network_architectures_qsurface\\Thresholds_best_prots_sIIIc_new_fit_mi\\"
                # folder_rel = f"Old\\Compare_different_network_architectures_qsurface\\Thresholds_best_prots_sIIIc_new_fit_mi\\threshold_sim\\"
                # folder_stats = f"{folder}Paper_test_comparisons\\Test_vary_nDD\\Thresholds_test_{set_name}_nDD_{nDD}\\"
                # folder_rel = f"Paper_test_comparisons\\Test_vary_nDD\\Thresholds_test_{set_name}_nDD_{nDD}\\threshold_sim\\"
                # folder_stats = f"{folder}Paper_test_comparisons\\Test_isotopically_purified_set10n_idle_coh_times\\Thresholds_best_prots_{set_for_which_prots_are_found_yield}_improved_isotopically_purified_T2_50\\"
                # folder_rel = f"Paper_test_comparisons\\Test_isotopically_purified_set10n_idle_coh_times\\Thresholds_best_prots_{set_for_which_prots_are_found_yield}_improved_isotopically_purified_T2_50\\threshold_sim\\"
                # folder_stats = f"{folder}Paper_test_comparisons\\Thresholds_best_prots_{set_for_which_prots_are_found_yield}_eta_link_test\\"
                # folder_rel = f"Paper_test_comparisons\\Thresholds_best_prots_{set_for_which_prots_are_found_yield}_eta_link_test\\threshold_sim\\"

            elif review_old_results is False:
                if plot_cut_off_time_add_cut_off_to_location:
                    folder_stats = f"{folder}{folder_known_protocols}_in_{set_name_yield}_vary_cut_off\\"
                    folder_rel = f"{folder_known_protocols}_in_{set_name_yield}_vary_cut_off\\threshold_sim\\"
                else:
                    folder_stats = f"{folder}{folder_known_protocols}_in_{set_name_yield}\\"
                    folder_rel = f"{folder_known_protocols}_in_{set_name_yield}\\threshold_sim\\"
                # folder_stats = f"{folder}{folder_known_protocols}\\"
                # folder_rel = f"{folder_known_protocols}\\threshold_sim\\"
                # folder_rel = f"{csv_folder_name_new_prots}\\threshold_sim\\"
                # folder_stats = f"{folder}{csv_folder_name_new_prots}\\"
            else:
                folder_stats = f"{folder}{folder_known_protocols}\\"
                folder_rel = f"{folder_known_protocols}\\threshold_sim\\"

            if p_g is None:
                p_g_v = [0, 1]
            elif not isinstance(p_g, list):
                p_g_v = [p_g, p_g]
            else:
                p_g_v = p_g

            for prot_name_short, prot_name, file_name in yield_protocols(n, set_name_yield, explore_new_protocols_yield,
                                                                         set_for_which_prots_are_found_yield):
                # print(prot_name_short, prot_name, file_name, folder + folder_rel)
                if update_csv_files:
                    update_result_files_sc(folder + folder_rel, file_name.split(".csv")[0])

                file_loc = folder + folder_rel + file_name
                try:
                    threshold_info = ct.calculate_threshold_information(file_loc,
                                                                        return_information_per_data_point=True)
                except FileNotFoundError:
                    continue

                if calculate_cut_off_times:
                    for error_prob in [1e-12, 1e-6, 0.0006, 0.00077]:
                        if "s" + set_name[3:] in ["s3h", "s3k"]:
                            file_stats = prot_name + "_node-s" + set_name[3:] + f"_p_g-{error_prob}_p_m-{error_prob}_no-dec_no-prob"
                        else:
                            file_stats = prot_name + "_node-s" + set_name[3:] + f"_p_g-{error_prob}_p_m-{error_prob}"
                        if update_csv_files:
                            update_result_files_so(folder_stats, file_stats)
                        if os.path.exists(folder_stats + file_stats + ".pkl"):
                            cut_off_times_per_protocol_and_per_percentage_for_10m12[prot_name_short] = \
                                defaultdict(list)
                            # Construct cut-off times:
                            characteristics = pickle.load(open(folder_stats + file_stats + ".pkl", 'rb'))
                            for percentage in GHZ_success_percentages:
                                conf_int = asd.confidence_interval(characteristics['dur'],
                                                                   1 - 2 * (100 - percentage * 100) / 100,
                                                                   require_unique_min=True,
                                                                   return_new_bound=True)
                                cut_off_time = round(conf_int[1], ACCURACY)
                                cut_off_times_per_protocol_and_per_percentage_for_10m12[prot_name_short][
                                    cut_off_time].append((percentage, conf_int[3], error_prob))
                            eval_dict = asd.GHZ_success_rate_for_cut_off_time(characteristics['dur'])
                            for cut_off_time in threshold_info.keys():
                                if cut_off_time not in cut_off_times_per_protocol_and_per_percentage_for_10m12[prot_name_short]:
                                    # Choose the closest cut-off time in eval_dict that is smaller than or equal to "cut_off_time"
                                    cot_diff = [cut_off_time - cot_dict for cot_dict in eval_dict.keys()]
                                    cot_diff_pos = [d for d in cot_diff if d >= 0]
                                    if len(cot_diff_pos) == 0:
                                        GHZ_success_rate = 0
                                        cut_off_time_sel = 0
                                    else:
                                        index_to_select = cot_diff.index(min(cot_diff_pos))
                                        cut_off_time_sel = list(eval_dict.keys())[index_to_select]
                                        GHZ_success_rate = eval_dict[cut_off_time_sel]["GHZ_success_rate"]*100
                                    cut_off_times_per_protocol_and_per_percentage_for_10m12[prot_name_short][cut_off_time].append((cut_off_time_sel, GHZ_success_rate, error_prob))
                            break

                p_g_value_in_threshold_info = False
                for cut_off_time in list(threshold_info.keys()):
                    for p_g_value in list(threshold_info[cut_off_time].keys()):
                        if p_g_value < p_g_v[0] or p_g_value > p_g_v[1]:
                            del threshold_info[cut_off_time][p_g_value]
                            if not threshold_info[cut_off_time]:
                                del threshold_info[cut_off_time]
                        else:
                            p_g_value_in_threshold_info = True
                if p_g_value_in_threshold_info:

                    # from fit_threshold import main as fit_Yves
                    #
                    # mean = pd.DataFrame()
                    # mean.index.name = 'p'
                    # mean.columns.name = 'r'
                    # coef = pd.DataFrame()
                    # coef.index.name = 'p'
                    # coef.columns.name = 'r'
                    #
                    # for cut_off_time in threshold_info.keys():
                    #     if prot_name_short == "simv3_sIIIc_4_7_1" and cut_off_time == 0.0226081:
                    #         for p_g_value in threshold_info[cut_off_time].keys():
                    #             for lattice_size in sorted(threshold_info[cut_off_time][p_g_value]["Success_rates"].keys()):
                    #                 if lattice_size != 4:
                    #                     rate = threshold_info[cut_off_time][p_g_value]["Success_rates"][lattice_size]
                    #                     mean.loc[p_g_value, lattice_size] = 1-rate[0]
                    #                     coef.loc[p_g_value, lattice_size] = rate[1]
                    #
                    #         print(mean.to_string())
                    #         print(coef.to_string())
                    #
                    #         fit_Yves(mean, coef)
                    #         exit(0)

                    cut_off_times = list(threshold_info.keys())
                    ghz_success_rates = {}
                    for cut_off_time in cut_off_times:
                        p_g_value = list(threshold_info[cut_off_time].keys())[0]
                        ghz_success_rates[threshold_info[cut_off_time][p_g_value]["GHZ_success"]] = cut_off_time
                    if target_ghz_success_rate is not None:
                        ghz_sr_diff = [ghz_success_rate - target_ghz_success_rate for ghz_success_rate in ghz_success_rates.keys()]
                        pos_differences = [d for d in ghz_sr_diff if d >= 0]
                        if len(pos_differences) == 0:
                            continue
                        index_to_select = ghz_sr_diff.index(min(pos_differences))
                        cut_off_times = [ghz_success_rates[list(ghz_success_rates.keys())[index_to_select]]]
                    for cut_off_time in cut_off_times:
                        for p_g_value in list(threshold_info[cut_off_time].keys()):
                            if calculate_cut_off_times:
                                if cut_off_time in cut_off_times_per_protocol_and_per_percentage_for_10m12[prot_name_short].keys():
                                    cot_info = cut_off_times_per_protocol_and_per_percentage_for_10m12[prot_name_short][cut_off_time][0]
                                    est_GHZ_rate = f"{round(cot_info[1]/100, 5)} ({cot_info[2]})"
                                else:
                                    est_GHZ_rate = "Unknown"
                                index_key = (prot_name_short, cut_off_time, est_GHZ_rate, p_g_value)
                            else:
                                index_key = (prot_name_short, cut_off_time, p_g_value)
                            threshold_df.sort_index(inplace=True)
                            threshold_df.loc[index_key, "below_threshold"] = threshold_info[cut_off_time][p_g_value]["Below_threshold"]
                            threshold_df.sort_index(inplace=True)
                            threshold_df.loc[index_key, "possibly_below_threshold"] = threshold_info[cut_off_time][p_g_value]["Might_be_below_threshold"]
                            threshold_df.sort_index(inplace=True)
                            threshold_df.loc[index_key, "possibly_above_threshold"] = threshold_info[cut_off_time][p_g_value]["Might_be_above_threshold"]
                            threshold_df.sort_index(inplace=True)
                            threshold_df.loc[index_key, "order"] = list(threshold_info[cut_off_time][p_g_value]["Success_rates"].keys())
                            threshold_df.sort_index(inplace=True)
                            threshold_df.loc[index_key, "success_rates"] = [round(r[0], 5) for r in list(threshold_info[cut_off_time][p_g_value]["Success_rates"].values())]
                            threshold_df.sort_index(inplace=True)
                            threshold_df.loc[index_key, "GHZ_success_rate"] = threshold_info[cut_off_time][p_g_value]["GHZ_success"]
                            threshold_df.sort_index(inplace=True)
                            if isinstance(threshold_info[cut_off_time][p_g_value]["Number_of_iterations"], dict):
                                threshold_df.loc[index_key, "number_of_iterations"] = list(threshold_info[cut_off_time][p_g_value]["Number_of_iterations"].values())
                            else:
                                threshold_df.loc[index_key, "number_of_iterations"] = threshold_info[cut_off_time][p_g_value]["Number_of_iterations"]
                            threshold_df.sort_index(inplace=True)

                            if include_superoperator_statistics_stab_fids \
                                    or include_superoperator_statistics_decoherence \
                                    or include_superoperator_statistics_weighted_sum:
                                max_tries = 2
                                found_supop = False
                                try_nr = 1
                                while found_supop is False and try_nr <= max_tries:
                                    if try_nr == 1:
                                        file_prefix = prot_name + "_node-s" + set_name[3:] + f"_p_g-{p_g_value}_p_m-{p_g_value}"
                                    elif nDD is not None:
                                        file_prefix = prot_name + "_node-s" + set_name[3:] + f"_n_DD-{float(nDD)}" + f"_p_g-{p_g_value}_p_m-{p_g_value}"
                                    else:
                                        found_supop = False
                                        break
                                    if "s" + set_name[3:] in ["s3h", "s3k"]:
                                        supop_file = file_prefix + "_no-dec_no-prob"
                                    elif cut_off_time != np.infty:
                                        supop_file = file_prefix + f"_cut_off_time-{cut_off_time}"
                                    else:
                                        supop_file = file_prefix
                                    if update_csv_files:
                                        update_result_files_so(folder_stats, supop_file)
                                    try:
                                        df_supop = pd.read_csv(folder_stats + supop_file + ".csv", sep=';', float_precision='round_trip')
                                        found_supop = True
                                    except FileNotFoundError:
                                        found_supop = False
                                    try_nr += 1

                                if found_supop:
                                    index = ['error_config', 'lie'] if 'error_idle' not in df_supop else ['error_stab', 'error_idle', 'lie']
                                    df_supop = df_supop.set_index(index)
                                    if include_superoperator_statistics_decoherence:
                                        average_prot_time = df_supop.loc[('IIII', False), 'avg_duration']
                                        total_link_attempts = df_supop.loc[('IIII', False), 'total_link_attempts']
                                        total_iterations = df_supop.loc[('IIII', False), 'written_to']
                                        link_attempt_time = df_supop.loc[('IIII', False), 't_link']
                                        T1n_link = df_supop.loc[('IIII', False), 'T1n_link']
                                        T1n_idle = df_supop.loc[('IIII', False), 'T1n_idle']
                                        T2n_link = df_supop.loc[('IIII', False), 'T2n_link']
                                        T2n_idle = df_supop.loc[('IIII', False), 'T2n_idle']
                                        nDD_data = df_supop.loc[('IIII', False), 'n_DD']
                                        average_link_generating_time = total_link_attempts / total_iterations / 2 * link_attempt_time
                                        # print(f'\nAverage protocol time        = {average_prot_time} seconds.')
                                        # print(f'Total amount of iterations   = {total_iterations}.')
                                        # print(f'Total link attempts          = {total_link_attempts}.')
                                        # print(f'Link generation attempt time = {link_attempt_time} seconds.')
                                        # print(f'Average time spend on links  = {average_link_generating_time} seconds.')
                                        threshold_df.loc[index_key, "avg_prot_time"] = average_prot_time
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "link_gen_perc"] = average_link_generating_time / average_prot_time * 100
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "dec_T1n_link"] = 1 - math.exp(-average_link_generating_time / T1n_link)
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "dec_T1n_idle"] = 1 - math.exp(-(average_prot_time - average_link_generating_time) / T1n_idle)
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "dec_T2n_link"] = 1 - math.exp(-average_link_generating_time / T2n_link)
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "dec_T2n_idle"] = 1 - math.exp(-(average_prot_time - average_link_generating_time) / T2n_idle)
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "n_DD"] = nDD_data
                                        threshold_df.sort_index(inplace=True)
                                    if include_superoperator_statistics_stab_fids:
                                        stab_fid_p = df_supop.iloc[0, df_supop.columns.get_loc('p')]
                                        stab_fid_s = df_supop.iloc[0, df_supop.columns.get_loc('s')]
                                        threshold_df.loc[index_key, "stab_fid_p"] = stab_fid_p
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "stab_fid_s"] = stab_fid_s
                                        threshold_df.sort_index(inplace=True)
                                    if include_superoperator_statistics_weighted_sum:
                                        superoperator = Superoperator(folder_stats + supop_file + ".csv")
                                        weighted_sum_info = calculate_weighted_sum(superoperator)
                                        threshold_df.loc[index_key, "weighted_sum"] = weighted_sum_info['s'][1] + weighted_sum_info['s'][2] + weighted_sum_info['s'][3] * 2 + weighted_sum_info['s'][4] / 2 + \
                                                                                      weighted_sum_info['p'][1] + weighted_sum_info['p'][2] + weighted_sum_info['p'][3] * 2 + weighted_sum_info['p'][4] / 2
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "success_rate_12"] = threshold_info[cut_off_time][p_g_value]["Success_rates"][12][0]
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "success_rate_12_int"] = threshold_info[cut_off_time][p_g_value]["Success_rates"][12][1]
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "success_rate_4"] = threshold_info[cut_off_time][p_g_value]["Success_rates"][4][0]
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "weighted_sum_p_x"] = weighted_sum_info['p'][1]
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "weighted_sum_p_z"] = weighted_sum_info['p'][2]
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "weighted_sum_p_y"] = weighted_sum_info['p'][3]
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "weighted_sum_p_meas"] = weighted_sum_info['p'][4]
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "weighted_sum_s_x"] = weighted_sum_info['s'][1]
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "weighted_sum_s_z"] = weighted_sum_info['s'][2]
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "weighted_sum_s_y"] = weighted_sum_info['s'][3]
                                        threshold_df.sort_index(inplace=True)
                                        threshold_df.loc[index_key, "weighted_sum_s_meas"] = weighted_sum_info['s'][4]
                                        threshold_df.sort_index(inplace=True)

                thresholds, GHZ_success_rates = ct.calculate_threshold_per_cut_off_time(threshold_info)
                for cut_off_time in thresholds.keys():
                    # Add thresholds calculated manually to the dictionary "threshold_dict":
                    thresholds_dict[prot_name_short][cut_off_time]["threshold"] = thresholds[cut_off_time]
                    thresholds_dict[prot_name_short][cut_off_time]["GHZ_success_rates"] = GHZ_success_rates[cut_off_time]
                    key_rate = min(GHZ_success_rates[cut_off_time].keys(), key=lambda x: abs(x - thresholds[cut_off_time][0]))
                    GHZ_rate_found = GHZ_success_rates[cut_off_time][key_rate] if key_rate in GHZ_success_rates[cut_off_time].keys() else 0
                    points_involved_in_threshold = [index_name[2]
                                                    for index_name in threshold_df.index
                                                    if (index_name[0] == prot_name_short
                                                        and index_name[1] == cut_off_time
                                                        and index_name[2] >= thresholds[cut_off_time][1][0]
                                                        and index_name[2] <= thresholds[cut_off_time][1][1])]
                    iterations_involved_in_threshold = [threshold_df.loc[index_name, "number_of_iterations"]
                                                        for index_name in threshold_df.index
                                                        if (index_name[0] == prot_name_short
                                                            and index_name[1] == cut_off_time
                                                            and index_name[2] in points_involved_in_threshold)]
                    thresholds_dict[prot_name_short][cut_off_time]["number_of_points"] = len(points_involved_in_threshold)
                    thresholds_dict[prot_name_short][cut_off_time]["minimum_number_iterations"] = min(iterations_involved_in_threshold) if len(iterations_involved_in_threshold) > 0 else 0
                    # thresholds_dict[prot_name_short][cut_off_time]["iterations"] =
                    # print(f"{prot_name_short}, {cut_off_time}: {thresholds[cut_off_time]}. GHZ success rates: {GHZ_success_rates[cut_off_time]}.")

                    # Add thresholds found with fitting to a different dictionary "threshold_fit_dict":
                    # Construct DataFrame required for the fit function to work:
                    fit_dataframe = pd.DataFrame(columns=["L", "p_g", "N", "success"])
                    fit_dataframe = fit_dataframe.set_index(["L", "p_g"])
                    for p_g_value in threshold_info[cut_off_time].keys():
                        for lattice_size in sorted(threshold_info[cut_off_time][p_g_value]["Success_rates"].keys()):
                            rate = threshold_info[cut_off_time][p_g_value]["Success_rates"][lattice_size]
                            iter = threshold_info[cut_off_time][p_g_value]["Number_of_iterations"]
                            if isinstance(iter, dict):
                                iter = iter[lattice_size]
                            fit_dataframe.loc[(lattice_size, p_g_value), "success"] = rate[0]*iter
                            fit_dataframe.loc[(lattice_size, p_g_value), "N"] = iter
                    try:
                        _, par, perr, chisqred = zoomed_fit_Mark(fit_dataframe, modified_ansatz=True,
                                                                 print_results=False, limit_bounds=True,
                                                                 return_chi_squared_red=True,
                                                                 zoomed_in_fit=zoomed_in_fit, zoom_number=zoom_number)
                    except Exception:
                        try:
                            _, par, perr, chisqred = zoomed_fit_Mark(fit_dataframe, modified_ansatz=True,
                                                                     print_results=False, limit_bounds=False,
                                                                     return_chi_squared_red=True,
                                                                     zoomed_in_fit=zoomed_in_fit,
                                                                     zoom_number=zoom_number)
                        except Exception:
                            par, perr, chisqred = [0], 0, 0
                    degrees_freedom = len(fit_dataframe.index) - len(par)
                    tci = cci.find_t(p=0.95, df=degrees_freedom)[1]
                    if show_fits or output_fits:
                        show_fit_itself = True if par[0] != 0 else False
                        output_fits_i = output_fits +  f"_{i_output_fits}" if output_fits else ""
                        font_size_big_fit = font_size_big_start if output_fits else 20
                        font_size_small_fit = font_size_small_start if output_fits else 20*0.8
                        prot_name_plot = prot_name_short.replace('_', '-') if prot_name_short != "simv3_sIIIc_4_7_1" \
                            else r"Septimum ($k=7$)"
                        plot_fit_Mark(modified_ansatz=True, data=fit_dataframe, interactive_plot=interactive_plot,
                                      plot_title=f"Threshold for {prot_name_plot}" + "\n"
                                                 r"at $\eta^{\ast}_\mathrm{link}=2\cdot10^3$ in Fig. 7, " +
                                                 r"$t_\mathrm{GHZ}\approx $" +
                                                 f"{round(cut_off_time,3)} s" +
                                                 r" ($p_\mathrm{GHZ}\approx $" +
                                                 f"{round(GHZ_rate_found,3)*100}%)",
                                      par=par, perr=perr*tci,
                                      accuracy=None, include_fit=show_fit_itself, include_rescaled_error_rates=False,
                                      chi_squared_red=chisqred, plotstyle=plot_style,
                                      font_size_big=font_size_big_fit, font_size_small=font_size_small_fit,
                                      number_scaling_tries=number_scaling_tries,
                                      output=output_fits_i, show_plot=show_fits)
                        i_output_fits += 1
                    thresholds_fit_dict[prot_name_short][cut_off_time]["threshold"] = par[0], (par[0] - perr*tci, par[0] + perr*tci)
                    thresholds_fit_dict[prot_name_short][cut_off_time]["GHZ_success_rates"] = GHZ_success_rates[cut_off_time]
                    thresholds_fit_dict[prot_name_short][cut_off_time]["number_of_points"] = len(points_involved_in_threshold)
                    thresholds_fit_dict[prot_name_short][cut_off_time]["minimum_number_iterations"] = min(iterations_involved_in_threshold) if len(iterations_involved_in_threshold) > 0 else 0



            # Only keep the protocols that were below threshold:
            threshold_df_filtered = threshold_df # threshold_df[threshold_df['possibly_below_threshold'].isin([True])]

            for index_name in threshold_df_filtered.index:
                # if index_name[0] != "4_7_1":
                #     new_protocol_settings.append(f"(dyn_prot_simv3_sIIIc_{index_name[0]}, {index_name[1]}, [{new_p_g}])")
                p_name = f"dyn_prot_{index_name[0]}" if explore_new_protocols_yield is True else f"{index_name[0]}"
                # if p_name in new_protocol_settings.keys():
                #     continue
                # elif not (index_name[0][-1] == "1" and index_name[0][-2] == "_"):
                #     continue
                if p_name not in new_protocol_settings.keys():
                    new_protocol_settings[p_name] = {}
                use_range_p_g = new_p_g
                inc = new_p_g_increment
                if index_name[0] in thresholds_fit_dict.keys():
                    if index_name[1] in thresholds_fit_dict[index_name[0]].keys():
                        th_p_name = math.floor(thresholds_fit_dict[index_name[0]][index_name[1]]["threshold"][0]*1e4)/1e4
                        use_range_p_g = [round(th_p_name - 2*inc + inc*x, 5) for x in range(6)]
                new_protocol_settings[p_name][index_name[1]] = use_range_p_g
                # for percentage in cut_off_times_per_protocol_and_per_percentage_for_10m12[index_name[0]].keys():
                #     cut_off_time = cut_off_times_per_protocol_and_per_percentage_for_10m12[index_name[0]][percentage][0]
                #     new_protocol_settings[f"dyn_prot_{index_name[0]}"][cut_off_time] = new_p_g

            if not (plot_link_efficiency_thresholds or plot_bell_succ_prob_thresholds or plot_bell_quality_thresholds):
                print(f"\n{colored('All surface code data:', 'grey', 'on_white', attrs=['bold'])}")
                df_columns = threshold_df.columns.tolist()
                threshold_df = threshold_df[df_columns[(-1*re_order_threshold_df_columns):] + df_columns[:(-1*re_order_threshold_df_columns)]]
                if output_threshold_success_rates_dataframe:
                    threshold_df.to_csv("C:\\Users\\sebastiandebon\\surfdrive\\Code\\threshold_df.csv")
                print(threshold_df.to_string(formatters={'__index__': lambda x: str(x) if isinstance(x, float) else x}))

            if calculate_cut_off_times:
                if plot_link_efficiency_thresholds is False:
                    print(f"\n\n{colored('Cut-off times per GHZ success percentage:', 'grey', 'on_white', attrs=['bold'])}")
                for prot in cut_off_times_per_protocol_and_per_percentage_for_10m12.keys():
                    for percentage in cut_off_times_per_protocol_and_per_percentage_for_10m12[prot].keys():
                        print(f"{prot}, {percentage}, {cut_off_times_per_protocol_and_per_percentage_for_10m12[prot][percentage]}.")

            for dict_used in [thresholds_fit_dict]:     # [thresholds_dict, thresholds_fit_dict]:
                if dict_used:
                    print(f"\n\n{colored('Thresholds found:', 'grey', 'on_white', attrs=['bold'])}") # bold, dark, underline, blink, reverse, concealed
                    for prot in dict_used.keys():
                        identical_prots_in_best_prots = []
                        similar_prots_in_best_prots = []
                        try:
                            identical_protocols = protocol_information[prot]["identical"]
                            similar_protocols = protocol_information[prot]["similar"]
                            for identical_protocol in identical_protocols:
                                if identical_protocol in best_prots_found:
                                    identical_prots_in_best_prots.append(identical_protocol)
                            for similar_protocol in similar_protocols:
                                if similar_protocol in best_prots_found:
                                    similar_prots_in_best_prots.append(similar_protocol)
                        except:
                            pass
                        dict_used[prot] = dict(sorted(dict_used[prot].items(), key=lambda item: item[0]))
                        for cut_off_time in dict_used[prot].keys():
                            threshold = dict_used[prot][cut_off_time]['threshold']
                            GHZ_rates = dict_used[prot][cut_off_time]['GHZ_success_rates']
                            key_to_take = min(GHZ_rates.keys(), key=lambda x: abs(x - threshold[0]))
                            GHZ_rate_threshold = GHZ_rates[key_to_take] if key_to_take in GHZ_rates.keys() else 0
                            errorrates = round(threshold[1][0], 6), round(threshold[1][1], 6)
                            print(f"{prot}, {cut_off_time} ({colored('{:1.5f}'.format(GHZ_rate_threshold), 'blue')}): "
                                  f"{colored(round(threshold[0], 6), 'red', attrs=['bold'])} {colored(errorrates, 'red')}. "
                                  f"Numb. points involved: {colored(dict_used[prot][cut_off_time]['number_of_points'], 'blue')}. "
                                  f"Min. iterations: {colored(dict_used[prot][cut_off_time]['minimum_number_iterations'], 'blue')}. "
                                  f"Identical protocols in list of best protocols: {identical_prots_in_best_prots}. "
                                  f"Similar protocols in list of best protocols: {similar_prots_in_best_prots}.")
                            calculated_thresholds_per_set = ct.update_list_of_calculated_thresholds_per_set(set_name_new=set_name_yield, prot_name_new=prot)
                        print("")

                print(f"\n{colored(f'Best (completely calculated) threshold found for this set ({set_name_yield}):', 'grey', 'on_white', attrs=['bold'])}")
                max_threshold = 0
                max_threshold_full = None
                prot_mt = None
                prot_mcot = None
                for prot in dict_used.keys():
                    max_threshold_prot = 0
                    max_threshold_full_prot = 0
                    mcot = None
                    for cut_off_time in dict_used[prot].keys():
                        if dict_used[prot][cut_off_time]['threshold'][0] > max_threshold \
                                and dict_used[prot][cut_off_time]['threshold'][1][0] > 0 \
                                and dict_used[prot][cut_off_time]['threshold'][1][1] < 1:
                            prot_mt = prot
                            prot_mcot = cut_off_time
                            max_threshold_full = dict_used[prot][cut_off_time]['threshold']
                            max_threshold = max_threshold_full[0]
                        if dict_used[prot][cut_off_time]['threshold'][0] > max_threshold_prot \
                                and dict_used[prot][cut_off_time]['threshold'][1][0] > 0 \
                                and dict_used[prot][cut_off_time]['threshold'][1][1] < 1:
                            mcot = cut_off_time
                            max_threshold_full_prot = dict_used[prot][cut_off_time]['threshold']
                            max_threshold_prot = max_threshold_full_prot[0]
                    if mcot is not None and (explore_new_protocols_yield is False or
                                             (explore_new_protocols_yield is True and prot in best_new_prot_list
                                              and (prot, mcot, max_threshold_full_prot) not in protocols_calculated_per_set[set_name_yield])):
                        save_info = False
                        if explore_new_protocols_yield is False:
                            # print("\n\n\n\nWarning! Sometimes version 1 of the known protocols is now saved.n\n\n\n\n")
                            if set_name_yield in ["Set3h", "Set3k"]:
                                if prot in ['basic2', 'medium2', 'refined2', 'minimum4x_40_2']:
                                    prot = prot[:-1] if "minimum4x_40" not in prot else prot[:-2]
                                    save_info = True
                                elif prot not in ['basic1', 'medium1', 'refined1', 'minimum4x_40_1']:
                                    save_info = True
                            else:
                                if prot in ['basic2', 'medium2', 'refined2', 'minimum4x_40_2']:
                                    prot = prot[:-1] if "minimum4x_40" not in prot else prot[:-2]
                                    save_info = True
                                elif prot not in ['basic1', 'medium1', 'refined1', 'minimum4x_40_1']:
                                    save_info = True
                        # elif "sIIIe" not in prot:
                        #     save_info = True
                        else:
                            save_info = True
                        if save_info is True:
                            print(colored(f"Appended {(prot, mcot, max_threshold_full_prot)} for set name {set_name_yield}.", "yellow"))
                            protocols_calculated_per_set[set_name_yield].append((prot, mcot, max_threshold_full_prot))


                if max_threshold_full is not None:
                    best_thresholds_per_set[set_name_yield] = [(prot_mt, prot_mcot, max_threshold_full)]
                    print(f"{prot_mt}, {prot_mcot}: {round(max_threshold_full[0], 6)} ({round(max_threshold_full[1][0], 6)}, {round(max_threshold_full[1][1], 6)}).\n")
                    if explore_new_protocols_yield is True and round_yield == 1:
                        best_new_prot_per_set[set_name_yield] = (prot_mt, prot_mcot)
                    # if explore_new_protocols_yield is True:
                    #     protocols_calculated_per_set[set_name_yield].append((prot_mt, prot_mcot))
                    #

            if plot_cut_off_time_dependence or plot_cut_off_time_dependence_subplots:
                if cut_off_dep_plot_data_yield is not None:
                    protocol = cut_off_dep_plot_data_yield[0]
                    subplot_number = cut_off_dep_plot_data_yield[1]
                    cut_off_dict_set[subplot_number] = thresholds_fit_dict[protocol]
                    cut_off_information[subplot_number] = cut_off_dep_plot_data_yield
                else:
                    protocol = prot_cut_off_time_dependence  # list(thresholds_dict.keys())[0]
                    subplot_number = None
                    cut_off_dict_set[0] = thresholds_fit_dict[protocol]


        if output_file is True:
            print(f"\n{colored('All surface code data that is used for new calculations:', 'grey', 'on_white', attrs=['bold'])}")
            print(threshold_df_filtered.to_string(formatters={'__index__': lambda x: str(x) if isinstance(x, float) else x}))
            print(f"\n{colored(f'Settings used for new protocol calculations (set {set_name}):', 'grey', 'on_white', attrs=['bold'])}")
            # print(new_protocol_settings)
            number_of_lines = 0
            with open(r'protocols_to_calculate.txt', 'w') as fp:
                for key, value in new_protocol_settings.items():
                    for key2, value2 in value.items():
                        # write each item on a new line
                        new_line = f"({key}, {key2}, {value2})"
                        fp.write(f"\n{new_line}")
                        print(new_line)
                        number_of_lines += len(value2) if isinstance(value2, list) else 1

            print(f"\nNumber of array jobs required: {number_of_lines}.")

        if plot_cut_off_time_dependence or plot_cut_off_time_dependence_subplots:
            if cut_off_dict_set_plot is not None:
                cut_off_dict_set = cut_off_dict_set_plot
                cut_off_information = cut_off_information_plot

            print(f"cut_off_information_plot = {cut_off_information}")
            print(f"cut_off_dict_set_plot = {cut_off_dict_set}")
            markers = ["o", "s", "v", "D", "p", "^", "h", "X", "<", "P", "*", ">", "H", "d", 4, 5, 6, 7, 8, 9, 10, 11]
            for i in range(number_scaling_tries):
                font_size_big = font_size_big_start * figure_scale_factor
                font_size_small = font_size_small_start * figure_scale_factor

                if plot_cut_off_time_dependence_subplots and not plot_cut_off_subplot_one_figure:
                    fig, ax = plt.subplots(2, 2, figsize=(13.33339, 11)) if plot_style == "paper" \
                        else plt.subplots(2, 2, figsize=(9.48290, 8.5))
                    colors = plot_settings(axes=[ax[0,0], ax[0,1], ax[1,0], ax[1,1]], style=plot_style, number_colors=4, set_grid=True)
                    # ax[2, 1].set_axis_off()
                    fig.suptitle("GHZ cycle time dependence on the surface code threshold", fontsize=font_size_big)
                elif plot_cut_off_time_dependence_subplots and plot_cut_off_subplot_one_figure:
                    fig = plt.figure(figsize=(6.4314, 6.7)) if plot_style == "paper" \
                        else plt.figure(figsize=(0.8*9.48290, 7.7))
                    ax = plt.gca()
                    colors = plot_settings(axes=[ax], style=plot_style, number_colors=4, set_grid=True)
                else:
                    fig = plt.figure(figsize=(6.4314, 5))
                    ax = plt.gca()
                    colors = plot_settings(axes=[ax], style=plot_style, number_colors=4, set_grid=True)

                colors = [value for value in colors.values()]

                for subplot_nmb in [0, 3, 1, 2]: #range(len(cut_off_dict_set.items())):
                    cut_off_dict = cut_off_dict_set[subplot_nmb]
                    x_values = sorted(list(cut_off_dict.keys()))
                    x_values = [cot for cot in x_values if cot not in plot_cut_off_time_filtered_out_cots]
                    threshold_values = [cut_off_dict[cot]['threshold'] for cot in x_values]
                    y_values = [v[0] for v in threshold_values]
                    y_err_below = [v[0] - v[1][0] for v in threshold_values]
                    y_err_above = [v[1][1] - v[0] for v in threshold_values]
                    y_err = [y_err_below, y_err_above]

                    if plot_cut_off_time_dependence_subplots and not plot_cut_off_subplot_one_figure:
                        sp_row = int(subplot_nmb / 2)
                        sp_col = subplot_nmb % 2
                        plot_object = ax[sp_row, sp_col]
                        axis = plot_object
                    else:
                        plot_object = plt
                        axis = ax

                    if plot_cut_off_time_dependence_subplots:
                        protocol = convert_name[cut_off_information[subplot_nmb][0]]
                    else:
                        protocol = prot_cut_off_time_dependence

                    offset = 0.000015
                    GHZ_rates_data = []
                    for i_x, x in enumerate(x_values):
                        GHZ_rates = cut_off_dict[x]['GHZ_success_rates']
                        threshold_value = y_values[i_x]
                        key_to_take = min(GHZ_rates.keys(), key=lambda x: abs(x - threshold_value))
                        GHZ_rate_threshold = GHZ_rates[key_to_take] if key_to_take in GHZ_rates.keys() else 0
                        if round(GHZ_rate_threshold, 3) >= 0.9999:
                            if x == 0.0442801:
                                GHZ_rates_data.append(99.999)
                            elif x == 0.0904881:
                                GHZ_rates_data.append(99.991)
                            elif x == 0.0977841:
                                GHZ_rates_data.append(99.997)
                            else:
                                GHZ_rates_data.append(round(GHZ_rate_threshold * 100, 2))
                        else:
                            GHZ_rates_data.append(round(GHZ_rate_threshold * 100, 1))
                        # GHZ_success_rate = round(GHZ_rate_threshold * 100, 3)
                        # y_value = threshold_values[i_x][1][0] - offset if i_x != 4 else threshold_values[i_x][1][1] + offset/3
                        # plt.text(x, y_value, f"{GHZ_success_rate}%", size=12, ha='center')

                    if plot_cut_off_time_dependence_subplots and plot_cut_off_subplot_one_figure:
                        x_values = GHZ_rates_data
                        x_values_new, y_values_new, y_err_below_new, y_err_above_new = [], [], [], []
                        for i_value, x_value in enumerate(x_values):
                            if x_value >= 86:
                                x_values_new.append(x_value)
                                y_values_new.append(y_values[i_value])
                                y_err_below_new.append(y_err[0][i_value])
                                y_err_above_new.append(y_err[1][i_value])
                        x_values, y_values, y_err = x_values_new, y_values_new, [y_err_below_new, y_err_above_new]

                    labelplot = f"{protocol}"
                    colorplot = 'blue'
                    markerplot = 'o'
                    colorline = 'grey'
                    alphaline = 1
                    widthline = 1
                    markersizeplot = 5
                    markerfillstyle = 'full'
                    markeralpha = 1
                    if plot_cut_off_time_dependence_subplots:
                        # protocol = convert_name[cut_off_information[subplot_nmb][0]]
                        par_value = cut_off_information[subplot_nmb][2]
                        fig_information = cut_off_information[subplot_nmb][3]
                        titleplot = protocol + ' at ' + r'$' + str(par_value) + r'$' + f' in {fig_information}'
                        colorplot = colors[subplot_nmb]
                        markerplot = markers[subplot_nmb]
                        colorline = colorplot
                        alphaline = 0.4
                        widthline = 1.5
                        markerfillstyle = 'none'
                        markersizeplot = 8
                        markeralpha = 0.8
                        if plot_cut_off_subplot_one_figure:
                            labelplot = titleplot
                            if plot_style == "paper":
                                titleplot = f'GHZ completion probability dependence on threshold'
                            else:
                                titleplot = f'GHZ completion probability dependence on the surface code threshold'
                    else:
                        titleplot = f'GHZ cycle time dependence on threshold'  # {protocol.replace("_","-")}'

                    plot_object.plot(x_values, y_values, color=colorline, linestyle=':', alpha=alphaline,
                                     linewidth=widthline)
                    plot_object.errorbar(x_values, y_values, yerr=y_err, markersize=markersizeplot, color=colorplot,
                                         linestyle='None', marker=markerplot, fillstyle=markerfillstyle,
                                         label=labelplot, alpha=markeralpha)

                    # plot_object.grid(axis='y', alpha=0.75)
                    # plot_object.grid(axis='x', alpha=0.75)
                    if plot_cut_off_time_dependence_subplots and plot_cut_off_subplot_one_figure:
                        xlabel = r'GHZ completion probability $p_\mathrm{GHZ}$ (%)'
                    else:
                        xlabel = r'GHZ cycle time $t_\mathrm{GHZ}$ (s)'
                    ylabel = r'Error probability threshold $p_\mathrm{g}=p_\mathrm{m}$ (%)'
                    # axis.grid(color='0.85', linestyle='-', linewidth=1)

                    if plot_cut_off_time_dependence_subplots and plot_cut_off_subplot_one_figure:
                        # plot_object.xlim([89, 100])
                        # plot_object.xscale('log')
                        legendlabelspacing = 0.05 if plot_style == "paper" else 0.085
                        plot_object.legend(fontsize=font_size_small, loc='upper center', ncol=1,
                                           labelspacing=legendlabelspacing, bbox_to_anchor=(0, -0.32, 1, 0.2)) #bbox_to_anchor=(0, 1.02, 1, 0.2)) #, mode='expand')
                        # title_shift_upwards = 1.33 if plot_style == "paper" else 1.28
                        # plot_object.title(titleplot, fontsize=font_size_big, y=title_shift_upwards)
                        plot_object.title(titleplot, fontsize=font_size_big)
                        axis.set_xticks([86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100])
                    else:
                        axis.set_title(titleplot, fontsize=font_size_big)
                    # if not (plot_cut_off_time_dependence_subplots and sp_row in [0, 1]):
                    axis.set_xlabel(xlabel, fontsize=font_size_small)
                    if not (plot_cut_off_time_dependence_subplots
                            and not plot_cut_off_subplot_one_figure
                            and sp_col == 1):
                        axis.set_ylabel(ylabel, fontsize=font_size_small)
                    # axis.patch.set_facecolor('0.97')
                    # axis.spines["top"].set_visible(False)
                    # axis.spines["right"].set_visible(False)
                    # axis.spines["bottom"].set_visible(False)
                    # axis.spines["left"].set_visible(False)

                    def update_ticks_y(y, pos):
                        if y == 0:
                            return 0
                        else:
                            return round(y * 100, 2)

                    min_y_tick, max_y_tick = axis.get_ylim()
                    min_y_tick = math.ceil(min_y_tick * 1e4) / 1e4
                    max_y_tick = math.floor(max_y_tick * 1e4) / 1e4
                    tick_spacing = 2e-4 if (plot_cut_off_time_dependence_subplots and plot_cut_off_subplot_one_figure) \
                        else 1e-4
                    axis.set_yticks(np.arange(min_y_tick, max_y_tick + 1e-4, tick_spacing))
                    axis.yaxis.set_major_formatter(mpl.ticker.FuncFormatter(update_ticks_y))
                    if not plot_cut_off_subplot_one_figure:
                        ax2 = axis.twiny()
                        ax2.spines["top"].set_visible(False)
                        ax2.spines["right"].set_visible(False)
                        ax2.spines["bottom"].set_visible(False)
                        ax2.spines["left"].set_visible(False)
                        ax2.set_xbound(axis.get_xbound())
                        # ax.set_xticks(minor_ticks, minor=True)
                        ax2.set_xlabel(r'GHZ completion probability $p_\mathrm{GHZ}$ (%)', fontsize=font_size_small)
                        ax2.set_xticks(x_values)
                        ax2.set_xticks([], minor=True)
                        ax2.set_xticklabels(GHZ_rates_data, fontsize=font_size_small*0.93, rotation=45)

                    # axis.tick_params(axis='both', labelsize=font_size_small)

                if plot_style != "dissertation":
                    w, h = fig.get_size_inches()
                    plt.savefig(f"Cut_off_time_dependence_plot_{i}.png", bbox_inches='tight', pad_inches=0)
                    im = Image.open(f"Cut_off_time_dependence_plot_{i}.png")
                    figure_scale_factor = im.size[0] / (w * fig.dpi)
                    print(im.size[0], w * fig.dpi, figure_scale_factor)

            if plot_style != "dissertation":
                plt.savefig("Cut_off_time_dependence_plot.pdf", bbox_inches='tight', pad_inches=0)
            else:
                plt.savefig("Cut_off_time_dependence_plot.pdf")
            plt.show()

        if plot_link_efficiency_thresholds:
            # sets_link_efficiency = {'Set3b': 200, 'Set3d': 500, 'Set3c': 800, 'Set3e': 2000, 'Set3f': 20000,
            #                         'Set3g': 200000, 'Set3m': 322581, 'Set3h': 10000000, 'Set3k': 100000000}
            # sets_in_plot = {'Set3u': 300, 'Set3r': 400, 'Set3d': 500, 'Set3q': 600, 'Set3c': 800, 'Set3e': 2000,
            #                 'Set3f': 20000, 'Set3g': 200000, 'Set3h': 10000000, 'Set3k': 100000000}
            sets_in_plot = {'Set3r': 400, 'Set3d': 500, 'Set3q': 600, 'Set3c': 800, 'Set3e': 2000,
                            'Set3f': 20000, 'Set3g': 200000, 'Set3h': 10000000, 'Set3k': 100000000}
        elif plot_bell_succ_prob_thresholds:
            # f_eta_values = {'Set6f': 5, 'Set6k': 5.5, 'Set6g': 6, 'Set6m': 6.5, 'Set6h': 7, 'Set3e': 8}
            f_eta_values = {'Set6k': 5.5, 'Set6g': 6, 'Set6m': 6.5, 'Set6h': 7, 'Set3e': 8}
            # p_link values:
            sets_in_plot = {k: 0.002*10**(4+0.25*v) for k, v in f_eta_values.items()}
            # eta values:
            # sets_in_plot = {'Set6d': 0.1061, 'Set6e': 0.141421, 'Set6f': 0.1886, 'Set6k': 0.2178, 'Set6g': 0.2516,
            #                         'Set6m': 0.2904, 'Set6h': 0.3354}
            # eta^2 values:
            # sets_in_plot = {k: v**2 for k, v in sets_in_plot.items()}
            top_ticks_labels = [round(math.sqrt(2)*10**(0.125*v-1.5), 4) for v in f_eta_values.values()]
        elif plot_bell_quality_thresholds:
            # f_phi_values = {'Set5u': -4, 'Set5r': -3, 'Set5q': -2, 'Set5p': -1, 'Set5a': 0, 'Set5b': 1, 'Set5c': 2, 'Set5d': 3, 'Set5e': 4}
            f_phi_values = {'Set5r': -3, 'Set5q': -2, 'Set5p': -1, 'Set5a': 0, 'Set5b': 1, 'Set5c': 2, 'Set5d': 3, 'Set5e': 4}
            phi_values = {k: 0.84 + 0.03 * v for k, v in f_phi_values.items()}
            sets_in_plot = {k: 0.5 * (1 + v**2) for k, v in phi_values.items()}
            top_ticks_labels = [round(1-(v/(math.sqrt(0.95)*(2*0.999-1)**2))**(1/2), 2) for v in phi_values.values()]
            # sets_in_plot = {'Set5p': 0.086556, 'Set5a': 0.069796, 'Set5b': 0.0533309, 'Set5c': 0.037147,
            #                         'Set5d': 0.021231, 'Set5e': 0.0055701}
            # top_ticks_labels = [round(v,2) for k, v in phi_values.items()]

        markers = ["o", "s", "v", "D", "p", "^", "h", "X", "<", "P", "*", ">", "H", "d", 4, 5, 6, 7, 8, 9, 10, 11]
        # colors = [f"C{i % 10}" for i in range(len(markers))]
        if plot_link_efficiency_thresholds or plot_bell_succ_prob_thresholds or plot_bell_quality_thresholds:
            if protocols_to_be_plotted_per_set is None:
                protocols_calculated_per_set = dict(protocols_calculated_per_set)
            else:
                protocols_calculated_per_set = protocols_to_be_plotted_per_set
            print(f"\n\n\n\nProtocols to be plotted per set: {protocols_calculated_per_set}.")

            y_values = {}
            # for set in sets_link_efficiency.keys():
            #     best_thresholds_per_set[set] = best_thresholds_per_set['Set3c']
            # best_thresholds_per_set['Set3e'] = [('simv3_sIIIc_4_7_1', 0.0226081, [0.00166, (0.001640, 0.001685)])]

            for i in range(number_scaling_tries):
                font_size_big = font_size_big_start * figure_scale_factor
                font_size_small = font_size_small_start * figure_scale_factor

                if plot_style == "paper":
                    fig = plt.figure(figsize=(6.4314, 6))
                elif plot_style == "dissertation":
                    fig = plt.figure(figsize=(0.8 * 9.48290, 7))
                colors = plot_settings(style=plot_style, number_colors=len(markers), set_grid=True, set_minor_grid=True)
                colors = [value for value in colors.values()]
                # fig = plt.figure(figsize=(8.4314, 6))
                # plt.rc('mathtext', fontset='cm')  # 'dejavusans', 'dejavuserif', 'cm', 'stix', and 'stixsans'.
                # plt.rc('font', family='cmss10')

                prots = defaultdict(dict)

                # # Add threshold values zero to protocols that doesn't have a threshold in a set:
                # all_protocols_in_plot = []
                # for set in protocols_calculated_per_set.keys():
                #     for prot in protocols_calculated_per_set[set]:
                #         if prot[0] not in all_protocols_in_plot:
                #             all_protocols_in_plot.append(prot[0])
                #             # for set_acc in sets_link_efficiency.keys():
                #             for set_acc in sets_in_plot.keys():
                #                 prots[prot[0]][sets_in_plot[set_acc]] = [-0.0004, (0, 0)]

                for set in sets_in_plot.keys():
                    if set in protocols_calculated_per_set.keys():
                        for prot in protocols_calculated_per_set[set]:
                            # prots[(prot[0], prot[1])][sets_link_efficiency[set]] = prot[2]
                            prots[prot[0]][sets_in_plot[set]] = prot[2]

                i_col_mar = 0
                for prot in prots.keys():
                    color = colors[i_col_mar]
                    marker = markers[i_col_mar]
                    i_col_mar += 1
                    label = convert_name[prot] if prot in convert_name.keys() else prot
                    # label = prot.replace("_", "-")
                    x_values = prots[prot].keys()
                    y_values = [prots[prot][x][0] for x in x_values]
                    y_err_below = [prots[prot][x][0] - prots[prot][x][1][0] for x in x_values]
                    y_err_above = [prots[prot][x][1][1] - prots[prot][x][0] for x in x_values]
                    y_err = [y_err_below, y_err_above]
                    if prot == "Best protocol GHZ optimization":
                        if plot_link_efficiency_thresholds:
                            # This loop can be removed (to remove the new protocol data points without decoherence)
                            x_values = list(x_values)[:-2]
                            y_values = y_values[:-2]
                            y_err = [y_err_below[:-2], y_err_above[:-2]]
                            k_values = [6, 6, 6, 7, 7, 11, 11]
                            for i_value in range(len(x_values)):
                                plt.text(x_values[i_value]-x_values[i_value]/7, y_values[i_value],
                                         str(k_values[i_value]), size=font_size_small, ha='right')
                        if plot_bell_succ_prob_thresholds:
                            k_values = [6, 7, 7, 7, 7]
                            for i_value in range(len(x_values)):
                                plt.text(list(x_values)[i_value]-list(x_values)[i_value]/30, y_values[i_value],
                                         str(k_values[i_value]), size=font_size_small, ha='right')
                        if plot_bell_quality_thresholds:
                            k_values = [12, 11, 10, 9, 9, 10, 7, 7]
                            for i_value in range(len(x_values)):
                                plt.text(list(x_values)[i_value]-0.003, y_values[i_value],
                                         str(k_values[i_value]), size=font_size_small, ha='right')
                    # elif prot != "Best protocol GHZ optimization" and plot_link_efficiency_thresholds:
                    #     x_values = [xvalue for i_x, xvalue in enumerate(x_values) if i_x not in [3, 4, 5, 6]]
                    #     y_values = [yvalue for i_y, yvalue in enumerate(y_values) if i_y not in [3, 4, 5, 6]]
                    #     y_err = [[yerrb for i_yerrb, yerrb in enumerate(y_err_below) if i_yerrb not in [3, 4, 5, 6]], [yerra for i_yerra, yerra in enumerate(y_err_above) if i_yerra not in [3, 4, 5, 6]]]
                    plt.plot(x_values, y_values, color=color, linestyle=':', alpha=0.4, linewidth=1.5)
                    plt.errorbar(x_values, y_values, yerr=y_err, markersize=8, fillstyle="none", color=color,
                                 linestyle='None', marker=marker, label=f'{label}', alpha=0.8)

                # fig = plt.plot(x_values, y_values, color='grey', linestyle=':')
                # plt.errorbar(x_values, y_values, yerr=y_err, markersize=5, color='blue', linestyle='None', marker='o', label=f'{protocol}')

                # offset = 0.000015
                # for i_x, x in enumerate(x_values):
                #     GHZ_success_rate = round(thresholds_dict[protocol][x]['GHZ_success_rates'][y_values[i_x]]*100, 3)
                #     y_value = threshold_values[i_x][1][0] - offset if i_x != 4 else threshold_values[i_x][1][1] + offset/3
                #     plt.text(x, y_value, f"{GHZ_success_rate}%", size=12, ha='center')

                ax = plt.gca()
                ax.set_xscale('log') if (plot_link_efficiency_thresholds or plot_bell_succ_prob_thresholds) else None
                # plt.grid(axis='y', alpha=0.75)
                # plt.grid(axis='x', alpha=0.75)
                if plot_link_efficiency_thresholds:
                    xlabel = r'Link efficiency $\eta^{\ast}_\mathrm{link}$'
                    title = f'Thresholds at different link efficiency values'
                elif plot_bell_succ_prob_thresholds:
                    xlabel = r'Link efficiency $\eta^{\ast}_\mathrm{link}$'
                    title = f'Thresholds at different entanglement success rates'
                elif plot_bell_quality_thresholds:
                    xlabel = r'Bell pair fidelity $F_\mathrm{link}$'
                    title = f'Thresholds at different Bell pair fidelities'

                ylabel = r'Error probability threshold $p_\mathrm{g}=p_\mathrm{m}$ (%)'
                # plt.legend(bbox_to_anchor=(1.04,1), loc="upper left", prop={'size': 13})
                plt.legend(prop={'size': font_size_small})
                # ax.grid(color='0.85', linestyle='-', linewidth=1)
                ax.set_title(title, fontsize=font_size_big)
                ax.set_xlabel(xlabel, fontsize=font_size_small)
                ax.set_ylabel(ylabel, fontsize=font_size_small)
                # ax.patch.set_facecolor('0.97')
                # ax.spines["top"].set_visible(False)
                # ax.spines["right"].set_visible(False)
                # ax.spines["bottom"].set_visible(False)
                # ax.spines["left"].set_visible(False)

                x_limits = plt.gca().get_xlim()
                ypos = plt.gca().get_ylim()  # y position of the "break"
                if plot_link_efficiency_thresholds:
                    break_loc_x = [math.sqrt(10)*10**6, math.sqrt(10)*10**7] if 'Set3h' in sets_in_plot else [math.sqrt(10) * 10 ** 6]
                    major_ticks = ax.xaxis.get_ticklocs()
                    minor_ticks = ax.xaxis.get_ticklocs(minor=True)
                    minor_ticks = [x for x in minor_ticks if (x >= x_limits[0] and x <= x_limits[1] and x <= 10**6)]
                    # plt.scatter(break_loc, ypos[0], color='white', marker='s', s=80, clip_on=False, zorder=100)
                    # # draw "//" on the same place as text
                    for bl in break_loc_x:
                        plt.text(bl, ypos[0], r'//', fontsize=10, zorder=101, horizontalalignment='center',
                                 verticalalignment='center')
                    # ax.set_ylim(ypos)
                # break_loc_y = -0.0002
                # y_axis_break = r"/"
                # plt.text(x_limits[0], break_loc_y+0.00003, y_axis_break, fontsize=10, zorder=101, horizontalalignment='center',
                #          verticalalignment='center', rotation=-45)
                # plt.text(x_limits[0], break_loc_y-0.00003, y_axis_break, fontsize=10, zorder=101, horizontalalignment='center',
                #          verticalalignment='center', rotation=-45)

                y_ticks = ax.yaxis.get_ticklocs()
                y_ticks = [y for y in y_ticks if y >= 0] #+ [-0.0004]
                ax.set_yticks(y_ticks)
                def update_ticks_y(y, pos):
                    if y == -0.0004:
                        return "NT"
                    elif y == 0:
                        return 0
                    else:
                        return round(y*100, 2)
                def update_ticks_x(x, pos):
                    if x == 0:
                        return 0
                    elif x < 1 and (plot_bell_quality_thresholds or plot_link_efficiency_thresholds):
                        return round(x, 2)
                    elif x == 10**7:
                        return '(*)'
                    elif x == 10**8:
                        return '($\ddag$)'
                    else:
                        return "$10^{" + f"{int(math.log10(x))}" + "}$"
                ax.xaxis.set_major_formatter(mpl.ticker.FuncFormatter(update_ticks_x))
                ax.yaxis.set_major_formatter(mpl.ticker.FuncFormatter(update_ticks_y))

                if plot_bell_succ_prob_thresholds:
                    minor_ticks = ax.xaxis.get_ticklocs(minor=True)
                    minor_ticks = [x for x in minor_ticks if (x >= x_limits[0] and x <= x_limits[1])]
                    minor_ticks = [400] + minor_ticks
                    minor_tick_labels = [r"$" + f"{int(t / 100) if t < 1000 else int(t/1000)}" + r"\cdot10^" + f"{2 if t < 1000 else 3}" + "$" if t in [400, 600, 800, 2000] else "" for t
                                         in minor_ticks]
                    ax.set_xticks(minor_ticks, minor=True)
                    # ax.grid(which='minor', color='0.9', linestyle='-', linewidth=0.5)
                    # ax.grid(b=True, which='minor', color='r', linestyle='--')

                ax2 = ax.twiny()
                if plot_link_efficiency_thresholds or plot_bell_succ_prob_thresholds:
                    ax2.set_xscale('log')

                plot_settings(axes=[ax2], style=plot_style, set_grid=False)
                # ax2.spines["top"].set_visible(False)
                # ax2.spines["right"].set_visible(False)
                # ax2.spines["bottom"].set_visible(False)
                # ax2.spines["left"].set_visible(False)
                ax2.set_xbound(ax.get_xbound())
                if plot_link_efficiency_thresholds:
                    ax.set_xticks(minor_ticks, minor=True)
                    ax2.set_xlabel(r'Coherence time scaling factor $f_\mathrm{dec}$', fontsize=font_size_small)
                    # ax2.set_xticks([300, 400, 500, 600, 800, 2000, 20000, 200000])
                    ax2.set_xticks([400, 500, 600, 800, 2000, 20000, 200000])
                    ax2.set_xticks([], minor=True)
                    ax2.set_xticklabels([2, "", 3, 4, 10, 100, 1000], fontsize=font_size_small) #, rotation=45)
                elif plot_bell_succ_prob_thresholds:
                    ax.set_xticklabels(minor_tick_labels, fontsize=font_size_big, minor=True)
                    ax2.set_xlabel(r'Total photon detection probability $\eta_\mathrm{ph}$', fontsize=font_size_small)
                    ax2.set_xticks(list(sets_in_plot.values()))
                    ax2.set_xticks([], minor=True)
                    ax2.set_xticklabels(top_ticks_labels, fontsize=font_size_small)
                else:
                    ax2.set_xlabel(r'Excitation error probability $p_\mathrm{EE}$', fontsize=font_size_small)
                    ax2.set_xticks(list(sets_in_plot.values()))
                    ax2.set_xticks([], minor=True)
                    ax2.set_xticklabels(top_ticks_labels, fontsize=font_size_small)

                # THIS IS NOT IN AGREEMENT WITH THE OTHER FIGURES, I THINK:
                # ax.tick_params(axis='both', labelsize=font_size_small)

                if number_scaling_tries > 1:
                    # tight_bbox_raw = ax.get_tightbbox(fig.canvas.get_renderer())
                    # print(tight_bbox_raw.x0 / fig.dpi, tight_bbox_raw.x1 / fig.dpi, tight_bbox_raw.y0 / fig.dpi, tight_bbox_raw.y1 / fig.dpi)
                    # fig = ax.get_figure()
                    w, h = fig.get_size_inches()
                    # print(w, h)
                    # figure_scale_factor = (tight_bbox_raw.x1 - tight_bbox_raw.x0) / fig.dpi / w

                    # if i == number_scaling_tries - 1:
                    plt.savefig(f"Link_efficiency_plot_{i}.png", bbox_inches='tight', pad_inches=0)
                    im = Image.open(f"Link_efficiency_plot_{i}.png")
                    figure_scale_factor = im.size[0] / (w * fig.dpi)
                    print(im.size[0], w*fig.dpi, figure_scale_factor)

            if plot_style == "paper":
                plt.savefig("Link_efficiency_plot.pdf", bbox_inches='tight', pad_inches=0)
            else:
                plt.savefig("Link_efficiency_plot.pdf")
            plt.show()

    elif calc_type == 13:
        print("Running instance 13 of main.")
        # Test code to divide a requested number of iterations to multiple CPUs.

        iterations = 30
        threads = 7
        a = [*range(iterations)]

        iterations, remaining_iterations = divmod(iterations, threads)

        divisions = []
        for i in range(threads):
            iterations_thread = iterations + 1 if i < remaining_iterations else iterations
            iter_offset = iterations * i + min(i, remaining_iterations)
            divisions.append(a[iter_offset: iter_offset + iterations_thread])

        print(divisions)

    elif calc_type == 14:
        print("Running instance 14 of main.")
        # Estimate how long the dynamic algorithm will take.

        das.dynamic_algorith_calculation_estimation()

    elif calc_type == 15:
        print("Running instance 15 of main.")
        # Calculate threshold and GHZ success rate per cut-off time for a certain protocol.

        from oopsc.threshold.sim import update_result_files
        import calculate_thresholds as ct

        folder = "C:\\Users\\sebastiandebon\\surfdrive\\Code\\Sebastian\\oopsc\\superoperator\\"
        folder_rel = "Thresholds_best_prots_sIIIc\\threshold_sim\\"

        n = 4
        k = 7
        b = 1
        version = "simv3"
        set_prot = "sIIIc"
        set_calc = "Set3c"

        update_csv_files = True

        prot_name = f'dyn_prot_{version}_{set_prot}_{n}_{k}_{b}'
        file_name = "recipe_" + prot_name + "_swap_" + set_calc + "_toric_graph_3D_uf_thres.csv"

        if update_csv_files:
            update_result_files(folder + folder_rel, file_name.split(".csv")[0])

        file_location = folder + folder_rel + file_name

        ct.calculate_threshold_information(file_location, print_information=True)

    elif calc_type == 16:
        print("Running instance 16 of main.")
        # Write protocols to a file to be run in the circuit simulator.

        folder = "C:\\Users\\sebastiandebon\\surfdrive\\Code\\Sebastian\\circuit_simulation\\protocol_recipes\\"
        n = 4
        number_of_lines = 0
        p_g = 0.00219
        with open(r'protocols_to_calculate.txt', 'w') as fp:
            for k in range(4, 14):
                for b in range(1, 3):
                    file_name = f"dyn_prot_simv8T_sIIIf_{n}_{k}_{b}"
                    if os.path.exists(folder + file_name):
                        # write each item on a new line
                        line = f"({file_name}, None, {p_g})"
                        print(line)
                        fp.write(f"\n{line}")
                        number_of_lines += 1

        print(f"Number of array jobs required: {number_of_lines}.")

    elif calc_type == 17:
        print("Running instance 17 of main.")
        # Plot pdf-versions of the binary trees of protocols

        # data = pickle.load(open(f"calc_data/SetIIIe_v4/20220802_data_4_13_9_1_SetIIIe_50_mpuP_algv1.pkl", "rb"))
        #
        # # for k in range(3, 14):
        # #     for i in range(len(data[4][k])):
        # #         print(f"{k}, {i}: " + str(data[4][k][i].state))
        # # exit(0)
        #
        # n = 3
        # k = [5, 7, 8, 10]
        # b = [1, 1, 1, 1]
        # for i_k, i_b in zip(k, b):
        #     name = f"dyn_prot_simv4_sIIIe_{n}_{i_k}_{i_b}"
        #     protocol = dap.identify_protocol(data, n, i_k, i_b - 1)
        #     protocol = dap.protocol_add_meta_data(protocol)
        #     plotp.plot_protocol(protocol, 1, name="protocols_pdf/" + name)
        #     # protocol_recipe = pr.ProtocolRecipe(protocol)
        #     # protocol_recipe.print_protocol_recipe()
        # exit(0)

        folder = "C:\\Users\\sebastiandebon\\surfdrive\\Code\\GHZ_prot_II\\protocols\\ProtocolRecipe\\"
        # folder = "C:\\Users\\sebastiandebon\\surfdrive\\Code\\Sebastian\\circuit_simulation\\protocol_recipes\\"

        # prot_1 = "dyn_prot_simv4_sIIIe_3_5_1"
        # prot_2 = "dyn_prot_simv4_sIIIe_3_7_1"
        # prot_3 = "dyn_prot_simv4_sIIIe_3_8_1"
        # prot_4 = "dyn_prot_simv4_sIIIe_3_10_1"

        prot_1 = "dyn_prot_simv3_sIIIc_4_7_1"
        prot_2 = "dyn_prot_simv1_sVa_4_10_10"
        prot_3 = "dyn_prot_simv5_sIIIf_4_11_14"
        prots = [prot_1, prot_2, prot_3]

        for prot in prots: #[prot_1, prot_2, prot_3, prot_4]:
            protocol_recipe = dill.load(open(folder + prot, "rb"))
            # for i_ts in range(len(protocol_recipe.time_blocks)):
            #     for i_ssys in range(len(protocol_recipe.time_blocks[i_ts])):
            #         for i_op in range(len(protocol_recipe.time_blocks[i_ts][i_ssys].list_of_operations)):
            #             operation = protocol_recipe.time_blocks[i_ts][i_ssys].list_of_operations[i_op]
            #             print_operation(operation)
            plotp.plot_protocol(protocol_recipe.protocol, 1, name="protocols_pdf/" + prot)
            plotp.plot_protocol(protocol_recipe.protocol, 1, name="protocols_pdf2/" + prot)
            # protocol_recipe.print_protocol_recipe()


    elif calc_type == 18:
        print("Running instance 18 of main.")
        # Plot a white rectangle on the x-axis-spine to "break" it
        label_size = 10
        xpos = 10  # x position of the "break"

        ypos = plt.gca().get_ylim()[0]  # y position of the "break"
        plt.scatter(xpos, ypos, color='white', marker='s', s=80, clip_on=False, zorder=100)
        # draw "//" on the same place as text
        plt.text(xpos, ypos, r'//', fontsize=label_size, zorder=101, horizontalalignment='center',
                 verticalalignment='center')

        ax = plt.gca()
        a = ax.get_xticks().tolist()
        a[1] = 'change'
        ax.set_xticklabels(a)
        print(a)

        plt.ylim(0, 10)
        plt.show()

    elif calc_type == 19:
        print("Running instance 19 of main.")
        # Plot success probabilities and fidelity of single-click and double-click state
        mu = 0.9
        # phi_range = np.linspace(0.84, 0.98)
        phi_range = []
        alpha = 0.0667
        eta = 0.0046       #0.10605105374     #0.4474
        F_prep = 0.99
        labda = 0.984
        p_DE = 0.04
        F_1c = []
        p_succ_1c = []
        F_2c = []
        p_succ_2c = []

        phi = math.sqrt(mu) * ((2 * F_prep - 1) ** 2) * ((1 - p_DE) ** 2)

        alpha_range = []
        for alpha in np.linspace(1e-12, 0.99999999):
            alpha_range.append(alpha)
            # phi = math.sqrt(mu) * ((2 * F_prep - 1) ** 2) * ((1 - p_DE) ** 2)
            # p_succ_1c_i = (2 * eta * (1 - eta) + (1 + mu)/2 * eta ** 2) * alpha ** 2 + 2 * eta * alpha * (1 - alpha)
            p_succ_1c_i = 2 * eta * alpha + 0.5 * (eta * alpha)**2 * (mu-3)
            F_1c_i = 1 / p_succ_1c_i * (1 + phi * (2*labda - 1)) * eta * alpha * (1 - alpha)
            p_succ_1c.append(p_succ_1c_i)
            F_1c.append(F_1c_i)
            p_succ_2c.append(eta**2/2)
            F_2c.append(0.5*(1+phi**2))

        # Find alpha for fixed p_succ_1c, mu and eta:
        p_target = eta**2/2
        print(f"p_link_dc = {p_target}.")
        alpha_target = (-2*eta + math.sqrt(2*eta**2*((mu-3)*p_target + 2)))/(eta**2*(mu-3))
        F_alpha_target = 1 / p_target * (1 + phi * (2*labda - 1)) * eta * alpha_target * (1 - alpha_target)
        print(f"F_link_sc(p_link_dc) = {F_alpha_target}.")
        print(f"alpha_target = {alpha_target}.")
        print(f"F_link_dc = {F_2c[-1]}.")

        plt.figure(figsize=(14, 8))
        plt.rc('mathtext', fontset='cm')  # 'dejavusans', 'dejavuserif', 'cm', 'stix', and 'stixsans'.
        plt.rc('font', family='cmr10')

        print("")
        print(p_succ_1c)
        print(F_1c)
        print(p_succ_2c)
        print(F_2c)
        print("")

        ax = plt.gca()
        y_range = alpha_range if len(alpha_range) > 0 else phi_range
        plt.plot(y_range, F_1c, label="F-single-click", color='r')
        plt.plot(y_range, F_2c, label="F-double-click", color="y")
        ax2 = ax.twinx()
        ax2.plot(y_range, p_succ_1c, label="p-single-click")
        ax2.plot(y_range, p_succ_2c, label="p-double-click")
        ax.legend()
        ax2.legend()
        plt.show()

    elif calc_type == 20:
        print("Running instance 20 of main.")
        # Calculate fidelity of one of Fenglei's first states with the ideal GHZ state
        from scipy import sparse as sp
        import cmath
        import functools as ft
        from circuit_simulation.basic_operations.basic_operations import fidelity, trace_distance

        phi_0 = -0.3674 * np.pi
        c1 = 0.9990
        phi_1 = -0.0924 * np.pi
        cerr = 0.0456
        phi_err = -0.0789 * np.pi

        ket0 = np.array([[1], [0]])
        ket1 = np.array([[0], [1]])

        psi = 1/cmath.sqrt(2) * (cmath.exp(1j*phi_0) * ft.reduce(np.kron, [ket0] * 4)
                                 + ft.reduce(np.kron, [c1 * cmath.exp(1j*phi_1) * ket1
                                                       + cerr * cmath.exp(1j*phi_err) * ket0] * 4))

        # Normalizing psi:
        psi = psi / cmath.sqrt((np.matrix(psi).H @ np.matrix(psi))[0, 0])

        rho = sp.csr_matrix(np.matrix(psi) @ np.matrix(psi).H)

        density_matrix_target = sp.lil_matrix((16, 16))
        density_matrix_target[0, 0] = 0.5
        density_matrix_target[0, 15] = 0.5
        density_matrix_target[15, 0] = 0.5
        density_matrix_target[15, 15] = 0.5

        print(fidelity(rho, density_matrix_target))
        print(trace_distance(rho, density_matrix_target))

    elif calc_type == 21:
        print(f"Running instance 21 of main.")
        # Runs the surface code, but it forgot what exactly the point was here
        from utilities.files import get_full_path
        from oopsc.threshold.sim import sim_thresholds

        logical_successes = {}
        multithreading = True
        lattices = [6]
        sc_iters = 20*50000
        decoders = __import__("oopsc.decoder", fromlist=['uf'])
        decode = getattr(decoders, 'uf')
        surface_code_args = {'lattice_type': 'toric', 'iters': sc_iters, 'lattices': lattices, 'perror': [],
                             'superoperator_filenames': None, 'superoperator_filenames_failed': None,
                             'supop_date_and_time_values': None, 'superoperator_filenames_additional': None,
                             'superoperator_filenames_additional_failed': None, 'networked_architecture': True,
                             'measurement_error': True, 'multithreading': multithreading, 'threads': 128,
                             'modified_ansatz': False, 'save_result': False, 'show_result': False,
                             'progressbar': False, 'fbloom': 0.5, 'dg_connections': False, 'directed_graph': False,
                             'debug': False, 'cycles': None, 'iid': False, 'space_weight': 2,
                             'skip_surface_code': False, 'network_architecture_type': 'weight-4'}

        for variation in range(1):
            if variation == 1:
                supop_fn = "20221103_153108_dyn_prot_simv3_sIIIc_4_7_1_node-s3c_p_g-0.00105_p_m-0.00105_cut_off_time-0.0226081"
                supop_fol = "oopsc/superoperator/Test_stand_dev/Thresholds_best_prots_sIIIc_new_fit_lli"
                key_log_dict = "lli"
            else:
                supop_fn = "20221103_001847_dyn_prot_simv3_sIIIc_4_7_1_node-s3c_p_g-0.00105_p_m-0.00105_cut_off_time-0.0226081"
                supop_fol = "oopsc/superoperator/Test_stand_dev/Thresholds_best_prots_sIIIc_new_fit_mi"
                key_log_dict = "mi"
            logical_successes[key_log_dict] = {}
            for lat in lattices:
                logical_successes[key_log_dict][lat] = []
            fn = os.path.join(get_full_path(supop_fol), supop_fn)
            succeeded = pd.read_csv(fn + ".csv", sep=";", float_precision='round_trip')
            index = ['error_config', 'lie'] if 'error_idle' not in succeeded else ['error_stab', 'error_idle', 'lie']
            succeeded = succeeded.set_index(index)
            if os.path.exists(fn + "_failed.csv"):
                cut_off = pd.read_csv(fn + "_failed.csv", sep=";", float_precision='round_trip')
                index = ['error_config', 'lie'] if 'error_idle' not in cut_off else ['error_stab', 'error_idle', 'lie']
                cut_off = cut_off.set_index(index)
            else:
                cut_off = None

            surface_code_args['superoperator_filenames'] = [succeeded]
            surface_code_args['superoperator_filenames_failed'] = [cut_off] if cut_off is not None else None
            successful_iterations = succeeded.loc[succeeded.index[0], 'written_to']
            failed_iterations = cut_off.loc[cut_off.index[0], 'written_to'] if cut_off is not None else 0
            surface_code_args['GHZ_successes'] = [float(successful_iterations) / (float(successful_iterations) + float(failed_iterations))]

            for i in range(1000):
                data = sim_thresholds(decode, **surface_code_args)
                for i_lat, lat in enumerate(lattices):
                    logical_successes[key_log_dict][lat].append(data.loc[data.index[i_lat], 'success'])

        print(logical_successes)

    elif calc_type == 22:
        print(f"Running instance 22 of main.")
        # Learn more about  the standard deviation of the logical success rate

        # # for N = 50000:
        # N = 50000
        # logical_successes = {'lli': {6: [46025, 46025, 46188, 46105, 46115, 46119, 46089, 46085, 46040, 46094, 46211, 45998, 45997, 46125, 46013, 46040, 46042, 46083, 46062, 46083, 46096, 46172, 46136, 46027, 46036, 46082, 45988, 46165, 46093, 46017, 46117, 46155, 46130, 46079, 46098, 46116, 46051, 46090, 46021, 46156, 46053, 46185, 46023, 46060, 46055, 46141, 45993, 46128, 46102, 46088, 46057, 46155, 46094, 46176, 46154, 46072, 46086, 46127, 45976, 46107, 46112, 46154, 46040, 46168, 46104, 46078, 45935, 46026, 46132, 46083, 46196, 46084, 46125, 46070, 46098, 46064, 46066, 46057, 46037, 46108, 46183, 46087, 46126, 46058, 46085, 46019, 46159, 46058, 46058, 46109, 46081, 46169, 46102, 46091, 46295, 46023, 46121, 46063, 46011, 46058, 46062, 46050, 46085, 46100, 46083, 46008, 46110, 46144, 46145, 45977, 46103, 46006, 46060, 46134, 46004, 46086, 46053, 46122, 46136, 46120, 46078, 46078, 46079, 46114, 46089, 45987, 46001, 46145, 46164, 46169, 46119, 46112, 46170, 46075, 46143, 46076, 46036, 46116, 46010, 46053, 46057, 46103, 46085, 45982, 46095, 46081, 46088, 45996, 46019, 46104, 46087, 46087, 46074, 46011, 46141, 46024, 46153, 46078, 46202, 46127, 46088, 46134, 46125, 46166, 46168, 46010, 45953, 46003, 46101, 46137, 46084, 46125, 46067, 46062, 46017, 46127, 46082, 46066, 46028, 46088, 45996, 46139, 46017, 46073, 46096, 46168, 46096, 46138, 46008, 46129, 45963, 46029, 46052, 46057, 46012, 46100, 46006, 46079, 46116, 46057],
        #                              10: [46616, 46455, 46513, 46529, 46542, 46477, 46370, 46517, 46593, 46451, 46679, 46532, 46487, 46511, 46512, 46504, 46540, 46655, 46571, 46590, 46484, 46553, 46538, 46629, 46589, 46654, 46497, 46554, 46590, 46557, 46650, 46464, 46538, 46500, 46600, 46480, 46490, 46596, 46530, 46596, 46473, 46613, 46533, 46530, 46572, 46515, 46592, 46539, 46526, 46426, 46650, 46590, 46555, 46524, 46606, 46546, 46549, 46560, 46619, 46557, 46474, 46561, 46459, 46569, 46501, 46472, 46485, 46524, 46562, 46479, 46547, 46614, 46546, 46541, 46419, 46496, 46479, 46595, 46580, 46496, 46574, 46549, 46514, 46535, 46547, 46504, 46566, 46603, 46555, 46537, 46535, 46515, 46459, 46632, 46618, 46562, 46533, 46538, 46623, 46475, 46501, 46524, 46546, 46679, 46552, 46490, 46579, 46542, 46528, 46575, 46546, 46568, 46474, 46603, 46528, 46569, 46517, 46578, 46580, 46476, 46551, 46652, 46546, 46566, 46449, 46519, 46518, 46550, 46467, 46499, 46582, 46580, 46552, 46548, 46505, 46494, 46483, 46454, 46563, 46450, 46502, 46480, 46461, 46489, 46533, 46524, 46544, 46613, 46465, 46538, 46495, 46414, 46395, 46504, 46526, 46428, 46509, 46482, 46519, 46512, 46526, 46556, 46559, 46442, 46557, 46440, 46561, 46504, 46504, 46615, 46565, 46552, 46626, 46650, 46530, 46455, 46564, 46465, 46577, 46449, 46618, 46519, 46500, 46512, 46520, 46628, 46430, 46424, 46521, 46521, 46515, 46432, 46627, 46540, 46458, 46568, 46502, 46570, 46507, 46607]},
        #                      'mi': {6: [46091, 46191, 46163, 46096, 46111, 46221, 46110, 46127, 46177, 46034, 45947, 46145, 46191, 46071, 46133, 46059, 46185, 46050, 46124, 46040, 46092, 46057, 46214, 46210, 46146, 46062, 46048, 46090, 46195, 46115, 46115, 46208, 46158, 46216, 46144, 46143, 46167, 46167, 46154, 46118, 46096, 46202, 46160, 46066, 46053, 46116, 46094, 46177, 46108, 46142, 46045, 46074, 46143, 46078, 46288, 46191, 46114, 46049, 46163, 46155, 45998, 45977, 46125, 46131, 46130, 46099, 46086, 46109, 46083, 46206, 46051, 46199, 46154, 46090, 46169, 46134, 46119, 46119, 46204, 46064, 46151, 46148, 46164, 46093, 46105, 46101, 46183, 46139, 46081, 46076, 46053, 46022, 46169, 46081, 46063, 46079, 46131, 46044, 46077, 46183, 46131, 46054, 46061, 46134, 46086, 46141, 46048, 46198, 46067, 46089, 46164, 46161, 46022, 46140, 46190, 46106, 45960, 46064, 46172, 46152, 46131, 46153, 46061, 46182, 46190, 46099, 46126, 46094, 46016, 46033, 46088, 46180, 46094, 46086, 46003, 46168, 46035, 46143, 46121, 46054, 46098, 46010, 46208, 46133, 46031, 46098, 46104, 46110, 45967, 46097, 46063, 46093, 46068, 46020, 46118, 46174, 46042, 46136, 46122, 45931, 46070, 46029, 46067, 46106, 46116, 46176, 46025, 46106, 46112, 46029, 46126, 46044, 46125, 46161, 46200, 46123, 46105, 46134, 46022, 46085, 46152, 46096, 46143, 46118, 46058, 46210, 46075, 45889, 46130, 46079, 46040, 46233, 46152, 46110, 46186, 46066, 46141, 46125, 46060, 46087],
        #                             10: [46589, 46657, 46531, 46632, 46588, 46553, 46706, 46655, 46657, 46585, 46587, 46571, 46622, 46486, 46648, 46567, 46612, 46555, 46582, 46568, 46543, 46554, 46784, 46610, 46454, 46515, 46635, 46591, 46605, 46593, 46639, 46585, 46515, 46618, 46602, 46483, 46629, 46519, 46567, 46563, 46518, 46590, 46488, 46643, 46539, 46564, 46472, 46602, 46551, 46597, 46546, 46512, 46655, 46480, 46574, 46643, 46600, 46630, 46586, 46589, 46654, 46687, 46668, 46577, 46639, 46726, 46621, 46673, 46611, 46583, 46596, 46534, 46710, 46584, 46457, 46557, 46528, 46633, 46529, 46523, 46468, 46650, 46462, 46536, 46591, 46558, 46577, 46653, 46586, 46550, 46585, 46500, 46436, 46552, 46602, 46629, 46634, 46592, 46614, 46562, 46625, 46587, 46595, 46602, 46557, 46580, 46526, 46567, 46674, 46586, 46472, 46554, 46482, 46521, 46509, 46603, 46601, 46560, 46609, 46515, 46679, 46594, 46527, 46530, 46717, 46526, 46639, 46624, 46538, 46563, 46539, 46547, 46526, 46601, 46636, 46611, 46563, 46713, 46594, 46493, 46584, 46542, 46568, 46774, 46563, 46455, 46549, 46524, 46559, 46502, 46580, 46581, 46541, 46603, 46608, 46533, 46485, 46643, 46580, 46567, 46690, 46557, 46590, 46627, 46591, 46531, 46649, 46588, 46615, 46554, 46579, 46546, 46519, 46530, 46662, 46572, 46576, 46500, 46586, 46640, 46655, 46571, 46578, 46567, 46577, 46655, 46530, 46554, 46607, 46537, 46534, 46498, 46566, 46473, 46445, 46649, 46671, 46597, 46621, 46639]}}

        # # for N = 1000000 and 100 samples:
        # N = 1000000
        # logical_successes ={'mi': {
        #     6: [922037, 922037, 922284, 921891, 922289, 921903, 922043, 922364, 922460, 921630, 922393, 921922, 922180,
        #         922030, 922240, 922609, 922356, 921945, 922133, 922338, 922001, 922346, 922538, 922766, 922261, 922144,
        #         921787, 921906, 922546, 921887, 921757, 922793, 922131, 922115, 922212, 922011, 922224, 922622, 922355,
        #         922351, 921989, 922403, 922110, 922191, 921755, 921816, 922721, 922187, 922026, 921623, 922489, 921953,
        #         922116, 922113, 922140, 922025, 921951, 922069, 922142, 922271, 922576, 921987, 922651, 922359, 922209,
        #         922176, 922403, 922263, 921972, 922124, 922150, 921915, 922317, 922525, 921905, 922383, 922544, 922454,
        #         921699, 921913, 921836, 922034, 921945, 922080, 921714, 921839, 922277, 922005, 921739, 922079, 922384,
        #         921904, 921742, 922691, 922334, 921947, 921937, 922262, 922208, 921943]}}

        # for N = 1000000 and 1000 samples:
        N = 1000000
        logical_successes = {'mi': {6: [922216, 922336, 921896, 921996, 922496, 922155, 922120, 922314, 921992, 921712, 922014, 922327, 922436, 922106, 921607, 922123, 921806, 922312, 922529, 921977, 922221, 921923, 921893, 922845, 921744, 922133, 922212, 922326, 922433, 921745, 922196, 922165, 922254, 922239, 921491, 922150, 922301, 921973, 921363, 922187, 921873, 922702, 922166, 922219, 921951, 922008, 922231, 922484, 922507, 922513, 922279, 922255, 922033, 922060, 922468, 922503, 922623, 922111, 921711, 922414, 921910, 922243, 921933, 922007, 922175, 922323, 922316, 921933, 922334, 922121, 921979, 922035, 922127, 922366, 921972, 922054, 922465, 922373, 922004, 921794, 921983, 922286, 922409, 921834, 922242, 922496, 921905, 922495, 922021, 922339, 921891, 921981, 921519, 922377, 921504, 922055, 922117, 921984, 921893, 922341, 922162, 921691, 922459, 922144, 921858, 922034, 922266, 922196, 922028, 922209, 922602, 922147, 922119, 922561, 922254, 922142, 922074, 921943, 921858, 921937, 922265, 921844, 921988, 921776, 922772, 921879, 921994, 921918, 921751, 921957, 921662, 922222, 922069, 922611, 922000, 921509, 922136, 922302, 922095, 922563, 921887, 922302, 922117, 922014, 922447, 922020, 922184, 922153, 922418, 922155, 922407, 921857, 922014, 921956, 922337, 921816, 921951, 922661, 922481, 922069, 921732, 921721, 922126, 921940, 921854, 921554, 922125, 921960, 922390, 921997, 921531, 921403, 922119, 922212, 921991, 922180, 922113, 921931, 921957, 922423, 922326, 921943, 921807, 922244, 922283, 921984, 922218, 922166, 922041, 922226, 921826, 922200, 922282, 921919, 922030, 921977, 922068, 922152, 922163, 922159, 921980, 922478, 922085, 921951, 921990, 922335, 921885, 921925, 922392, 922330, 922101, 921711, 922512, 922104, 922372, 922275, 921858, 921499, 921960, 921912, 922351, 922624, 922276, 922363, 922106, 921876, 922079, 921575, 922093, 922006, 921843, 922179, 922460, 922073, 921628, 921801, 922065, 922234, 921808, 921895, 921480, 922330, 922604, 922016, 922508, 921767, 921919, 921882, 922529, 922337, 922486, 922080, 922097, 922001, 921946, 922275, 922075, 922457, 922440, 922166, 921790, 921654, 922138, 922265, 922231, 921902, 922499, 922073, 922110, 922068, 922379, 922382, 922266, 922450, 922382, 921741, 922003, 922520, 922664, 922190, 921629, 922199, 922794, 921984, 922272, 922153, 922126, 922258, 922504, 921624, 921836, 922547, 921999, 921968, 922006, 922140, 922355, 921605, 921822, 922248, 921960, 921791, 921492, 922203, 921983, 922408, 922204, 922333, 922057, 922519, 921930, 921698, 922145, 922096, 921844, 921644, 921894, 922509, 922210, 921884, 921796, 922501, 921994, 921759, 922304, 922653, 922096, 922053, 921957, 922195, 922035, 922088, 922203, 922437, 922135, 922440, 922307, 922498, 922412, 921799, 922265, 922519, 922383, 921974, 922146, 922586, 922007, 922698, 921894, 922438, 922047, 921651, 922054, 922504, 922352, 922176, 921884, 922432, 921585, 922360, 922529, 922094, 922189, 922651, 922058, 921984, 921384, 922402, 922310, 922515, 921579, 922268, 922376, 922090, 922151, 922557, 922337, 922232, 921668, 922659, 922404, 922331, 922157, 922352, 922208, 922177, 922059, 921642, 922085, 922476, 921516, 921300, 922804, 922171, 921870, 921981, 921846, 922420, 921803, 922485, 922106, 922164, 921786, 922994, 922019, 922527, 921917, 921843, 922110, 922158, 921877, 921853, 921963, 921879, 922327, 922457, 922240, 921954, 922056, 922022, 922201, 921764, 921983, 922037, 922007, 921914, 922433, 922760, 922008, 922028, 922460, 921698, 922353, 922027, 921888, 922323, 921315, 922159, 922098, 922350, 921707, 922054, 922474, 922044, 921751, 922208, 922332, 922176, 921885, 921977, 922433, 922042, 922462, 921821, 922221, 922181, 922129, 922368, 922271, 922399, 921865, 922331, 922315, 922379, 921956, 922037, 922171, 922032, 922321, 922489, 922579, 922464, 922063, 921614, 921900, 921808, 921912, 921986, 922221, 922505, 922359, 922133, 922047, 921252, 922442, 922287, 922288, 921944, 922052, 921962, 922306, 922199, 922012, 921978, 921678, 921880, 922411, 922014, 922194, 922361, 922113, 922324, 922005, 922211, 922548, 922390, 922532, 922192, 922305, 922298, 921896, 922286, 922539, 922403, 922099, 922105, 922063, 921885, 922157, 922288, 922573, 922354, 922271, 922059, 922530, 922060, 921805, 922232, 921807, 921828, 922371, 921683, 922209, 922028, 921958, 922051, 921735, 922204, 921684, 921971, 921852, 922018, 921809, 922331, 922378, 921813, 922045, 922114, 922140, 922113, 922275, 922009, 922292, 922184, 921917, 922266, 922486, 922185, 922262, 922062, 921496, 921777, 922058, 922584, 922197, 922432, 922017, 921896, 921474, 922131, 921919, 922145, 922455, 921978, 922362, 922106, 922526, 921938, 922286, 922257, 922291, 922290, 921995, 922454, 921989, 921722, 922166, 921963, 921856, 921955, 922366, 922705, 922086, 922344, 922130, 922143, 922254, 922186, 922226, 921720, 922322, 922677, 922184, 922318, 922070, 922134, 921797, 922023, 921647, 921906, 922387, 922635, 922330, 922078, 922247, 922178, 922275, 922191, 922431, 922745, 922331, 922107, 922128, 922061, 922010, 922059, 922397, 922186, 921713, 922280, 922330, 921806, 922098, 921909, 922465, 922035, 922450, 922019, 921986, 922206, 922197, 922431, 922145, 922072, 922011, 922007, 922121, 921992, 921956, 921857, 921512, 921729, 922150, 921647, 921979, 922464, 921933, 922227, 922152, 922576, 921961, 921727, 922168, 921449, 922256, 922205, 921703, 922129, 922195, 922155, 922238, 921943, 921724, 921768, 922083, 922570, 922364, 922607, 922074, 922265, 922409, 922175, 922491, 922182, 922433, 921894, 921830, 922499, 922134, 922228, 921963, 922038, 921992, 921840, 922135, 922185, 922291, 922349, 922246, 921623, 922265, 921951, 922134, 922306, 922342, 921859, 922265, 922135, 922093, 922004, 922163, 921734, 922446, 922299, 921711, 922333, 921941, 922264, 921874, 921915, 922051, 921940, 921685, 922126, 922387, 922652, 922257, 922549, 921774, 922100, 922280, 921783, 922285, 922326, 921847, 921880, 921967, 922058, 921910, 922141, 921927, 921873, 922002, 922208, 922069, 922127, 922158, 922081, 922205, 922220, 921957, 922044, 922462, 921950, 922032, 922699, 921838, 921897, 922437, 921711, 922323, 921646, 922125, 922474, 922243, 922529, 922233, 922384, 922399, 922000, 922622, 922503, 921920, 921899, 922115, 922045, 921900, 921853, 921998, 922243, 922238, 922343, 922412, 922205, 922555, 921869, 922107, 921713, 922202, 921884, 921794, 921555, 922195, 922003, 922328, 922138, 922018, 922050, 922466, 922137, 921929, 922283, 922280, 922323, 921952, 922098, 921811, 922444, 921576, 922171, 922500, 922041, 921936, 922134, 922043, 921927, 922179, 922234, 922573, 921905, 922144, 922146, 922015, 921517, 922472, 922615, 921985, 921874, 922326, 921801, 921709, 922266, 922086, 922664, 921794, 922219, 922245, 922266, 922251, 922210, 921935, 922589, 922430, 922070, 922435, 922398, 921915, 922286, 922011, 922319, 921859, 922242, 921939, 922483, 922234, 922058, 922247, 922031, 922227, 922032, 921617, 922583, 922271, 922050, 922463, 922098, 921901, 921851, 922303, 922144, 921761, 922004, 921766, 922111, 921874, 921991, 921856, 922442, 922113, 921658, 921841, 921655, 922637, 922597, 922833, 921981, 922296, 922018, 922124, 922420, 921719, 922448, 922159, 922360, 922113, 922066, 921973, 922161, 922128, 921865, 922261, 921876, 921210, 922115, 922258, 922052, 922266, 922248, 922233, 922164, 921859, 922197, 921511, 922108, 922196, 922182, 921511, 922148, 922451, 922153, 922395, 922475, 922319, 922277, 922005, 922168, 921954, 922109, 922051, 922047, 921823, 921609, 921993, 921825, 922282, 921951, 922177, 921740, 921674, 921941, 922378, 922161, 921927, 921742, 922352, 922459, 921575, 921769, 921587, 921767, 921797, 921745, 922270, 922487, 922386, 922077, 922750, 921787, 922302, 921954, 921667, 922190, 922152, 922222, 922160, 922219, 922597, 922308, 922008, 922071, 921916, 922343, 921960, 922342, 922146, 921730, 922419, 922628, 922310, 921983, 922319, 921895, 922096, 922157, 921815, 921889, 922297, 922290, 922185, 921916, 921991, 922573, 922051, 921788, 921999, 922130, 922255, 921749, 922401, 921907]}}

        mode = "mi"
        lattice_size = 6

        print(f"Average logical success rate over all samples: {np.mean(logical_successes[mode][lattice_size])/N}.")
        stds_per_sample = []
        for i_sample in range(len(logical_successes[mode][lattice_size])):
            t = logical_successes[mode][lattice_size][i_sample]
            stds_per_sample.append(np.sqrt(t / N * (1 - t / N) / N))
        # print(f"Standard deviations per sample: {stds_per_sample}.")
        print(f"Average of individual standard deviation per sample: {np.mean(stds_per_sample)}.")

        # These two things are equivalent:
        # all_outc = logical_successes[mode]
        # print(np.std(all_outc))
        # print(np.sqrt((1/len(all_outc))*np.sum([np.square(outc - np.mean(all_outc)) for outc in all_outc])))

        all_outc_suc_prob = [succ/N for succ in logical_successes[mode][lattice_size]]
        print(f"Standard deviation of the {len(all_outc_suc_prob)} sampled logical success rates: {np.std(all_outc_suc_prob)}.")

        plt.hist(x=all_outc_suc_prob, bins=10, alpha=0.7, rwidth=0.85)
        plt.grid(axis='y', alpha=0.75)
        plt.title(f'Logical success probability for {len(all_outc_suc_prob)} samples \nof size {N} with lattice size {lattice_size}')
        plt.xlabel('Logical success probability')
        plt.ylabel('Frequency')
        plt.show()

    elif calc_type == 23:
        print(f"Running instance 23 of main.")
        # Plot the general effect (on the Bloch sphere) of the general amplitude damping channel and phase damping channel used in the simulations
        from mpl_toolkits.mplot3d import Axes3D
        from itertools import product, combinations
        from math import sqrt, cos, sin
        from cmath import exp

        ket0 = np.matrix([1, 0])
        ket1 = np.matrix([0, 1])

        I = np.matrix([[1, 0], [0, 1]])
        X = np.matrix([[0, 1], [1, 0]])
        Y = np.matrix([[0, -1j], [1j, 0]])
        Z = np.matrix([[1, 0], [0, -1]])

        gamma1 = 0
        gad1 = 1/sqrt(2) * np.matrix([[1, 0], [0, sqrt(1-gamma1)]])
        gad2 = 1/sqrt(2) * np.matrix([[0, sqrt(gamma1)], [0, 0]])
        gad3 = 1/sqrt(2) * np.matrix([[sqrt(1-gamma1), 0], [0, 1]])
        gad4 = 1/sqrt(2) * np.matrix([[0, 0], [sqrt(gamma1), 0]])

        gamma2 = 0.95
        pd1 = np.matrix([[1, 0], [0, sqrt(1-gamma2)]])
        pd2 = np.matrix([[0, 0], [0, sqrt(gamma2)]])

        fig = plt.figure()
        ax = fig.gca(projection='3d')

        # # draw cube
        # r = [-1, 1]
        # for s, e in combinations(np.array(list(product(r, r, r))), 2):
        #     if np.sum(np.abs(s - e)) == r[1] - r[0]:
        #         ax.plot3D(*zip(s, e), color="b")

        # draw sphere
        u, v = np.mgrid[0:2*np.pi:80j, 0:np.pi:40j]
        x = np.cos(u) * np.sin(v)
        y = np.sin(u) * np.sin(v)
        z = np.cos(v)
        x_new = deepcopy(x)
        y_new = deepcopy(y)
        z_new = deepcopy(z)
        for i in range(len(x)):
            for j in range(len(x[i])):
                state = 0.5 * (I + x[i][j]*X + y[i][j]*Y + z[i][j]*Z)
                state = gad1 @ state @ gad1.T + gad2 @ state @ gad2.T + \
                        gad3 @ state @ gad3.T + gad4 @ state @ gad4.T
                state = pd1 @ state @ pd1.T + pd2 @ state @ pd2.T
                x_new[i][j] = np.real(np.trace(X @ state))
                y_new[i][j] = np.real(np.trace(Y @ state))
                z_new[i][j] = np.real(np.trace(Z @ state))
                # if not(round(x[i][j], 6) == round(np.real(x_new), 6)
                #        and round(y[i][j], 6) == round(np.real(y_new),6)
                #        and round(z[i][j], 6) == round(np.real(z_new),6)):
                #     print(x[i][j], np.real(x_new))

        # x_t, y_t, z_t = x[0][8], y[0][8], z[0][8]
        # u_t, v_t = u[0][8], v[0][8]
        # test_state = 0.5 * (I + x_t*X + y_t*Y + z_t*Z)
        # ket_t = cos(v_t/2)*ket0 + sin(v_t/2)*exp(1j*u_t)*ket1
        # rho_t = ket_t.T @ ket_t
        ax.plot_wireframe(x, y, z, color="r", linewidth=0.5, alpha=0.2)
        ax.plot_wireframe(x_new, y_new, z_new, color='b', linewidth=0.5)

        # draw a point
        ax.scatter([0], [0], [0], color="g", s=100)
        ax.set_box_aspect((np.ptp(x), np.ptp(y), np.ptp(z)))
        ax.axis('off')
        plt.show()

    elif calc_type == 24:
        print(f"Running instance 24 of main.")
        # Import the characteristics object of a certain (set of) calculation(s), and investigate whether or not the
        # superoperators have actually converged for this/these calculation(s)
        from scipy import sparse as sp
        # import cmath
        from itertools import product as comb
        import math
        import functools as ft
        from circuit_simulation.basic_operations.basic_operations import fidelity, trace_distance, trace, \
            fidelity_elementwise
        from utilities.files import filename_contains_datestamp

        folder_root = "C:\\Users\\sebastiandebon\\surfdrive\\Code\\Sebastian\\oopsc\\superoperator\\"
        folder_fn_set = {}
        pickle_fn_set = {}
        list_iterations_set = {}
        list_trace_distances_set = {}
        import_from_pickle = True
        plot_style = "dissertation"
        font_size_big_start = 16 if plot_style == "paper" else 15.06
        font_size_small_start = 14 if plot_style == "paper" else 15.06
        figure_scale_factor = 1
        number_scaling_tries = 3 if plot_style == "paper" else 1

        if import_from_pickle is False or \
                not (os.path.exists('calc_data\\trace_distance\\data_list_trace_distances_set.pkl')
                     and os.path.exists('calc_data\\trace_distance\\data_list_iterations_set.pkl')):
            # folder_fn = folder_root + "Thresholds_best_prots_all_sV_in_Set5e\\"
            # pickle_fn = "20230523_141810_dyn_prot_simv1_sVc_4_11_5_node-s5e_p_g-0.0015_p_m-0.0015"
            folder_fn = folder_root + "Thresholds_best_prots_sIIIe_trace_distance\\"
            pickle_fn = "20230524_113317_dyn_prot_simv4_sIIIe_4_7_3_node-s3e_p_g-0.0018_p_m-0.0018_cut_off_time-0.0261161"
            folder_fn_set['simv4_sIIIe_4_7_3'] = folder_fn
            pickle_fn_set['simv4_sIIIe_4_7_3'] = pickle_fn

            folder_fn = folder_root + "Thresholds_best_prots_sVp_trace_distance\\"
            pickle_fn = "20230525_141307_dyn_prot_simv2_sVp_4_14_1_node-s5p_p_g-0.0005_p_m-0.0005_cut_off_time-0.0552241"
            folder_fn_set['simv2_sVp_4_14_1'] = folder_fn
            pickle_fn_set['simv2_sVp_4_14_1'] = pickle_fn

            folder_fn = folder_root + "Thresholds_best_prots_sIIIc_trace_distance\\"
            pickle_fn = "20230616_222535_dyn_prot_simv3_sIIIc_4_7_1_node-s3c_p_g-0.0011_p_m-0.0011_cut_off_time-0.0262921"
            folder_fn_set['simv3_sIIIc_4_7_1'] = folder_fn
            pickle_fn_set['simv3_sIIIc_4_7_1'] = pickle_fn

            folder_fn = folder_root + "Thresholds_known_prots_in_Set3g_trace_distance\\"
            pickle_fn = "20230616_223216_hc_expedient_node-s3g_p_g-0.0015_p_m-0.0015_cut_off_time-0.0661681"
            folder_fn_set['expedient'] = folder_fn
            pickle_fn_set['expedient'] = pickle_fn

            folder_fn = folder_root + "Thresholds_known_prots_in_Set3g_trace_distance\\"
            pickle_fn = "20230616_230421_hc_stringent_node-s3g_p_g-0.00057_p_m-0.00057_cut_off_time-0.1792561"
            folder_fn_set['stringent'] = folder_fn
            pickle_fn_set['stringent'] = pickle_fn

            # # folder_fn = folder_root + "Thresholds_TROW_OUT\\"
            # # pickle_fn = "20230609_162047_hc_plain_node-bsetippapersucc_no-dec"

            succ_fail = 'succ'

            # Defining the building blocks that we need:
            ket0 = np.array([[1], [0]])
            ket1 = np.array([[0], [1]])
            I2 = np.array([[1, 0], [0, 1]])
            X = np.array([[0, 1], [1, 0]])
            # SWAP = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])

            # The calculated state from the simulator has the form A1 A2 B1 B2 C1 C2 D1 D2. Here, A2, B2, C2 and D2 are the
            # Choi qubits (the qubits left ``untouched'' by the circuit) and A1, B1, C1 and D1 represent the data qubits.

            # Therefore, we make the maximally entangled state in the following way here:
            Phi_plus = 1 / math.sqrt(2) * (np.kron(ket0, ket0) + np.kron(ket1, ket1))
            psi = ft.reduce(np.kron, [Phi_plus] * 4)

            # To make a maximally entangled state of the form A1 B1 C1 D1 A2 B2 C2 D2, we could use the following:
            # psi = 0
            # for state_comb in comb([ket0, ket1], repeat=4):
            #     psi += ft.reduce(np.kron, state_comb + state_comb)
            # psi = (1 / math.sqrt(2**4)) * psi

            # We ``measure'' the stabilizer operator X X X X on data qubits A1 B1 C1 D1:
            I8 = ft.reduce(np.kron, [I2] * 8)
            operator = ft.reduce(np.kron, [X, I2, X, I2, X, I2, X, I2])
            psi_plus = (I8 + operator) / math.sqrt(2) @ psi

            # We construct the ideal state in density matrix format:
            rho = sp.csr_matrix(np.matrix(psi_plus) @ np.matrix(psi_plus).H)

            for protocol_name, folder_fn in folder_fn_set.items():
                files_in_folder = os.listdir(folder_fn)
                pickle_fn = pickle_fn_set[protocol_name]
                pickle_fn_main = pickle_fn[16:] if filename_contains_datestamp(pickle_fn) else pickle_fn
                pickle_fn_main = pickle_fn_main[:-4] if pickle_fn_main.endswith(".pkl") else pickle_fn_main
                char = {}
                for filename in files_in_folder:
                    if filename_contains_datestamp(filename) and filename.endswith(".pkl"):
                        if filename[16:-4] == pickle_fn_main:
                            char_file = pickle.load(open(folder_fn + filename, 'rb'))
                            for key in char_file.keys():
                                if key in char.keys():
                                    char[key] += char_file[key]
                                else:
                                    char[key] = char_file[key]

                # PUT THIS BACK TO CALCULATE THE CONVERGENCE OF THE STABILIZER FIDELITY FROM THE STABILIZER FIDELITY DATA
                # IN THE CHARACTERISTICS FILE ITSELF:
                # # Calculate how the average stabilizer fidelity evolves over the iterations:
                # sf_bool = False if succ_fail == 'succ' else True
                # stab_fids_filt = [char['stab_fid'][x] for x in range(len(char['stab_fid'])) if char['cut_off_time'][x] is sf_bool]
                # total_avg_stab_fid = sum(stab_fids_filt)/len(stab_fids_filt) if len(stab_fids_filt) > 0 else 0
                # avg_stab_fid_build_up = 0
                # iter_stab_fid = 0
                # list_stab_fid_build_up = []
                # list_stab_fid_iters = []
                # for stab_fid in stab_fids_filt[:-1]:
                #     avg_stab_fid_build_up = (avg_stab_fid_build_up * iter_stab_fid + stab_fid) / (iter_stab_fid + 1)
                #     iter_stab_fid += 1
                #     list_stab_fid_build_up.append(abs(total_avg_stab_fid - avg_stab_fid_build_up))
                #     list_stab_fid_iters.append(iter_stab_fid)

                # char_trace_dist_succ = []
                # for ix, x in enumerate(characteristics['trace_dist_X']):
                #     if characteristics['cut_off_time'][ix] is False:
                #         char_trace_dist_succ.append(x)
                #
                # print(char_trace_dist_succ)

                # We import the Choi states calculated in the simulations and make a total average state over all
                # iterations:
                states = [char['avg_supop'][x]['X'][succ_fail] for x in range(len(char['avg_supop']))]
                states_iters = [char['avg_supop_iters'][x]['X'][succ_fail] for x in range(len(char['avg_supop_iters']))]
                total_state = sp.csr_matrix((256, 256))
                total_iters = 0
                for s_iters, state in zip(states_iters, states):
                    if total_iters + s_iters > 0:
                        total_state = (state * s_iters + total_state * total_iters) / (total_iters + s_iters)
                    total_iters += s_iters

                # We calculate the stabilizer fidelity to make sure it is in agreement with the value in the .csv file:
                final_stab_fid = fidelity_elementwise(rho, total_state)
                print(f"Final stabilizer fidelity for {protocol_name}: {final_stab_fid}.")

                # Calculate the trace distance with the final state at each stage of the calculations:
                total_state_build_up = sp.csr_matrix((256, 256))
                total_bu_iters = 0
                list_trace_distances = []
                list_iterations = []
                # list_stab_fids = []
                for s_iters, state in zip(states_iters[:-1], states[:-1]):
                    if total_bu_iters + s_iters > 0:
                        total_state_build_up = (state * s_iters + total_state_build_up * total_bu_iters) / (total_bu_iters + s_iters)
                    total_bu_iters += s_iters
                    list_trace_distances.append(trace_distance(total_state_build_up, total_state))
                    list_iterations.append(total_bu_iters)
                    # list_stab_fids.append(abs(final_stab_fid - fidelity_elementwise(rho, total_state_build_up)))
                    # print(total_bu_iters, trace_distance(total_state_build_up, total_state), abs(final_stab_fid - fidelity_elementwise(rho, total_state_build_up)))
                # print(list_trace_distances)
                list_iterations_set[protocol_name] = list_iterations
                list_trace_distances_set[protocol_name] = list_trace_distances

            with open('calc_data\\trace_distance\\data_list_trace_distances_set.pkl', 'wb') as f:
                pickle.dump(list_trace_distances_set, f)
            with open('calc_data\\trace_distance\\data_list_iterations_set.pkl', 'wb') as f:
                pickle.dump(list_iterations_set, f)
        else:
            list_trace_distances_set = pickle.load(open('calc_data\\trace_distance\\data_list_trace_distances_set.pkl', 'rb'))
            list_iterations_set = pickle.load(open('calc_data\\trace_distance\\data_list_iterations_set.pkl', 'rb'))

        for key in list_iterations_set.keys():
            iters = []
            iter_prev = 0
            for iter in list_iterations_set[key]:
                iters.append(iter - iter_prev)
                iter_prev = iter
            print(min(iters), max(iters))
        exit(0)

        plot_fonts(style=plot_style)
        # Plot the results:
        markers = ["o", "s", "v", "D", "p"]
        linestyles = ['solid', 'dotted', 'dashed', 'dashdot', (5, (10, 3))]
        convert_name = {'simv3_sIIIc_4_7_1': 'Septimum ($k=7$)',
                        'simv4_sIIIe_4_7_3': 'New protocol ($k=7$)',
                        'basic': 'Basic ($k=8$)',
                        'minimum4x_22': 'Minimum4x22 ($k=22$)',
                        'medium': 'Medium ($k=16$)',
                        'expedient': 'Expedient ($k=22$)',
                        'refined': 'Refined ($k=40$)',
                        'refined2': 'Refined ($k=40$)',
                        'minimum4x_40': 'Minimum4x40 ($k=40$)',
                        'stringent': 'Stringent ($k=42$)',
                        'simv2_sVp_4_14_1': 'New protocol ($k=14$)',
                        'simv2_sVp_4_15_1': 'New protocol ($k=15$)'}
        labelinformation = {'simv3_sIIIc_4_7_1': [r'\eta^{\ast}_\mathrm{link}=2\cdot10^3', '7', '0.18\%', 0.026],
                            'simv4_sIIIe_4_7_3': [r'\eta^{\ast}_\mathrm{link}=8\cdot10^2', '7', '0.11\%', 0.026],
                            'expedient': [r'\eta^{\ast}_\mathrm{link}=2\cdot10^5', '7', '0.15\%', 0.066],
                            'stringent': [r'\eta^{\ast}_\mathrm{link}=2\cdot10^5', '7', '0.057\%', 0.179],
                            'simv2_sVp_4_14_1': [r'F_\mathrm{link}\approx0.83', '6', '0.05\%', 0.055]}

        for i in range(number_scaling_tries):
            print("")
            font_size_big = font_size_big_start * figure_scale_factor
            font_size_small = font_size_small_start * figure_scale_factor
            print(font_size_big, font_size_small)

            fig = plt.figure(figsize=(6.4314, 6.7)) if plot_style == "paper" \
                else plt.figure(figsize=(0.8 * 9.48290, 7.7))
            ax1 = plt.gca()
            colors = plot_settings(axes=[ax1], style=plot_style, number_colors=5, set_grid=True, set_minor_grid=True)
            colors = [value for value in colors.values()]

            # fig = plt.figure(figsize=(8.4314, 6))
            # plt.rc('mathtext', fontset='cm')  # 'dejavusans', 'dejavuserif', 'cm', 'stix', and 'stixsans'.
            # plt.rc('font', family='cmss10')
            # fig, ax1 = plt.subplots()
            # colors = ['r', 'blue', 'yellow', 'green', 'purple']
            i_col = 0

            for i_protocol_name in [0, 2, 1, 3, 4]: #range(len(list_trace_distances_set)):
                protocol_name = list(list_trace_distances_set.keys())[i_protocol_name]
                list_trace_distances = list_trace_distances_set[protocol_name]
                if plot_style == "dissertation":
                    labelinfo = convert_name[protocol_name] + ', ' + r'$' + labelinformation[protocol_name][0] + r'$' \
                                + f' in Fig. {labelinformation[protocol_name][1]}, ' + r'$p_\mathrm{g}=p_\mathrm{m}=' \
                                + labelinformation[protocol_name][2] + r'$' + ', ' + r'$t_\mathrm{GHZ}\approx' \
                                + f'{labelinformation[protocol_name][3]}' + r'$' + ' s'
                else:
                    labelinfo = convert_name[protocol_name] + ', ' + r'$' + labelinformation[protocol_name][0] + r'$' \
                                + f' in Fig. {labelinformation[protocol_name][1]}, ' + r'$p_\mathrm{g}=p_\mathrm{m}=' \
                                + labelinformation[protocol_name][2] + r'$' + ', ' + r'$t_\mathrm{GHZ}\approx' \
                                + f'{labelinformation[protocol_name][3]}' + r'$' + ' s'

                ax1.plot(list_iterations_set[protocol_name], list_trace_distances, color=colors[i_col],
                         linestyle='-', linewidth=0.8, alpha=0.8, label=labelinfo)
                i_col += 1
            # ax2 = ax1.twinx()
            # ax2.plot(list_iterations, list_stab_fids, color='b', linestyle='--', marker='o', alpha=0.6)
            plottitle = 'Convergence trace distance ' + r'$T$ ' \
                        + r'$(\overline{\rho}_\mathrm{Choi},\overline{\rho}^{(i)}_\mathrm{Choi})$'
            ax1.set_xlabel('Iteration number ' + r'$i$', fontsize=font_size_small)
            ax1.set_ylabel('Trace distance with final state', fontsize=font_size_small)
            # ax1.set_ylabel('Trace distance', color='r')
            # ax2.set_ylabel(r'$|\Delta F_\mathrm{stab}|$', color='b')
            # ax1.set_xscale('log')
            ax1.set_yscale('log')
            # ax2.set_yscale('log')
            ymin1, ymax1 = ax1.get_ylim()
            ymin2, ymax2 = ax1.get_ylim()   # Change this to ax2 if plotting ``list_stab_fids''
            xmin1 = 0
            all_x_values = [item for sublist in [x for x in list_iterations_set.values()] for item in sublist]
            xmax1 = round(max(all_x_values), -4)
            # ymin, ymax = 4e-6, 4e-4
            # ax2.plot(list_stab_fid_iters, list_stab_fid_build_up, color='b', linestyle=':', alpha=0.5)
            # ax1.plot(list_stab_fid_iters, list_stab_fid_build_up, color='b', linestyle=':', alpha=0.5, label=r'$|\Delta F_\mathrm{stab}|$')
            ax1.set_ylim([min(ymin1, ymin2), max(ymax1, ymax2)])
            # ax2.set_ylim([min(ymin1, ymin2), max(ymax1, ymax2)])
            ax1.set_xlim([xmin1, xmax1])

            print(xmax1)
            x_ax_ticks = [x * xmax1 / 5 for x in range(6)]
            x_ax_ticks_minor = [x * xmax1 / 20 for x in range(1, 20)]
            x_ax_ticks_labels = []
            for x_ax_val in x_ax_ticks:
                if x_ax_val == 0:
                    val_label = 0
                else:
                    val_exp = int(math.log10(x_ax_val))
                    val_base = x_ax_val / 10 ** (val_exp)
                    val_base = round(val_base) if int(val_base) == val_base else round(val_base, 1)
                    val_label = r"$" + f"{val_base}" + r"\cdot10^{" + f"{val_exp}" + r"}$"
                x_ax_ticks_labels.append(val_label)
            print(x_ax_ticks)
            print(x_ax_ticks_minor)
            print(x_ax_ticks_labels)
            ax1.set_xticks(x_ax_ticks)
            ax1.set_xticks(x_ax_ticks_minor, minor=True)
            ax1.set_xticklabels(x_ax_ticks_labels)

            # plt.legend(fontsize=font_size_small)  # prop={'size': font_size_small})
            legendlabelspacing = 0.05 if plot_style == "paper" else 0.085
            font_size_legend = font_size_small*0.84 if plot_style == "paper" else font_size_small
            leg = plt.legend(fontsize=font_size_legend, loc='upper center', ncol=1, labelspacing=legendlabelspacing,
                             bbox_to_anchor=(0, -0.33, 1, 0.2)) # bbox_to_anchor=(0, 1.02, 1, 0.2))  # , mode='expand')
            # set the linewidth of each legend object
            for legobj in leg.legendHandles:
                legobj.set_linewidth(1.35)
            # title_shift_upwards = 1.43 if plot_style == "paper" else 1.58
            # plt.title(plottitle, fontsize=font_size_big, y=title_shift_upwards)
            plt.title(plottitle, fontsize=font_size_big)

            if plot_style != "dissertation":
                w, h = fig.get_size_inches()
                plt.savefig(f"Trace_distance_convergence_{i}.png", bbox_inches='tight', pad_inches=0)
                im = Image.open(f"Trace_distance_convergence_{i}.png")
                figure_scale_factor = im.size[0] / (w * fig.dpi)
                print(im.size[0], w * fig.dpi, figure_scale_factor)

        if plot_style != "dissertation":
            plt.savefig("Trace_distance_convergence.pdf", bbox_inches='tight', pad_inches=0)
        else:
            # plt.tight_layout()
            plt.savefig("Trace_distance_convergence.pdf")
        plt.show()

    elif calc_type == 25:
        print(f"Running instance 25 of main.")
        # Mess around with the plot style and fonts in the dissertation document
        # from matplotlib import font_manager
        # font_manager.fontManager.addfont('C:/Windows/Fonts/LibertinusMath-Regular.otf')

        # from matplotlib import font_manager
        # print("List of all fonts currently available in the matplotlib:\n")
        # print(*font_manager.findSystemFonts(fontpaths=None, fontext='ttf'), sep="\n")

        x_data = [x/100*2*math.pi for x in range(100)]
        y_data = [math.sin(x) for x in x_data]

        fig = plt.figure(figsize=(6.4314, 5))
        ax1 = plt.gca()

        plot_fonts(style="dissertation")
        colors = plot_settings(style="dissertation")

        # fig, ax1 = plt.subplots()
        ax1.plot(x_data, y_data, color=colors['lightgreen'], linestyle='-', linewidth=2, alpha=0.6, label=r'$\mathrm{sin}(x)$')
        # ax2 = ax1.twinx()
        # ax2.plot(list_iterations, list_stab_fids, color='b', linestyle='--', marker='o', alpha=0.6)
        ax1.set_xlabel(r'$F_\mathrm{prep}$')
        ax1.set_ylabel('Value')
        # ax1.set_ylabel('Trace distance', color='r')
        # ax2.set_ylabel(r'$|\Delta F_\mathrm{stab}|$', color='b')
        # ax1.set_xscale('log')
        # ax1.set_yscale('log')
        # ax2.set_yscale('log')
        # ymin1, ymax1 = ax1.get_ylim()
        # ymin2, ymax2 = ax1.get_ylim()   # Change this to ax2 if plotting ``list_stab_fids''
        # xmin1 = 0
        # xmax1 = x_data[-1]
        # ymin, ymax = 4e-6, 4e-4

        # ax1.set_ylim([min(ymin1, ymin2), max(ymax1, ymax2)])
        # ax2.set_ylim([min(ymin1, ymin2), max(ymax1, ymax2)])
        # ax1.set_xlim([xmin1, xmax1])

        # x_ax_ticks = [x * xmax1 / 5 for x in range(6)]
        # x_ax_ticks_minor = [x * xmax1 / 20 for x in range(1, 20)]
        # x_ax_ticks_labels = []
        # for x_ax_val in x_ax_ticks:
        #     if x_ax_val == 0:
        #         val_label = 0
        #     else:
        #         val_base = int(math.log10(x_ax_val))
        #         val_label = r"$" + f"{int(round(x_ax_val / 10**(val_base), 0))}" + r"\cdot10^{" + f"{val_base}" + r"}$"
        #     x_ax_ticks_labels.append(val_label)
        # print(x_ax_ticks)
        # print(x_ax_ticks_minor)
        # print(x_ax_ticks_labels)
        # ax1.set_xticks(x_ax_ticks)
        # ax1.set_xticks(x_ax_ticks_minor, minor=True)
        # ax1.set_xticklabels(x_ax_ticks_labels)   #, fontsize=font_size_small)
        plt.legend()    #prop={'size': font_size_small})
        plt.title("test title")

        # plt.tight_layout()
        plt.show()

    elif calc_type == 26:
        print(f"Running instance 26 of main.")
        # Make the plot for the 3-page conference paper
        from utilities.files import filename_contains_datestamp
        from plot_superoperator.analyse_simulation_data import confidence_interval

        folder_root = "C:\\Users\\sebastiandebon\\surfdrive\\Code\\Sebastian\\oopsc\\superoperator\\"
        folder_fn_1 = folder_root + "Thresholds_best_prots_conference_paper\\"
        folder_fn_2 = folder_root + "Thresholds_best_prots_conference_paper_success\\"
        folders_fn = [folder_fn_1, folder_fn_2]
        protocols_considered = {'dyn_prot_simv3_sIIIc_4_7_1': r'New protocol ($K=7$)',
                                'dyn_prot_simv1_sVa_4_10_10': r'New protocol ($K=10$)',
                                'dyn_prot_simv5_sIIIf_4_11_14': r'New protocol ($K=11$)',
                                'hc_plain': r'Plain ($K=3$)',
                                'hc_modicum': r'Modicum ($K=4$)',
                                'hc_basic2': r'Basic ($K=8$)',
                                'hc_expedient': r'Expedient ($K=22$)'}
        large_font = 16
        small_font = 14
        figure_scale_factor = 0.93
        number_of_tries = 1

        char_normal = {}
        char_cutoff = {}
        for i_run, folder_fn in enumerate(folders_fn):
            files_in_folder = os.listdir(folder_fn)
            char_normal[i_run] = {}
            char_cutoff[i_run] = {}
            for protocol in protocols_considered.keys():
                pickle_fn = protocol + "_node-bsetippaper" if i_run == 0 else protocol + "_node-bsetippapersucc"
                # pickle_fn_main = pickle_fn[16:] if filename_contains_datestamp(pickle_fn) else pickle_fn
                # pickle_fn_main = pickle_fn_main[:-4] if pickle_fn_main.endswith(".pkl") else pickle_fn_main
                pickle_fn_main = pickle_fn
                char_normal[i_run][protocol] = {}
                char_cutoff[i_run][protocol] = {}
                for filename in files_in_folder:
                    if filename_contains_datestamp(filename) and filename.endswith(".pkl"):
                        open_file = False
                        if filename[16:-4] == pickle_fn_main:
                            open_file = 'normal'
                        if filename[16:].split("_cut_off_time-")[0] == pickle_fn_main:
                            open_file = 'cutoff'
                        if open_file in ['normal', 'cutoff']:
                            char_file = pickle.load(open(folder_fn + filename, 'rb'))
                            for key in char_file.keys():
                                if key in eval('char_' + open_file)[i_run][protocol].keys():
                                    eval('char_' + open_file)[i_run][protocol][key] += char_file[key]
                                else:
                                    eval('char_' + open_file)[i_run][protocol][key] = char_file[key]

        type_file = "normal"
        data = {}
        for i_run in range(len(folders_fn)):
            data[i_run] = {}
            for protocol in protocols_considered.keys():
                for type_key in ['ghz_fid', 'dur']:
                    if type_key in eval('char_' + type_file)[i_run][protocol].keys():
                        values = []
                        for value_key in eval('char_' + type_file)[i_run][protocol][type_key]:
                            if value_key is not None:
                                values.append(value_key)
                        if len(values) > 0:
                            if protocol not in data[i_run].keys():
                                data[i_run][protocol] = {}
                            data[i_run][protocol]['avg_' + type_key] = sum(values)/len(values)
                            if type_key == 'dur':
                                data[i_run][protocol]['dur_ci'] = \
                                    confidence_interval(eval('char_' + type_file)[i_run][protocol]['dur'])


        for i in range(number_of_tries):
            large_font = large_font * figure_scale_factor
            small_font = small_font * figure_scale_factor

            plot_fonts(style="paper", fontsize=large_font)
            fig = plt.figure(figsize=(6.4314, 7))
            plot_settings(style="paper")
            ax1 = plt.gca()
            colors = {'cyaan': '#00A6D6', 'darkblue': '#6F1D77', 'turkoois': '#00B8C8', 'blue': '#0076C2',
                      'purple': '#6F1D77', 'pink': '#EF60A3', 'framboos': '#A50034', 'red': '#E03C31',
                      'orange': '#EC6842',
                      'yellow': '#FFB81C', 'lightgreen': '#6CC24A', 'darkgreen': '#009B77'}
            colors = [colors['cyaan'], colors['darkgreen'], colors['purple'], colors['lightgreen'], colors['yellow'],
                     colors['red'], colors['turkoois']]
            markers = ['o', 's', 'v', '^', '<', '>', 'P']

            for i_run in range(len(folders_fn)):
                for i_pr, protocol in enumerate(data[i_run].keys()):
                    data_int = data[i_run][protocol]
                    color_used = colors[i_pr]
                    alpha_used = 1 if i_run == 0 else 0.2
                    face_used = 'full' if i_run == 0 else 'none'
                    label_used = protocols_considered[protocol] if i_run == 0 else ""
                    ax1.plot(data_int['avg_dur'], data_int['avg_ghz_fid'], color=color_used, marker=markers[i_pr],
                             markersize=10, linestyle=None, linewidth=2, alpha=alpha_used, label=label_used,
                             fillstyle=face_used)
                    ax1.plot([data_int['dur_ci'][0], data_int['dur_ci'][1]], [data_int['avg_ghz_fid'], data_int['avg_ghz_fid']],
                             color=color_used, marker=None, linestyle='-', linewidth=2, alpha=alpha_used)
            ax1.set_xlabel('Time (s)')
            ax1.set_xscale('log')
            xmin1, xmax1 = ax1.get_xlim()
            ax1.set_xlim([1e-1, xmax1])

            ax1.set_ylabel('GHZ state fidelity')
            ax1.legend(fontsize=small_font, loc='lower left', mode='expand', ncol=2, bbox_to_anchor=(0,1.02,1,0.2))

            # w, h = fig.get_size_inches()
            # plt.savefig(f"Test_figure.png", bbox_inches='tight', pad_inches=0)
            # im = Image.open(f"Test_figure.png")
            # figure_scale_factor = im.size[0] / (w * fig.dpi)
            # print(im.size[0], w * fig.dpi, figure_scale_factor)

        plt.savefig("Test_figure.pdf", bbox_inches='tight', pad_inches=0)
        # plt.show()

    elif calc_type == 27:
        print(f"Running instance 27 of main.")
        # Investigating the problem with the single-click state giving different results in the simulator for situations
        # without decoherence
        from scipy import sparse as sp
        import math
        import functools as ft
        import circuit_simulation.circuit_simulator as cs

        # Defining the building blocks that we need:
        ket0 = np.array([[1], [0]])
        ket1 = np.array([[0], [1]])
        I2 = np.array([[1, 0], [0, 1]])
        Z = np.array([[1, 0], [0, -1]])
        X = np.array([[0, 1], [1, 0]])
        CZ = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]])
        CX = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])
        SWAP = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])

        bell_params = {"ent_prot":"single_click","F_prep":0.999,"p_DE":0.02,"mu":0.925,"lambda":0.991832,"eta":0.0793,"alpha":0.06325}
        # bell_params = {"ent_prot":"double_click","F_prep":0.999,"p_DE":0.01,"mu":0.95,"lambda":1,"eta":0.4474}

        bell_pair = 1 / math.sqrt(2) * (np.kron(ket0, ket0) + np.kron(ket1, ket1))
        # error_term = 1 / math.sqrt(2) * (np.kron(ket0, ket0) - np.kron(ket1, ket1))
        error_term = np.kron(ket0, ket1)
        psi = bell_pair
        F_link = 0.9
        rho = F_link * sp.csr_matrix(np.matrix(psi) @ np.matrix(psi).H) \
              + (1 - F_link) * sp.csr_matrix(np.matrix(error_term) @ np.matrix(error_term).H)
        print("Bell state Conor's paper (after X correction):")
        print(rho)
        print("")
        # Try to see if I can pass this state to the circuit simulator
        # rho[0, 0] = 0.3

        rho2 = sp.kron(rho, rho, format='csr')
        cnot_gate = sp.csr_matrix(ft.reduce(np.kron, [I2, CX, I2]))

        rho3 = cnot_gate @ rho2 @ cnot_gate.conj().T
        measure0_gate = sp.csr_matrix(ft.reduce(np.kron, [I2]*4) + ft.reduce(np.kron, [I2, I2, Z, I2]))/2
        rho4_0 = (measure0_gate @ rho3 @ measure0_gate.conj().T) / (measure0_gate @ rho3 @ measure0_gate.conj().T).diagonal().sum()

        measure1_gate = sp.csr_matrix(ft.reduce(np.kron, [I2]*4) - ft.reduce(np.kron, [I2, I2, Z, I2]))/2
        rho4_1 = (measure1_gate @ rho3 @ measure1_gate.conj().T) / (measure1_gate @ rho3 @ measure1_gate.conj().T).diagonal().sum()
        corr_gate = sp.csr_matrix(ft.reduce(np.kron, [I2, I2, X, X]))
        rho4_1_corr = corr_gate @ rho4_1 @ corr_gate.conj().T

        swap_2_3 = sp.csr_matrix(ft.reduce(np.kron, [I2, I2, SWAP]))
        print("GHZ state after 0 measurement during fusion:")
        print(swap_2_3 @ rho4_0 @ swap_2_3.conj().T)
        print("")
        print("GHZ state after 1 measurement during fusion:")
        print(swap_2_3 @ rho4_1_corr @ swap_2_3.conj().T)
        print("")

        qc = cs.QuantumCircuit(8, 0, decoherence=False, p_g=0.01, p_m=0.01, p_m_1=0.1, F_link=0.1, noise=True,
                               bell_pair_type=3, network_noise_type=3, bell_pair_parameters=bell_params,
                               no_single_qubit_error=True)
        qc.define_node("C", qubits=[3, 7])
        qc.define_node("B", qubits=[2, 1, 6, 5])
        qc.define_node("A", qubits=[0, 4])
        qc.define_sub_circuit("ABC")
        qc.define_sub_circuit("AB")
        qc.define_sub_circuit("BC")

        qc.start_sub_circuit("AB")
        qc.create_bell_pair(1, 0)
        qc.start_sub_circuit("BC")
        qc.create_bell_pair(3, 2)
        qc.CNOT(1, 2)
        qc.measure(2, basis="Z", outcome=0)
        # qc._qubit_density_matrix_lookup.update({3: (rho, [3, 2]), 2: (rho, [3, 2]),
        #                                         1: (rho, [1, 0]), 0: (rho, [1, 0])})
        # qc._update_uninitialised_qubit_register([3, 2, 1, 0], update_type="remove")
        post_meas_state = qc.get_combined_density_matrix([0, 1, 2, 3])[0]
        print("GHZ state after 0 measurement during fusion, circuit simulator:")
        print(post_meas_state)
        print("")

        qc.start_sub_circuit("AB")
        qc.create_bell_pair(5, 4)
        qc.start_sub_circuit("BC")
        qc.create_bell_pair(7, 6)
        qc.CNOT(5, 6)
        qc.measure(6, basis="Z", outcome=1)
        qc.X(6)
        qc.X(7)
        post_meas_state = qc.get_combined_density_matrix([4, 5, 6, 7])[0]
        print("GHZ state after 1 measurement during fusion, circuit simulator:")
        print(post_meas_state)
